# include "Tree.inc"

typedef tTree * yyPtrtTree;

extern FILE * yyf;
extern yyRecursionLevel;
extern int yyLabelCount;

# define yyMapToTree(yyLabel) yyTreeStorePtr [yyLabel]
extern tTree * yyTreeStorePtr;

# define yyNil	0374
# define yyNoLabel	0375
# define yyLabelDef	0376
# define yyLabelUse	0377

static void yyPut
# if defined __STDC__ | defined __cplusplus
 (char * yyx, int yysize)
# else
 (yyx, yysize) char * yyx; int yysize;
# endif
{ (void) fwrite (yyx, 1, yysize, yyf); }

static void yyPutIdent
# if defined __STDC__ | defined __cplusplus
 (tIdent yyi)
# else
 (yyi) tIdent yyi;
# endif
{
 char yys [256];
 GetString (yyi, (tString) yys);
 (void) fprintf (yyf, "%s\n", yys);
}

static void yyPutTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyt)
# else
 (yyt) tTree yyt;
# endif
{
 static Tree_tLabel yyLabel;
 for (;;) {
  if (yyt == NoTree) {
   (void) putc (yyNil, yyf); return;
  } else if (yyt->yyHead.yyMark == 0) {
   (void) putc (yyLabelUse, yyf); yyLabel = yyMapToLabel (yyt); yyPut ((char *) & yyLabel, sizeof (yyLabel)); return;
  } else if (yyt->yyHead.yyMark > 1) {
   (void) putc (yyLabelDef, yyf); yyLabel = yyMapToLabel (yyt); yyPut ((char *) & yyLabel, sizeof (yyLabel));
   yyPut ((char *) & yyt->Kind, sizeof (Tree_tKind));
  } else if (yyt->Kind > 251) {
   (void) putc (yyNoLabel, yyf); yyPut ((char *) & yyt->Kind, sizeof (Tree_tKind));
  } else {
   (void) putc ((char) yyt->Kind, yyf);
  }
  yyt->yyHead.yyMark = 0;

  switch (yyt->Kind) {
case kCOMP_UNIT:
puttTree (yyt->COMP_UNIT.COMP_ELEMENTS)
yyt = yyt->COMP_UNIT.USED_ELEMENTS; break;
case kUNIT_LIST:
puttTree (yyt->UNIT_LIST.Elem)
yyt = yyt->UNIT_LIST.Next; break;
case kUNIT_NODE:
puttIdent (yyt->UNIT_NODE.Ident)
putint (yyt->UNIT_NODE.Line)
puttDefinitions (yyt->UNIT_NODE.Object)
return;
case kMODULE_DECL:
puttIdent (yyt->MODULE_DECL.Ident)
putint (yyt->MODULE_DECL.Line)
puttDefinitions (yyt->MODULE_DECL.Object)
yyt = yyt->MODULE_DECL.MODULE_BODY; break;
case kPROGRAM_DECL:
puttIdent (yyt->PROGRAM_DECL.Ident)
putint (yyt->PROGRAM_DECL.Line)
puttDefinitions (yyt->PROGRAM_DECL.Object)
puttTree (yyt->PROGRAM_DECL.FORMALS)
yyt = yyt->PROGRAM_DECL.PROGRAM_BODY; break;
case kPROC_DECL:
puttIdent (yyt->PROC_DECL.Ident)
putint (yyt->PROC_DECL.Line)
puttDefinitions (yyt->PROC_DECL.Object)
puttTree (yyt->PROC_DECL.FORMALS)
putbool (yyt->PROC_DECL.IsRecursive)
putbool (yyt->PROC_DECL.IsPure)
puttIdent (yyt->PROC_DECL.HPFExtrinsic)
yyt = yyt->PROC_DECL.PROC_BODY; break;
case kFUNC_DECL:
puttIdent (yyt->FUNC_DECL.Ident)
putint (yyt->FUNC_DECL.Line)
puttDefinitions (yyt->FUNC_DECL.Object)
puttTree (yyt->FUNC_DECL.FORMALS)
puttTree (yyt->FUNC_DECL.FUNC_BODY)
puttIdent (yyt->FUNC_DECL.RESULT_ID)
putbool (yyt->FUNC_DECL.IsRecursive)
putbool (yyt->FUNC_DECL.IsPure)
puttIdent (yyt->FUNC_DECL.HPFExtrinsic)
yyt = yyt->FUNC_DECL.RESULT_TYPE; break;
case kMODULE_PROC_DECL:
puttIdent (yyt->MODULE_PROC_DECL.Ident)
putint (yyt->MODULE_PROC_DECL.Line)
puttDefinitions (yyt->MODULE_PROC_DECL.Object)
return;
case kBLOCK_DATA_DECL:
puttIdent (yyt->BLOCK_DATA_DECL.Ident)
putint (yyt->BLOCK_DATA_DECL.Line)
puttDefinitions (yyt->BLOCK_DATA_DECL.Object)
yyt = yyt->BLOCK_DATA_DECL.DATA_BODY; break;
case kINTERFACE:
puttIdent (yyt->INTERFACE.Ident)
putint (yyt->INTERFACE.Line)
puttDefinitions (yyt->INTERFACE.Object)
puttTree (yyt->INTERFACE.SPEC)
yyt = yyt->INTERFACE.ITEMS; break;
case kBODY_NODE:
puttTree (yyt->BODY_NODE.DECLS)
puttTree (yyt->BODY_NODE.STATS)
yyt = yyt->BODY_NODE.INTERNALS; break;
case kINTEGER_TYPE:
yyt = yyt->INTEGER_TYPE.SIZE; break;
case kREAL_TYPE:
yyt = yyt->REAL_TYPE.SIZE; break;
case kBOOLEAN_TYPE:
yyt = yyt->BOOLEAN_TYPE.SIZE; break;
case kCOMPLEX_TYPE:
yyt = yyt->COMPLEX_TYPE.SIZE; break;
case kSTRING_TYPE:
yyt = yyt->STRING_TYPE.LENGTH; break;
case kTYPE_ID:
yyt = yyt->TYPE_ID.ID; break;
case kARRAY_TYPE:
puttTree (yyt->ARRAY_TYPE.ARRAY_INDEX_TYPES)
yyt = yyt->ARRAY_TYPE.ARRAY_COMP_TYPE; break;
case kRECORD_TYPE:
puttIdent (yyt->RECORD_TYPE.Name)
yyt = yyt->RECORD_TYPE.COMPONENTS; break;
case kPOINTER_TYPE:
yyt = yyt->POINTER_TYPE.PTR_COMP; break;
case kSHAPE_SPEC:
yyt = yyt->SHAPE_SPEC.Overlap; break;
case kEXPLICIT_SHAPE:
puttTree (yyt->EXPLICIT_SHAPE.Overlap)
puttTree (yyt->EXPLICIT_SHAPE.LOWER)
yyt = yyt->EXPLICIT_SHAPE.UPPER; break;
case kASSUMED_SHAPE:
puttTree (yyt->ASSUMED_SHAPE.Overlap)
yyt = yyt->ASSUMED_SHAPE.LOWER; break;
case kDEFERRED_SHAPE:
yyt = yyt->DEFERRED_SHAPE.Overlap; break;
case kASSUMED_SIZE:
puttTree (yyt->ASSUMED_SIZE.Overlap)
yyt = yyt->ASSUMED_SIZE.LOWER; break;
case kSHAPE_LIST:
puttTree (yyt->SHAPE_LIST.Elem)
yyt = yyt->SHAPE_LIST.Next; break;
case kDECL_NODE:
puttIdent (yyt->DECL_NODE.Ident)
putint (yyt->DECL_NODE.Line)
return;
case kTYPE_DECL:
puttIdent (yyt->TYPE_DECL.Ident)
putint (yyt->TYPE_DECL.Line)
yyt = yyt->TYPE_DECL.VAL; break;
case kNAME_DECL:
puttIdent (yyt->NAME_DECL.Ident)
putint (yyt->NAME_DECL.Line)
return;
case kRENAME_DECL:
puttIdent (yyt->RENAME_DECL.Ident)
putint (yyt->RENAME_DECL.Line)
puttIdent (yyt->RENAME_DECL.oldname)
return;
case kENTITY_DECL:
puttIdent (yyt->ENTITY_DECL.Ident)
putint (yyt->ENTITY_DECL.Line)
yyt = yyt->ENTITY_DECL.ATTRIBUTES; break;
case kVAR_DECL:
puttIdent (yyt->VAR_DECL.Ident)
putint (yyt->VAR_DECL.Line)
yyt = yyt->VAR_DECL.VAL; break;
case kCOMMON_DECL:
puttIdent (yyt->COMMON_DECL.Ident)
putint (yyt->COMMON_DECL.Line)
yyt = yyt->COMMON_DECL.IDS; break;
case kNAMELIST_DECL:
puttIdent (yyt->NAMELIST_DECL.Ident)
putint (yyt->NAMELIST_DECL.Line)
yyt = yyt->NAMELIST_DECL.IDS; break;
case kTEMPLATE_DECL:
puttIdent (yyt->TEMPLATE_DECL.Ident)
putint (yyt->TEMPLATE_DECL.Line)
yyt = yyt->TEMPLATE_DECL.DIMENSIONS; break;
case kRAGGED_DECL:
puttIdent (yyt->RAGGED_DECL.Ident)
putint (yyt->RAGGED_DECL.Line)
yyt = yyt->RAGGED_DECL.DIMENSIONS; break;
case kPROCESSORS_DECL:
puttIdent (yyt->PROCESSORS_DECL.Ident)
putint (yyt->PROCESSORS_DECL.Line)
yyt = yyt->PROCESSORS_DECL.DIMENSIONS; break;
case kSUB_PROCS_DECL:
puttIdent (yyt->SUB_PROCS_DECL.Ident)
putint (yyt->SUB_PROCS_DECL.Line)
puttTree (yyt->SUB_PROCS_DECL.TOP)
yyt = yyt->SUB_PROCS_DECL.DIMENSIONS; break;
case kINHERIT_DECL:
puttIdent (yyt->INHERIT_DECL.Ident)
putint (yyt->INHERIT_DECL.Line)
return;
case kRANGE_DECL:
puttIdent (yyt->RANGE_DECL.Ident)
putint (yyt->RANGE_DECL.Line)
yyt = yyt->RANGE_DECL.RANGES; break;
case kSTMT_FUNC_DECL:
puttIdent (yyt->STMT_FUNC_DECL.Ident)
putint (yyt->STMT_FUNC_DECL.Line)
puttTree (yyt->STMT_FUNC_DECL.FORMALS)
puttTree (yyt->STMT_FUNC_DECL.RESULT_TYPE)
yyt = yyt->STMT_FUNC_DECL.FFUNC_BODY; break;
case kINTERFACE_DECL:
puttIdent (yyt->INTERFACE_DECL.Ident)
putint (yyt->INTERFACE_DECL.Line)
puttTree (yyt->INTERFACE_DECL.SPEC)
yyt = yyt->INTERFACE_DECL.ITEMS; break;
case kVAR_PARAM_DECL:
puttIdent (yyt->VAR_PARAM_DECL.Ident)
putint (yyt->VAR_PARAM_DECL.Line)
yyt = yyt->VAR_PARAM_DECL.VAL; break;
case kRET_PARAM_DECL:
puttIdent (yyt->RET_PARAM_DECL.Ident)
putint (yyt->RET_PARAM_DECL.Line)
return;
case kIMPLICIT_DECL:
puttIdent (yyt->IMPLICIT_DECL.Ident)
putint (yyt->IMPLICIT_DECL.Line)
puttIdent (yyt->IMPLICIT_DECL.first)
puttIdent (yyt->IMPLICIT_DECL.last)
yyt = yyt->IMPLICIT_DECL.TYPE; break;
case kUSE_DECL:
puttIdent (yyt->USE_DECL.Ident)
putint (yyt->USE_DECL.Line)
puttIdent (yyt->USE_DECL.use)
yyt = yyt->USE_DECL.RENAMINGS; break;
case kONLY_USE_DECL:
puttIdent (yyt->ONLY_USE_DECL.Ident)
putint (yyt->ONLY_USE_DECL.Line)
puttIdent (yyt->ONLY_USE_DECL.use)
yyt = yyt->ONLY_USE_DECL.NAMES; break;
case kEQV_DECL:
puttIdent (yyt->EQV_DECL.Ident)
putint (yyt->EQV_DECL.Line)
yyt = yyt->EQV_DECL.VARS; break;
case kDATA_DECL:
puttIdent (yyt->DATA_DECL.Ident)
putint (yyt->DATA_DECL.Line)
puttTree (yyt->DATA_DECL.VARS)
yyt = yyt->DATA_DECL.VALS; break;
case kFORMAT_DECL:
puttIdent (yyt->FORMAT_DECL.Ident)
putint (yyt->FORMAT_DECL.Line)
putint (yyt->FORMAT_DECL.label)
yyt = yyt->FORMAT_DECL.FORMAT_PARAMS; break;
case kDIMENSION_DECL:
puttIdent (yyt->DIMENSION_DECL.Ident)
putint (yyt->DIMENSION_DECL.Line)
yyt = yyt->DIMENSION_DECL.INDEXES; break;
case kINIT_DATA_DECL:
puttIdent (yyt->INIT_DATA_DECL.Ident)
putint (yyt->INIT_DATA_DECL.Line)
yyt = yyt->INIT_DATA_DECL.VAL; break;
case kTYPESPEC_DECL:
puttIdent (yyt->TYPESPEC_DECL.Ident)
putint (yyt->TYPESPEC_DECL.Line)
yyt = yyt->TYPESPEC_DECL.VAL; break;
case kALLOCATABLE_DECL:
puttIdent (yyt->ALLOCATABLE_DECL.Ident)
putint (yyt->ALLOCATABLE_DECL.Line)
return;
case kEXTERNAL_DECL:
puttIdent (yyt->EXTERNAL_DECL.Ident)
putint (yyt->EXTERNAL_DECL.Line)
return;
case kINTENT_DECL:
puttIdent (yyt->INTENT_DECL.Ident)
putint (yyt->INTENT_DECL.Line)
putint (yyt->INTENT_DECL.intent)
return;
case kINTRINSIC_DECL:
puttIdent (yyt->INTRINSIC_DECL.Ident)
putint (yyt->INTRINSIC_DECL.Line)
puttIdent (yyt->INTRINSIC_DECL.generic)
putint (yyt->INTRINSIC_DECL.class)
putint (yyt->INTRINSIC_DECL.kind)
return;
case kOPTIONAL_DECL:
puttIdent (yyt->OPTIONAL_DECL.Ident)
putint (yyt->OPTIONAL_DECL.Line)
return;
case kPOINTER_DECL:
puttIdent (yyt->POINTER_DECL.Ident)
putint (yyt->POINTER_DECL.Line)
return;
case kSAVE_DECL:
puttIdent (yyt->SAVE_DECL.Ident)
putint (yyt->SAVE_DECL.Line)
return;
case kTARGET_DECL:
puttIdent (yyt->TARGET_DECL.Ident)
putint (yyt->TARGET_DECL.Line)
return;
case kPARAMETER_DECL:
puttIdent (yyt->PARAMETER_DECL.Ident)
putint (yyt->PARAMETER_DECL.Line)
yyt = yyt->PARAMETER_DECL.VAL; break;
case kPUBLIC_DECL:
puttIdent (yyt->PUBLIC_DECL.Ident)
putint (yyt->PUBLIC_DECL.Line)
return;
case kPRIVATE_DECL:
puttIdent (yyt->PRIVATE_DECL.Ident)
putint (yyt->PRIVATE_DECL.Line)
return;
case kSEQUENCE_DECL:
puttIdent (yyt->SEQUENCE_DECL.Ident)
putint (yyt->SEQUENCE_DECL.Line)
return;
case kDISTRIBUTE_DECL:
puttIdent (yyt->DISTRIBUTE_DECL.Ident)
putint (yyt->DISTRIBUTE_DECL.Line)
puttTree (yyt->DISTRIBUTE_DECL.DISTRIBUTION)
yyt = yyt->DISTRIBUTE_DECL.TARGET; break;
case kALIGN_DECL:
puttIdent (yyt->ALIGN_DECL.Ident)
putint (yyt->ALIGN_DECL.Line)
puttTree (yyt->ALIGN_DECL.ALIGN_SOURCE)
yyt = yyt->ALIGN_DECL.ALIGN_TARGET; break;
case kMAP_DECL:
puttIdent (yyt->MAP_DECL.Ident)
putint (yyt->MAP_DECL.Line)
yyt = yyt->MAP_DECL.MAP_TARGET; break;
case kOVERLAP_DECL:
puttIdent (yyt->OVERLAP_DECL.Ident)
putint (yyt->OVERLAP_DECL.Line)
yyt = yyt->OVERLAP_DECL.OVERLAPS; break;
case kSHARED_DECL:
puttIdent (yyt->SHARED_DECL.Ident)
putint (yyt->SHARED_DECL.Line)
putint (yyt->SHARED_DECL.kind)
return;
case kDYNAMIC_DECL:
puttIdent (yyt->DYNAMIC_DECL.Ident)
putint (yyt->DYNAMIC_DECL.Line)
return;
case kTRACE_DECL:
puttIdent (yyt->TRACE_DECL.Ident)
putint (yyt->TRACE_DECL.Line)
return;
case kTREE_DECL:
puttIdent (yyt->TREE_DECL.Ident)
putint (yyt->TREE_DECL.Line)
return;
case kSELECT_DECL:
puttIdent (yyt->SELECT_DECL.Ident)
putint (yyt->SELECT_DECL.Line)
yyt = yyt->SELECT_DECL.SELECTIONS; break;
case kHPF_SEQUENCE_DECL:
puttIdent (yyt->HPF_SEQUENCE_DECL.Ident)
putint (yyt->HPF_SEQUENCE_DECL.Line)
return;
case kNOSEQUENCE_DECL:
puttIdent (yyt->NOSEQUENCE_DECL.Ident)
putint (yyt->NOSEQUENCE_DECL.Line)
return;
case kLAYOUT_DECL:
puttIdent (yyt->LAYOUT_DECL.Ident)
putint (yyt->LAYOUT_DECL.Line)
puttStringRef (yyt->LAYOUT_DECL.kind)
return;
case kPASS_BY_DECL:
puttIdent (yyt->PASS_BY_DECL.Ident)
putint (yyt->PASS_BY_DECL.Line)
puttStringRef (yyt->PASS_BY_DECL.kind)
return;
case kMAP_TO_DECL:
puttIdent (yyt->MAP_TO_DECL.Ident)
putint (yyt->MAP_TO_DECL.Line)
puttStringRef (yyt->MAP_TO_DECL.kind)
return;
case kDECL_LIST:
puttTree (yyt->DECL_LIST.Elem)
yyt = yyt->DECL_LIST.Next; break;
case kRANGE_LIST:
puttTree (yyt->RANGE_LIST.Elem)
yyt = yyt->RANGE_LIST.Next; break;
case kSELECT_LIST:
puttTree (yyt->SELECT_LIST.Elem)
yyt = yyt->SELECT_LIST.Next; break;
case kSELECT_SPEC:
putint (yyt->SELECT_SPEC.selector)
return;
case kDISTRIBUTION_SPEC:
putint (yyt->DISTRIBUTION_SPEC.is_descriptive)
return;
case kANY_DISTRIBUTION:
putint (yyt->ANY_DISTRIBUTION.is_descriptive)
return;
case kDEFAULT_DISTRIBUTION:
putint (yyt->DEFAULT_DISTRIBUTION.is_descriptive)
return;
case kREPL_DISTRIBUTION:
putint (yyt->REPL_DISTRIBUTION.is_descriptive)
return;
case kNODE_DISTRIBUTION:
putint (yyt->NODE_DISTRIBUTION.is_descriptive)
yyt = yyt->NODE_DISTRIBUTION.MAPPING; break;
case kDIST_LIST:
puttTree (yyt->DIST_LIST.Elem)
yyt = yyt->DIST_LIST.Next; break;
case kBLOCK_DISTRIBUTION:
yyt = yyt->BLOCK_DISTRIBUTION.SIZE; break;
case kCYCLIC_DISTRIBUTION:
yyt = yyt->CYCLIC_DISTRIBUTION.SIZE; break;
case kGEN_BLOCK_DISTRIBUTION:
yyt = yyt->GEN_BLOCK_DISTRIBUTION.ARRAY; break;
case kINDIRECT_DISTRIBUTION:
yyt = yyt->INDIRECT_DISTRIBUTION.ARRAY; break;
case kARBITRARY_DISTRIBUTION:
puttTree (yyt->ARBITRARY_DISTRIBUTION.SIZE)
puttTree (yyt->ARBITRARY_DISTRIBUTION.LENGTH)
yyt = yyt->ARBITRARY_DISTRIBUTION.MAP; break;
case kPROCESSOR_ARRAY:
yyt = yyt->PROCESSOR_ARRAY.TOPNAME; break;
case kPROCESSOR_SUBSET:
puttTree (yyt->PROCESSOR_SUBSET.SUBTOP)
puttTree (yyt->PROCESSOR_SUBSET.FULLTOP)
yyt = yyt->PROCESSOR_SUBSET.SUBSCRIPTS; break;
case kALIGN_SPEC:
putint (yyt->ALIGN_SPEC.is_descriptive)
puttTree (yyt->ALIGN_SPEC.TEMPNAME)
yyt = yyt->ALIGN_SPEC.SUBSCRIPTS; break;
case kID_GENERIC_SPEC:
puttIdent (yyt->ID_GENERIC_SPEC.name)
return;
case kOP_GENERIC_SPEC:
yyt = yyt->OP_GENERIC_SPEC.OPERATOR; break;
case kACF_LIST:
puttTree (yyt->ACF_LIST.Elem)
yyt = yyt->ACF_LIST.Next; break;
case kACF_NODE:
putint (yyt->ACF_NODE.Label)
putint (yyt->ACF_NODE.Line)
return;
case kACF_DUMMY:
putint (yyt->ACF_DUMMY.Label)
putint (yyt->ACF_DUMMY.Line)
return;
case kACF_BASIC:
putint (yyt->ACF_BASIC.Label)
putint (yyt->ACF_BASIC.Line)
yyt = yyt->ACF_BASIC.BASIC_STMT; break;
case kACF_IF:
putint (yyt->ACF_IF.Label)
putint (yyt->ACF_IF.Line)
puttTree (yyt->ACF_IF.IF_EXP)
puttTree (yyt->ACF_IF.THEN_PART)
yyt = yyt->ACF_IF.ELSE_PART; break;
case kACF_WHERE:
putint (yyt->ACF_WHERE.Label)
putint (yyt->ACF_WHERE.Line)
puttTree (yyt->ACF_WHERE.WHERE_EXP)
puttTree (yyt->ACF_WHERE.TRUE_PART)
yyt = yyt->ACF_WHERE.FALSE_PART; break;
case kACF_CASE:
putint (yyt->ACF_CASE.Label)
putint (yyt->ACF_CASE.Line)
puttTree (yyt->ACF_CASE.CASE_EXP)
yyt = yyt->ACF_CASE.CASE_ALTS; break;
case kACF_WHILE:
putint (yyt->ACF_WHILE.Label)
putint (yyt->ACF_WHILE.Line)
puttTree (yyt->ACF_WHILE.WHILE_EXP)
yyt = yyt->ACF_WHILE.WHILE_BODY; break;
case kACF_REPEAT:
putint (yyt->ACF_REPEAT.Label)
putint (yyt->ACF_REPEAT.Line)
puttTree (yyt->ACF_REPEAT.REPEAT_BODY)
yyt = yyt->ACF_REPEAT.REPEAT_EXP; break;
case kACF_LOOP:
putint (yyt->ACF_LOOP.Label)
putint (yyt->ACF_LOOP.Line)
yyt = yyt->ACF_LOOP.LOOP_BODY; break;
case kACF_DO:
putint (yyt->ACF_DO.Label)
putint (yyt->ACF_DO.Line)
puttTree (yyt->ACF_DO.DO_ID)
puttTree (yyt->ACF_DO.DO_RANGE)
puttTree (yyt->ACF_DO.DO_BODY)
puttTree (yyt->ACF_DO.DO_DEP_INFO)
yyt = yyt->ACF_DO.DO_HOME_INFO; break;
case kACF_FORALL:
putint (yyt->ACF_FORALL.Label)
putint (yyt->ACF_FORALL.Line)
puttTree (yyt->ACF_FORALL.FORALL_ID)
puttTree (yyt->ACF_FORALL.FORALL_RANGE)
puttTree (yyt->ACF_FORALL.FORALL_BODY)
puttTree (yyt->ACF_FORALL.FORALL_DEP_INFO)
yyt = yyt->ACF_FORALL.FORALL_HOME_INFO; break;
case kACF_FLOW_GRAPH:
putint (yyt->ACF_FLOW_GRAPH.Label)
putint (yyt->ACF_FLOW_GRAPH.Line)
yyt = yyt->ACF_FLOW_GRAPH.FLOW_NODES; break;
case kACF_ENTRY:
putint (yyt->ACF_ENTRY.Label)
putint (yyt->ACF_ENTRY.Line)
yyt = yyt->ACF_ENTRY.ENTRY_DECL; break;
case kACF_HOME:
putint (yyt->ACF_HOME.Label)
putint (yyt->ACF_HOME.Line)
puttTree (yyt->ACF_HOME.HOME_VAR)
putbool (yyt->ACF_HOME.user_home)
yyt = yyt->ACF_HOME.HOME_BODY; break;
case kACF_RESIDENT:
putint (yyt->ACF_RESIDENT.Label)
putint (yyt->ACF_RESIDENT.Line)
puttTree (yyt->ACF_RESIDENT.RESIDENT_VAR)
yyt = yyt->ACF_RESIDENT.RESIDENT_BODY; break;
case kACF_NEW:
putint (yyt->ACF_NEW.Label)
putint (yyt->ACF_NEW.Line)
puttTree (yyt->ACF_NEW.NEW_VAR)
yyt = yyt->ACF_NEW.NEW_BODY; break;
case kACF_REDUCTION:
putint (yyt->ACF_REDUCTION.Label)
putint (yyt->ACF_REDUCTION.Line)
putbool (yyt->ACF_REDUCTION.is_zero)
puttTree (yyt->ACF_REDUCTION.REDUCTION_VAR)
puttTree (yyt->ACF_REDUCTION.REDUCTION_LVAR)
puttTree (yyt->ACF_REDUCTION.REDUCTION_FUNC)
yyt = yyt->ACF_REDUCTION.REDUCTION_BODY; break;
case kACF_TASK_REGION:
putint (yyt->ACF_TASK_REGION.Label)
putint (yyt->ACF_TASK_REGION.Line)
putint (yyt->ACF_TASK_REGION.task_flag)
yyt = yyt->ACF_TASK_REGION.TASK_BODY; break;
case kACF_PARALLEL:
putint (yyt->ACF_PARALLEL.Label)
putint (yyt->ACF_PARALLEL.Line)
puttTree (yyt->ACF_PARALLEL.REDUCTION_VARS)
puttTree (yyt->ACF_PARALLEL.SHARED_VARS)
puttTree (yyt->ACF_PARALLEL.LOCAL_VARS)
yyt = yyt->ACF_PARALLEL.PARALLEL_BODY; break;
case kACF_CRITICAL:
putint (yyt->ACF_CRITICAL.Label)
putint (yyt->ACF_CRITICAL.Line)
yyt = yyt->ACF_CRITICAL.CRITICAL_BODY; break;
case kACF_ON:
putint (yyt->ACF_ON.Label)
putint (yyt->ACF_ON.Line)
puttTree (yyt->ACF_ON.ON_HOME)
puttTree (yyt->ACF_ON.ON_SPECS)
yyt = yyt->ACF_ON.ON_BODY; break;
case kACF_RM_READ:
putint (yyt->ACF_RM_READ.Label)
putint (yyt->ACF_RM_READ.Line)
puttTree (yyt->ACF_RM_READ.LOCAL_VAR)
yyt = yyt->ACF_RM_READ.REMOTE_VAR; break;
case kACF_RM_WRITE:
putint (yyt->ACF_RM_WRITE.Label)
putint (yyt->ACF_RM_WRITE.Line)
putint (yyt->ACF_RM_WRITE.op)
puttTree (yyt->ACF_RM_WRITE.REMOTE_VAR)
yyt = yyt->ACF_RM_WRITE.LOCAL_DATA; break;
case kACF_MOVE:
putint (yyt->ACF_MOVE.Label)
putint (yyt->ACF_MOVE.Line)
yyt = yyt->ACF_MOVE.MOVE_STMT; break;
case kACF_BARRIER:
putint (yyt->ACF_BARRIER.Label)
putint (yyt->ACF_BARRIER.Line)
yyt = yyt->ACF_BARRIER.BARRIER_VARS; break;
case kACF_UPDATE:
putint (yyt->ACF_UPDATE.Label)
putint (yyt->ACF_UPDATE.Line)
yyt = yyt->ACF_UPDATE.OVERLAP_VAR; break;
case kON_VAR_CLAUSE:
yyt = yyt->ON_VAR_CLAUSE.ON_VAR; break;
case kON_PROC_CLAUSE:
yyt = yyt->ON_PROC_CLAUSE.ON_PROC; break;
case kON_PROCS:
puttObject (yyt->ON_PROCS.topid)
return;
case kON_VAR:
puttObject (yyt->ON_VAR.topid)
puttObject (yyt->ON_VAR.tempid)
puttObject (yyt->ON_VAR.arrayid)
return;
case kON_LIST:
puttTree (yyt->ON_LIST.Elem)
yyt = yyt->ON_LIST.Next; break;
case kON_SPEC:
putint (yyt->ON_SPEC.top_dim)
putint (yyt->ON_SPEC.temp_dim)
putint (yyt->ON_SPEC.array_dim)
yyt = yyt->ON_SPEC.ON_VAL; break;
case kSELECTED_ACF_LIST:
puttTree (yyt->SELECTED_ACF_LIST.Elem)
yyt = yyt->SELECTED_ACF_LIST.Next; break;
case kSELECTED_ACF_NODE:
puttTree (yyt->SELECTED_ACF_NODE.SELECT_LIST)
yyt = yyt->SELECTED_ACF_NODE.SELECT_ACFS; break;
case kASSIGN_STMT:
puttTree (yyt->ASSIGN_STMT.ASSIGN_VAR)
yyt = yyt->ASSIGN_STMT.ASSIGN_EXP; break;
case kPTR_ASSIGN_STMT:
puttTree (yyt->PTR_ASSIGN_STMT.ASSIGN_VAR)
yyt = yyt->PTR_ASSIGN_STMT.ASSIGN_EXP; break;
case kLABEL_ASSIGN_STMT:
putint (yyt->LABEL_ASSIGN_STMT.assign_label)
yyt = yyt->LABEL_ASSIGN_STMT.LABEL_VAR; break;
case kCALL_STMT:
puttTree (yyt->CALL_STMT.CALL_ID)
yyt = yyt->CALL_STMT.CALL_PARAMS; break;
case kIO_STMT:
puttTree (yyt->IO_STMT.ID)
puttTree (yyt->IO_STMT.IO_SPECS)
yyt = yyt->IO_STMT.IO_ITEMS; break;
case kGOTO_STMT:
putint (yyt->GOTO_STMT.GOTO_LABEL)
return;
case kASS_GOTO_STMT:
puttTree (yyt->ASS_GOTO_STMT.GOTO_VAR)
yyt = yyt->ASS_GOTO_STMT.LABELS; break;
case kCOMP_GOTO_STMT:
puttTree (yyt->COMP_GOTO_STMT.GOTO_LABELS)
yyt = yyt->COMP_GOTO_STMT.GOTO_EXP; break;
case kCOMP_IF_STMT:
putint (yyt->COMP_IF_STMT.IF_LT_LABEL)
putint (yyt->COMP_IF_STMT.IF_EQ_LABEL)
putint (yyt->COMP_IF_STMT.IF_GT_LABEL)
yyt = yyt->COMP_IF_STMT.IF_EXP; break;
case kRETURN_STMT:
yyt = yyt->RETURN_STMT.RETURN_EXP; break;
case kPAUSE_STMT:
yyt = yyt->PAUSE_STMT.PAUSE_CONST; break;
case kEXIT_STMT:
puttIdent (yyt->EXIT_STMT.loopid)
return;
case kCYCLE_STMT:
puttIdent (yyt->CYCLE_STMT.loopid)
return;
case kSTOP_STMT:
yyt = yyt->STOP_STMT.STOP_CONST; break;
case kFORMAT_STMT:
yyt = yyt->FORMAT_STMT.FORMAT_PARAMS; break;
case kINHERIT_DSP_STMT:
puttTree (yyt->INHERIT_DSP_STMT.VAR)
putint (yyt->INHERIT_DSP_STMT.flag)
putint (yyt->INHERIT_DSP_STMT.safety)
yyt = yyt->INHERIT_DSP_STMT.SHAPE; break;
case kCREATE_DSP_STMT:
puttTree (yyt->CREATE_DSP_STMT.VAR)
putint (yyt->CREATE_DSP_STMT.flag)
yyt = yyt->CREATE_DSP_STMT.SHAPE; break;
case kFREE_DSP_STMT:
putint (yyt->FREE_DSP_STMT.flag)
yyt = yyt->FREE_DSP_STMT.VAR; break;
case kALLOCATE_STMT:
puttTree (yyt->ALLOCATE_STMT.PARAMS)
yyt = yyt->ALLOCATE_STMT.STATUS; break;
case kDEALLOCATE_STMT:
puttTree (yyt->DEALLOCATE_STMT.PARAMS)
yyt = yyt->DEALLOCATE_STMT.STATUS; break;
case kNULLIFY_STMT:
yyt = yyt->NULLIFY_STMT.PARAMS; break;
case kREDUCE_STMT:
puttTree (yyt->REDUCE_STMT.RED_FUNC)
yyt = yyt->REDUCE_STMT.RED_PARAMS; break;
case kMOVE_STMT:
puttTree (yyt->MOVE_STMT.TARGET)
puttTree (yyt->MOVE_STMT.SOURCE)
yyt = yyt->MOVE_STMT.MASK; break;
case kSCATTER_STMT:
puttIdent (yyt->SCATTER_STMT.func)
puttTree (yyt->SCATTER_STMT.ARRAY)
puttTree (yyt->SCATTER_STMT.BASE)
puttTree (yyt->SCATTER_STMT.INDEXES)
yyt = yyt->SCATTER_STMT.MASK; break;
case kREALIGN_STMT:
puttTree (yyt->REALIGN_STMT.ALIGNEE)
puttTree (yyt->REALIGN_STMT.ALIGN_SOURCE)
puttDefinitions (yyt->REALIGN_STMT.distribution)
yyt = yyt->REALIGN_STMT.ALIGN_TARGET; break;
case kREDISTRIBUTE_STMT:
puttTree (yyt->REDISTRIBUTE_STMT.DISTRIBUTEE)
puttTree (yyt->REDISTRIBUTE_STMT.DISTRIBUTION)
puttDefinitions (yyt->REDISTRIBUTE_STMT.distribution)
yyt = yyt->REDISTRIBUTE_STMT.TARGET; break;
case kREG_SHADOW_GET:
puttTree (yyt->REG_SHADOW_GET.SHADOW_VAR)
yyt = yyt->REG_SHADOW_GET.SHADOW_SIZE; break;
case kREG_SHADOW_PUT:
putint (yyt->REG_SHADOW_PUT.op)
puttTree (yyt->REG_SHADOW_PUT.SHADOW_VAR)
yyt = yyt->REG_SHADOW_PUT.SHADOW_SIZE; break;
case kREG_SHADOW_SET:
puttTree (yyt->REG_SHADOW_SET.SHADOW_VAR)
puttTree (yyt->REG_SHADOW_SET.SHADOW_SIZE)
yyt = yyt->REG_SHADOW_SET.VAL; break;
case kIND_SHADOW_CREATE:
putint (yyt->IND_SHADOW_CREATE.id)
puttTree (yyt->IND_SHADOW_CREATE.SHADOW_VAR)
puttTree (yyt->IND_SHADOW_CREATE.INDEX_VAR)
puttTree (yyt->IND_SHADOW_CREATE.TMP_INDEX_VAR)
yyt = yyt->IND_SHADOW_CREATE.MASK; break;
case kIND_SHADOW_FREE:
putint (yyt->IND_SHADOW_FREE.id)
yyt = yyt->IND_SHADOW_FREE.INDEX_VAR; break;
case kIND_SHADOW_GET:
putint (yyt->IND_SHADOW_GET.id)
yyt = yyt->IND_SHADOW_GET.SHADOW_VAR; break;
case kIND_SHADOW_PUT:
putint (yyt->IND_SHADOW_PUT.id)
putint (yyt->IND_SHADOW_PUT.op)
yyt = yyt->IND_SHADOW_PUT.SHADOW_VAR; break;
case kIND_SHADOW_SET:
putint (yyt->IND_SHADOW_SET.id)
puttTree (yyt->IND_SHADOW_SET.SHADOW_VAR)
yyt = yyt->IND_SHADOW_SET.VAL; break;
case kBROADCAST_STMT:
puttTree (yyt->BROADCAST_STMT.ELEMS)
puttTree (yyt->BROADCAST_STMT.OWNER_TOP)
yyt = yyt->BROADCAST_STMT.OWNER_SPEC; break;
case kREDUCTION_STMT:
puttTree (yyt->REDUCTION_STMT.REDFUNC)
puttTree (yyt->REDUCTION_STMT.ELEMS)
puttTree (yyt->REDUCTION_STMT.RED_TOP)
yyt = yyt->REDUCTION_STMT.RED_SPEC; break;
case kDIRTY_STMT:
yyt = yyt->DIRTY_STMT.ELEMS; break;
case kLABEL_LIST:
putint (yyt->LABEL_LIST.Elem)
yyt = yyt->LABEL_LIST.Next; break;
case kOP_DEFINED:
puttIdent (yyt->OP_DEFINED.opname)
return;
case kCONST_EXP:
yyt = yyt->CONST_EXP.C; break;
case kOP_EXP:
puttTree (yyt->OP_EXP.EXP_OP)
puttTree (yyt->OP_EXP.OPND1)
yyt = yyt->OP_EXP.OPND2; break;
case kOP1_EXP:
puttTree (yyt->OP1_EXP.EXP_OP1)
yyt = yyt->OP1_EXP.OPND; break;
case kSLICE_EXP:
puttTree (yyt->SLICE_EXP.START)
puttTree (yyt->SLICE_EXP.STOP)
yyt = yyt->SLICE_EXP.INC; break;
case kVAR_EXP:
yyt = yyt->VAR_EXP.V; break;
case kFUNC_CALL_EXP:
puttTree (yyt->FUNC_CALL_EXP.FUNC_ID)
yyt = yyt->FUNC_CALL_EXP.FUNC_PARAMS; break;
case kNAMED_EXP:
puttIdent (yyt->NAMED_EXP.Name)
yyt = yyt->NAMED_EXP.VAL; break;
case kDO_EXP:
puttTree (yyt->DO_EXP.DO_ID)
puttTree (yyt->DO_EXP.RANGE)
yyt = yyt->DO_EXP.BODY; break;
case kTYPE_EXP:
puttTree (yyt->TYPE_EXP.ID)
yyt = yyt->TYPE_EXP.ELEMENTS; break;
case kARRAY_EXP:
yyt = yyt->ARRAY_EXP.ELEMENTS; break;
case kPERM_EXP:
putPermutation (yyt->PERM_EXP.perm)
yyt = yyt->PERM_EXP.VAL; break;
case kBOUND_EXP:
putint (yyt->BOUND_EXP.dim)
putint (yyt->BOUND_EXP.kind)
putint (yyt->BOUND_EXP.local)
yyt = yyt->BOUND_EXP.VAR; break;
case kRANK_EXP:
putint (yyt->RANK_EXP.kind)
return;
case kBTE_LIST:
puttTree (yyt->BTE_LIST.Elem)
yyt = yyt->BTE_LIST.Next; break;
case kUSED_VAR:
yyt = yyt->USED_VAR.VARNAME; break;
case kINDEXED_VAR:
puttTree (yyt->INDEXED_VAR.IND_VAR)
yyt = yyt->INDEXED_VAR.IND_EXPS; break;
case kSUBSTRING_VAR:
puttTree (yyt->SUBSTRING_VAR.IND_VAR)
yyt = yyt->SUBSTRING_VAR.IND_EXP; break;
case kSELECTED_VAR:
puttTree (yyt->SELECTED_VAR.SELEC_VAR)
yyt = yyt->SELECTED_VAR.SELECTOR; break;
case kLOOP_VAR:
yyt = yyt->LOOP_VAR.LOOP_VARNAME; break;
case kPERM_VAR:
putPermutation (yyt->PERM_VAR.perm)
yyt = yyt->PERM_VAR.VAL; break;
case kADDR:
yyt = yyt->ADDR.E; break;
case kDO_VAR:
puttTree (yyt->DO_VAR.DO_ID)
puttTree (yyt->DO_VAR.RANGE)
yyt = yyt->DO_VAR.BODY; break;
case kREMOTE_VAR:
yyt = yyt->REMOTE_VAR.VAR; break;
case kBTV_LIST:
puttTree (yyt->BTV_LIST.Elem)
yyt = yyt->BTV_LIST.Next; break;
case kVAR_PARAM:
puttDefinitions (yyt->VAR_PARAM.formal)
putint (yyt->VAR_PARAM.intent)
putint (yyt->VAR_PARAM.layout)
putint (yyt->VAR_PARAM.pass_by)
putint (yyt->VAR_PARAM.trace)
yyt = yyt->VAR_PARAM.V; break;
case kVALUE_PARAM:
yyt = yyt->VALUE_PARAM.E; break;
case kNAMED_PARAM:
puttIdent (yyt->NAMED_PARAM.Name)
yyt = yyt->NAMED_PARAM.VAL; break;
case kPROC_PARAM:
yyt = yyt->PROC_PARAM.P; break;
case kFUNC_PARAM:
yyt = yyt->FUNC_PARAM.F; break;
case kFORMAT_PARAM:
puttStringRef (yyt->FORMAT_PARAM.S)
return;
case kRETURN_PARAM:
putint (yyt->RETURN_PARAM.label)
return;
case kNO_PARAM:
putint (yyt->NO_PARAM.type)
return;
case kBTP_LIST:
puttTree (yyt->BTP_LIST.Elem)
yyt = yyt->BTP_LIST.Next; break;
case kBOOL_CONSTANT:
putint (yyt->BOOL_CONSTANT.value)
return;
case kINT_CONSTANT:
putint (yyt->INT_CONSTANT.value)
return;
case kREAL_CONSTANT:
puttStringRef (yyt->REAL_CONSTANT.value)
putint (yyt->REAL_CONSTANT.kind)
return;
case kSTRING_CONSTANT:
puttStringRef (yyt->STRING_CONSTANT.value)
return;
case kCOMPLEX_CONSTANT:
puttStringRef (yyt->COMPLEX_CONSTANT.rvalue)
puttStringRef (yyt->COMPLEX_CONSTANT.ivalue)
return;
case kPROC_OBJ:
puttIdent (yyt->PROC_OBJ.Ident)
puttDefinitions (yyt->PROC_OBJ.Object)
return;
case kVAR_OBJ:
putint (yyt->VAR_OBJ.Pos)
puttIdent (yyt->VAR_OBJ.Ident)
puttDefinitions (yyt->VAR_OBJ.Object)
puttReaching (yyt->VAR_OBJ.Reaching)
return;
case kTYPE_OBJ:
puttIdent (yyt->TYPE_OBJ.Ident)
puttDefinitions (yyt->TYPE_OBJ.Object)
return;
case kREC_COMP:
puttIdent (yyt->REC_COMP.Ident)
puttDefinitions (yyt->REC_COMP.Object)
return;
case kTOP_OBJ:
puttIdent (yyt->TOP_OBJ.Ident)
puttDefinitions (yyt->TOP_OBJ.Object)
return;
case kTEMP_OBJ:
puttIdent (yyt->TEMP_OBJ.Ident)
puttDefinitions (yyt->TEMP_OBJ.Object)
return;
case kOVERLAP_SPEC:
putint (yyt->OVERLAP_SPEC.left_size)
putint (yyt->OVERLAP_SPEC.right_size)
putint (yyt->OVERLAP_SPEC.left_update)
putint (yyt->OVERLAP_SPEC.right_update)
return;
case kINDEP_INFO:
putbool (yyt->INDEP_INFO.user_independent)
putint (yyt->INDEP_INFO.selection)
putInfoTree (yyt->INDEP_INFO.red_vars)
putInfoTree (yyt->INDEP_INFO.new_vars)
return;
case kVECTOR_INFO:
putbool (yyt->VECTOR_INFO.is_independent)
putint (yyt->VECTOR_INFO.selection)
putint (yyt->VECTOR_INFO.loop_cnt)
putInfoTree (yyt->VECTOR_INFO.independent_vars)
return;
case kCOMM_INFO:
putInfoTree (yyt->COMM_INFO.home_var)
putbool (yyt->COMM_INFO.is_local)
return;
case kON_INFO:
puttObject (yyt->ON_INFO.on_id)
putint (yyt->ON_INFO.on_dim)
putint (yyt->ON_INFO.base)
putint (yyt->ON_INFO.stride)
return;
case kPARALLEL_INFO:
putInfoTree (yyt->PARALLEL_INFO.reduction_vars)
putInfoTree (yyt->PARALLEL_INFO.home_var)
putInfoTree (yyt->PARALLEL_INFO.new_vars)
putInfoTree (yyt->PARALLEL_INFO.resident_vars)
return;
case kREACHING_INFO:
putInfoTree (yyt->REACHING_INFO.var_allocate)
putInfoTree (yyt->REACHING_INFO.temp_allocate)
puttDefinitions (yyt->REACHING_INFO.var_distribution)
puttDefinitions (yyt->REACHING_INFO.temp_distribution)
return;
  default: return;
  }
 }
}

void PutTree
# if defined __STDC__ | defined __cplusplus
 (FILE * yyyf, tTree yyt)
# else
 (yyyf, yyt) FILE * yyyf; tTree yyt;
# endif
{
 yyf = yyyf;
 if (yyRecursionLevel ++ == 0) yyLabelCount = 0;
 yyMark (yyt);
 yyPutTree (yyt);
 yyRecursionLevel --;
}

static void yyGet
# if defined __STDC__ | defined __cplusplus
 (char * yyx, int yysize)
# else
 (yyx, yysize) char * yyx; int yysize;
# endif
{ (void) fread (yyx, 1, yysize, yyf); }

static void yyGetIdent
# if defined __STDC__ | defined __cplusplus
 (tIdent * yyi)
# else
 (yyi) tIdent * yyi;
# endif
{
 char yys [256];
 (void) fscanf (yyf, "%s%*c", yys);
 * yyi = MakeIdent ((tString) yys, strlen (yys));
}

static void yyGetTree
# if defined __STDC__ | defined __cplusplus
 (yyPtrtTree yyt)
# else
 (yyt) yyPtrtTree yyt;
# endif
{
 static Tree_tLabel yyLabel;
 static Tree_tKind yyKind;
 for (;;) {
  switch (yyKind = getc (yyf)) {
  case yyNil		: * yyt = NoTree; return;
  case yyLabelUse	: yyGet ((char *) & yyLabel, sizeof (yyLabel));
   * yyt = yyMapToTree (yyLabel); return;
  case yyLabelDef	: yyGet ((char *) & yyLabel, sizeof (yyLabel));
   yyGet ((char *) & yyKind, sizeof (yyKind));
   * yyt = MakeTree (yyKind);
   if (yyLabel != yyMapToLabel (* yyt)) { (void) fputs ("Tree: error in GetTree\n", stderr); Tree_Exit (); } break;
  case yyNoLabel	: yyGet ((char *) & yyKind, sizeof (yyKind));
   * yyt = MakeTree (yyKind); break;
  default	: * yyt = MakeTree (yyKind);
  }

  switch (yyKind) {
case kCOMP_UNIT:
gettTree (& ((* yyt)->COMP_UNIT.COMP_ELEMENTS))
yyt = & ((* yyt)->COMP_UNIT.USED_ELEMENTS); break;
case kUNIT_LIST:
gettTree (& ((* yyt)->UNIT_LIST.Elem))
yyt = & ((* yyt)->UNIT_LIST.Next); break;
case kUNIT_NODE:
gettIdent ((* yyt)->UNIT_NODE.Ident)
getint ((* yyt)->UNIT_NODE.Line)
gettDefinitions ((* yyt)->UNIT_NODE.Object)
return;
case kMODULE_DECL:
gettIdent ((* yyt)->MODULE_DECL.Ident)
getint ((* yyt)->MODULE_DECL.Line)
gettDefinitions ((* yyt)->MODULE_DECL.Object)
yyt = & ((* yyt)->MODULE_DECL.MODULE_BODY); break;
case kPROGRAM_DECL:
gettIdent ((* yyt)->PROGRAM_DECL.Ident)
getint ((* yyt)->PROGRAM_DECL.Line)
gettDefinitions ((* yyt)->PROGRAM_DECL.Object)
gettTree (& ((* yyt)->PROGRAM_DECL.FORMALS))
yyt = & ((* yyt)->PROGRAM_DECL.PROGRAM_BODY); break;
case kPROC_DECL:
gettIdent ((* yyt)->PROC_DECL.Ident)
getint ((* yyt)->PROC_DECL.Line)
gettDefinitions ((* yyt)->PROC_DECL.Object)
gettTree (& ((* yyt)->PROC_DECL.FORMALS))
getbool ((* yyt)->PROC_DECL.IsRecursive)
getbool ((* yyt)->PROC_DECL.IsPure)
gettIdent ((* yyt)->PROC_DECL.HPFExtrinsic)
yyt = & ((* yyt)->PROC_DECL.PROC_BODY); break;
case kFUNC_DECL:
gettIdent ((* yyt)->FUNC_DECL.Ident)
getint ((* yyt)->FUNC_DECL.Line)
gettDefinitions ((* yyt)->FUNC_DECL.Object)
gettTree (& ((* yyt)->FUNC_DECL.FORMALS))
gettTree (& ((* yyt)->FUNC_DECL.FUNC_BODY))
gettIdent ((* yyt)->FUNC_DECL.RESULT_ID)
getbool ((* yyt)->FUNC_DECL.IsRecursive)
getbool ((* yyt)->FUNC_DECL.IsPure)
gettIdent ((* yyt)->FUNC_DECL.HPFExtrinsic)
yyt = & ((* yyt)->FUNC_DECL.RESULT_TYPE); break;
case kMODULE_PROC_DECL:
gettIdent ((* yyt)->MODULE_PROC_DECL.Ident)
getint ((* yyt)->MODULE_PROC_DECL.Line)
gettDefinitions ((* yyt)->MODULE_PROC_DECL.Object)
return;
case kBLOCK_DATA_DECL:
gettIdent ((* yyt)->BLOCK_DATA_DECL.Ident)
getint ((* yyt)->BLOCK_DATA_DECL.Line)
gettDefinitions ((* yyt)->BLOCK_DATA_DECL.Object)
yyt = & ((* yyt)->BLOCK_DATA_DECL.DATA_BODY); break;
case kINTERFACE:
gettIdent ((* yyt)->INTERFACE.Ident)
getint ((* yyt)->INTERFACE.Line)
gettDefinitions ((* yyt)->INTERFACE.Object)
gettTree (& ((* yyt)->INTERFACE.SPEC))
yyt = & ((* yyt)->INTERFACE.ITEMS); break;
case kBODY_NODE:
gettTree (& ((* yyt)->BODY_NODE.DECLS))
gettTree (& ((* yyt)->BODY_NODE.STATS))
yyt = & ((* yyt)->BODY_NODE.INTERNALS); break;
case kINTEGER_TYPE:
yyt = & ((* yyt)->INTEGER_TYPE.SIZE); break;
case kREAL_TYPE:
yyt = & ((* yyt)->REAL_TYPE.SIZE); break;
case kBOOLEAN_TYPE:
yyt = & ((* yyt)->BOOLEAN_TYPE.SIZE); break;
case kCOMPLEX_TYPE:
yyt = & ((* yyt)->COMPLEX_TYPE.SIZE); break;
case kSTRING_TYPE:
yyt = & ((* yyt)->STRING_TYPE.LENGTH); break;
case kTYPE_ID:
yyt = & ((* yyt)->TYPE_ID.ID); break;
case kARRAY_TYPE:
gettTree (& ((* yyt)->ARRAY_TYPE.ARRAY_INDEX_TYPES))
yyt = & ((* yyt)->ARRAY_TYPE.ARRAY_COMP_TYPE); break;
case kRECORD_TYPE:
gettIdent ((* yyt)->RECORD_TYPE.Name)
yyt = & ((* yyt)->RECORD_TYPE.COMPONENTS); break;
case kPOINTER_TYPE:
yyt = & ((* yyt)->POINTER_TYPE.PTR_COMP); break;
case kSHAPE_SPEC:
yyt = & ((* yyt)->SHAPE_SPEC.Overlap); break;
case kEXPLICIT_SHAPE:
gettTree (& ((* yyt)->EXPLICIT_SHAPE.Overlap))
gettTree (& ((* yyt)->EXPLICIT_SHAPE.LOWER))
yyt = & ((* yyt)->EXPLICIT_SHAPE.UPPER); break;
case kASSUMED_SHAPE:
gettTree (& ((* yyt)->ASSUMED_SHAPE.Overlap))
yyt = & ((* yyt)->ASSUMED_SHAPE.LOWER); break;
case kDEFERRED_SHAPE:
yyt = & ((* yyt)->DEFERRED_SHAPE.Overlap); break;
case kASSUMED_SIZE:
gettTree (& ((* yyt)->ASSUMED_SIZE.Overlap))
yyt = & ((* yyt)->ASSUMED_SIZE.LOWER); break;
case kSHAPE_LIST:
gettTree (& ((* yyt)->SHAPE_LIST.Elem))
yyt = & ((* yyt)->SHAPE_LIST.Next); break;
case kDECL_NODE:
gettIdent ((* yyt)->DECL_NODE.Ident)
getint ((* yyt)->DECL_NODE.Line)
return;
case kTYPE_DECL:
gettIdent ((* yyt)->TYPE_DECL.Ident)
getint ((* yyt)->TYPE_DECL.Line)
yyt = & ((* yyt)->TYPE_DECL.VAL); break;
case kNAME_DECL:
gettIdent ((* yyt)->NAME_DECL.Ident)
getint ((* yyt)->NAME_DECL.Line)
return;
case kRENAME_DECL:
gettIdent ((* yyt)->RENAME_DECL.Ident)
getint ((* yyt)->RENAME_DECL.Line)
gettIdent ((* yyt)->RENAME_DECL.oldname)
return;
case kENTITY_DECL:
gettIdent ((* yyt)->ENTITY_DECL.Ident)
getint ((* yyt)->ENTITY_DECL.Line)
yyt = & ((* yyt)->ENTITY_DECL.ATTRIBUTES); break;
case kVAR_DECL:
gettIdent ((* yyt)->VAR_DECL.Ident)
getint ((* yyt)->VAR_DECL.Line)
yyt = & ((* yyt)->VAR_DECL.VAL); break;
case kCOMMON_DECL:
gettIdent ((* yyt)->COMMON_DECL.Ident)
getint ((* yyt)->COMMON_DECL.Line)
yyt = & ((* yyt)->COMMON_DECL.IDS); break;
case kNAMELIST_DECL:
gettIdent ((* yyt)->NAMELIST_DECL.Ident)
getint ((* yyt)->NAMELIST_DECL.Line)
yyt = & ((* yyt)->NAMELIST_DECL.IDS); break;
case kTEMPLATE_DECL:
gettIdent ((* yyt)->TEMPLATE_DECL.Ident)
getint ((* yyt)->TEMPLATE_DECL.Line)
yyt = & ((* yyt)->TEMPLATE_DECL.DIMENSIONS); break;
case kRAGGED_DECL:
gettIdent ((* yyt)->RAGGED_DECL.Ident)
getint ((* yyt)->RAGGED_DECL.Line)
yyt = & ((* yyt)->RAGGED_DECL.DIMENSIONS); break;
case kPROCESSORS_DECL:
gettIdent ((* yyt)->PROCESSORS_DECL.Ident)
getint ((* yyt)->PROCESSORS_DECL.Line)
yyt = & ((* yyt)->PROCESSORS_DECL.DIMENSIONS); break;
case kSUB_PROCS_DECL:
gettIdent ((* yyt)->SUB_PROCS_DECL.Ident)
getint ((* yyt)->SUB_PROCS_DECL.Line)
gettTree (& ((* yyt)->SUB_PROCS_DECL.TOP))
yyt = & ((* yyt)->SUB_PROCS_DECL.DIMENSIONS); break;
case kINHERIT_DECL:
gettIdent ((* yyt)->INHERIT_DECL.Ident)
getint ((* yyt)->INHERIT_DECL.Line)
return;
case kRANGE_DECL:
gettIdent ((* yyt)->RANGE_DECL.Ident)
getint ((* yyt)->RANGE_DECL.Line)
yyt = & ((* yyt)->RANGE_DECL.RANGES); break;
case kSTMT_FUNC_DECL:
gettIdent ((* yyt)->STMT_FUNC_DECL.Ident)
getint ((* yyt)->STMT_FUNC_DECL.Line)
gettTree (& ((* yyt)->STMT_FUNC_DECL.FORMALS))
gettTree (& ((* yyt)->STMT_FUNC_DECL.RESULT_TYPE))
yyt = & ((* yyt)->STMT_FUNC_DECL.FFUNC_BODY); break;
case kINTERFACE_DECL:
gettIdent ((* yyt)->INTERFACE_DECL.Ident)
getint ((* yyt)->INTERFACE_DECL.Line)
gettTree (& ((* yyt)->INTERFACE_DECL.SPEC))
yyt = & ((* yyt)->INTERFACE_DECL.ITEMS); break;
case kVAR_PARAM_DECL:
gettIdent ((* yyt)->VAR_PARAM_DECL.Ident)
getint ((* yyt)->VAR_PARAM_DECL.Line)
yyt = & ((* yyt)->VAR_PARAM_DECL.VAL); break;
case kRET_PARAM_DECL:
gettIdent ((* yyt)->RET_PARAM_DECL.Ident)
getint ((* yyt)->RET_PARAM_DECL.Line)
return;
case kIMPLICIT_DECL:
gettIdent ((* yyt)->IMPLICIT_DECL.Ident)
getint ((* yyt)->IMPLICIT_DECL.Line)
gettIdent ((* yyt)->IMPLICIT_DECL.first)
gettIdent ((* yyt)->IMPLICIT_DECL.last)
yyt = & ((* yyt)->IMPLICIT_DECL.TYPE); break;
case kUSE_DECL:
gettIdent ((* yyt)->USE_DECL.Ident)
getint ((* yyt)->USE_DECL.Line)
gettIdent ((* yyt)->USE_DECL.use)
yyt = & ((* yyt)->USE_DECL.RENAMINGS); break;
case kONLY_USE_DECL:
gettIdent ((* yyt)->ONLY_USE_DECL.Ident)
getint ((* yyt)->ONLY_USE_DECL.Line)
gettIdent ((* yyt)->ONLY_USE_DECL.use)
yyt = & ((* yyt)->ONLY_USE_DECL.NAMES); break;
case kEQV_DECL:
gettIdent ((* yyt)->EQV_DECL.Ident)
getint ((* yyt)->EQV_DECL.Line)
yyt = & ((* yyt)->EQV_DECL.VARS); break;
case kDATA_DECL:
gettIdent ((* yyt)->DATA_DECL.Ident)
getint ((* yyt)->DATA_DECL.Line)
gettTree (& ((* yyt)->DATA_DECL.VARS))
yyt = & ((* yyt)->DATA_DECL.VALS); break;
case kFORMAT_DECL:
gettIdent ((* yyt)->FORMAT_DECL.Ident)
getint ((* yyt)->FORMAT_DECL.Line)
getint ((* yyt)->FORMAT_DECL.label)
yyt = & ((* yyt)->FORMAT_DECL.FORMAT_PARAMS); break;
case kDIMENSION_DECL:
gettIdent ((* yyt)->DIMENSION_DECL.Ident)
getint ((* yyt)->DIMENSION_DECL.Line)
yyt = & ((* yyt)->DIMENSION_DECL.INDEXES); break;
case kINIT_DATA_DECL:
gettIdent ((* yyt)->INIT_DATA_DECL.Ident)
getint ((* yyt)->INIT_DATA_DECL.Line)
yyt = & ((* yyt)->INIT_DATA_DECL.VAL); break;
case kTYPESPEC_DECL:
gettIdent ((* yyt)->TYPESPEC_DECL.Ident)
getint ((* yyt)->TYPESPEC_DECL.Line)
yyt = & ((* yyt)->TYPESPEC_DECL.VAL); break;
case kALLOCATABLE_DECL:
gettIdent ((* yyt)->ALLOCATABLE_DECL.Ident)
getint ((* yyt)->ALLOCATABLE_DECL.Line)
return;
case kEXTERNAL_DECL:
gettIdent ((* yyt)->EXTERNAL_DECL.Ident)
getint ((* yyt)->EXTERNAL_DECL.Line)
return;
case kINTENT_DECL:
gettIdent ((* yyt)->INTENT_DECL.Ident)
getint ((* yyt)->INTENT_DECL.Line)
getint ((* yyt)->INTENT_DECL.intent)
return;
case kINTRINSIC_DECL:
gettIdent ((* yyt)->INTRINSIC_DECL.Ident)
getint ((* yyt)->INTRINSIC_DECL.Line)
gettIdent ((* yyt)->INTRINSIC_DECL.generic)
getint ((* yyt)->INTRINSIC_DECL.class)
getint ((* yyt)->INTRINSIC_DECL.kind)
return;
case kOPTIONAL_DECL:
gettIdent ((* yyt)->OPTIONAL_DECL.Ident)
getint ((* yyt)->OPTIONAL_DECL.Line)
return;
case kPOINTER_DECL:
gettIdent ((* yyt)->POINTER_DECL.Ident)
getint ((* yyt)->POINTER_DECL.Line)
return;
case kSAVE_DECL:
gettIdent ((* yyt)->SAVE_DECL.Ident)
getint ((* yyt)->SAVE_DECL.Line)
return;
case kTARGET_DECL:
gettIdent ((* yyt)->TARGET_DECL.Ident)
getint ((* yyt)->TARGET_DECL.Line)
return;
case kPARAMETER_DECL:
gettIdent ((* yyt)->PARAMETER_DECL.Ident)
getint ((* yyt)->PARAMETER_DECL.Line)
yyt = & ((* yyt)->PARAMETER_DECL.VAL); break;
case kPUBLIC_DECL:
gettIdent ((* yyt)->PUBLIC_DECL.Ident)
getint ((* yyt)->PUBLIC_DECL.Line)
return;
case kPRIVATE_DECL:
gettIdent ((* yyt)->PRIVATE_DECL.Ident)
getint ((* yyt)->PRIVATE_DECL.Line)
return;
case kSEQUENCE_DECL:
gettIdent ((* yyt)->SEQUENCE_DECL.Ident)
getint ((* yyt)->SEQUENCE_DECL.Line)
return;
case kDISTRIBUTE_DECL:
gettIdent ((* yyt)->DISTRIBUTE_DECL.Ident)
getint ((* yyt)->DISTRIBUTE_DECL.Line)
gettTree (& ((* yyt)->DISTRIBUTE_DECL.DISTRIBUTION))
yyt = & ((* yyt)->DISTRIBUTE_DECL.TARGET); break;
case kALIGN_DECL:
gettIdent ((* yyt)->ALIGN_DECL.Ident)
getint ((* yyt)->ALIGN_DECL.Line)
gettTree (& ((* yyt)->ALIGN_DECL.ALIGN_SOURCE))
yyt = & ((* yyt)->ALIGN_DECL.ALIGN_TARGET); break;
case kMAP_DECL:
gettIdent ((* yyt)->MAP_DECL.Ident)
getint ((* yyt)->MAP_DECL.Line)
yyt = & ((* yyt)->MAP_DECL.MAP_TARGET); break;
case kOVERLAP_DECL:
gettIdent ((* yyt)->OVERLAP_DECL.Ident)
getint ((* yyt)->OVERLAP_DECL.Line)
yyt = & ((* yyt)->OVERLAP_DECL.OVERLAPS); break;
case kSHARED_DECL:
gettIdent ((* yyt)->SHARED_DECL.Ident)
getint ((* yyt)->SHARED_DECL.Line)
getint ((* yyt)->SHARED_DECL.kind)
return;
case kDYNAMIC_DECL:
gettIdent ((* yyt)->DYNAMIC_DECL.Ident)
getint ((* yyt)->DYNAMIC_DECL.Line)
return;
case kTRACE_DECL:
gettIdent ((* yyt)->TRACE_DECL.Ident)
getint ((* yyt)->TRACE_DECL.Line)
return;
case kTREE_DECL:
gettIdent ((* yyt)->TREE_DECL.Ident)
getint ((* yyt)->TREE_DECL.Line)
return;
case kSELECT_DECL:
gettIdent ((* yyt)->SELECT_DECL.Ident)
getint ((* yyt)->SELECT_DECL.Line)
yyt = & ((* yyt)->SELECT_DECL.SELECTIONS); break;
case kHPF_SEQUENCE_DECL:
gettIdent ((* yyt)->HPF_SEQUENCE_DECL.Ident)
getint ((* yyt)->HPF_SEQUENCE_DECL.Line)
return;
case kNOSEQUENCE_DECL:
gettIdent ((* yyt)->NOSEQUENCE_DECL.Ident)
getint ((* yyt)->NOSEQUENCE_DECL.Line)
return;
case kLAYOUT_DECL:
gettIdent ((* yyt)->LAYOUT_DECL.Ident)
getint ((* yyt)->LAYOUT_DECL.Line)
gettStringRef ((* yyt)->LAYOUT_DECL.kind)
return;
case kPASS_BY_DECL:
gettIdent ((* yyt)->PASS_BY_DECL.Ident)
getint ((* yyt)->PASS_BY_DECL.Line)
gettStringRef ((* yyt)->PASS_BY_DECL.kind)
return;
case kMAP_TO_DECL:
gettIdent ((* yyt)->MAP_TO_DECL.Ident)
getint ((* yyt)->MAP_TO_DECL.Line)
gettStringRef ((* yyt)->MAP_TO_DECL.kind)
return;
case kDECL_LIST:
gettTree (& ((* yyt)->DECL_LIST.Elem))
yyt = & ((* yyt)->DECL_LIST.Next); break;
case kRANGE_LIST:
gettTree (& ((* yyt)->RANGE_LIST.Elem))
yyt = & ((* yyt)->RANGE_LIST.Next); break;
case kSELECT_LIST:
gettTree (& ((* yyt)->SELECT_LIST.Elem))
yyt = & ((* yyt)->SELECT_LIST.Next); break;
case kSELECT_SPEC:
getint ((* yyt)->SELECT_SPEC.selector)
return;
case kDISTRIBUTION_SPEC:
getint ((* yyt)->DISTRIBUTION_SPEC.is_descriptive)
return;
case kANY_DISTRIBUTION:
getint ((* yyt)->ANY_DISTRIBUTION.is_descriptive)
return;
case kDEFAULT_DISTRIBUTION:
getint ((* yyt)->DEFAULT_DISTRIBUTION.is_descriptive)
return;
case kREPL_DISTRIBUTION:
getint ((* yyt)->REPL_DISTRIBUTION.is_descriptive)
return;
case kNODE_DISTRIBUTION:
getint ((* yyt)->NODE_DISTRIBUTION.is_descriptive)
yyt = & ((* yyt)->NODE_DISTRIBUTION.MAPPING); break;
case kDIST_LIST:
gettTree (& ((* yyt)->DIST_LIST.Elem))
yyt = & ((* yyt)->DIST_LIST.Next); break;
case kBLOCK_DISTRIBUTION:
yyt = & ((* yyt)->BLOCK_DISTRIBUTION.SIZE); break;
case kCYCLIC_DISTRIBUTION:
yyt = & ((* yyt)->CYCLIC_DISTRIBUTION.SIZE); break;
case kGEN_BLOCK_DISTRIBUTION:
yyt = & ((* yyt)->GEN_BLOCK_DISTRIBUTION.ARRAY); break;
case kINDIRECT_DISTRIBUTION:
yyt = & ((* yyt)->INDIRECT_DISTRIBUTION.ARRAY); break;
case kARBITRARY_DISTRIBUTION:
gettTree (& ((* yyt)->ARBITRARY_DISTRIBUTION.SIZE))
gettTree (& ((* yyt)->ARBITRARY_DISTRIBUTION.LENGTH))
yyt = & ((* yyt)->ARBITRARY_DISTRIBUTION.MAP); break;
case kPROCESSOR_ARRAY:
yyt = & ((* yyt)->PROCESSOR_ARRAY.TOPNAME); break;
case kPROCESSOR_SUBSET:
gettTree (& ((* yyt)->PROCESSOR_SUBSET.SUBTOP))
gettTree (& ((* yyt)->PROCESSOR_SUBSET.FULLTOP))
yyt = & ((* yyt)->PROCESSOR_SUBSET.SUBSCRIPTS); break;
case kALIGN_SPEC:
getint ((* yyt)->ALIGN_SPEC.is_descriptive)
gettTree (& ((* yyt)->ALIGN_SPEC.TEMPNAME))
yyt = & ((* yyt)->ALIGN_SPEC.SUBSCRIPTS); break;
case kID_GENERIC_SPEC:
gettIdent ((* yyt)->ID_GENERIC_SPEC.name)
return;
case kOP_GENERIC_SPEC:
yyt = & ((* yyt)->OP_GENERIC_SPEC.OPERATOR); break;
case kACF_LIST:
gettTree (& ((* yyt)->ACF_LIST.Elem))
yyt = & ((* yyt)->ACF_LIST.Next); break;
case kACF_NODE:
getint ((* yyt)->ACF_NODE.Label)
getint ((* yyt)->ACF_NODE.Line)
return;
case kACF_DUMMY:
getint ((* yyt)->ACF_DUMMY.Label)
getint ((* yyt)->ACF_DUMMY.Line)
return;
case kACF_BASIC:
getint ((* yyt)->ACF_BASIC.Label)
getint ((* yyt)->ACF_BASIC.Line)
yyt = & ((* yyt)->ACF_BASIC.BASIC_STMT); break;
case kACF_IF:
getint ((* yyt)->ACF_IF.Label)
getint ((* yyt)->ACF_IF.Line)
gettTree (& ((* yyt)->ACF_IF.IF_EXP))
gettTree (& ((* yyt)->ACF_IF.THEN_PART))
yyt = & ((* yyt)->ACF_IF.ELSE_PART); break;
case kACF_WHERE:
getint ((* yyt)->ACF_WHERE.Label)
getint ((* yyt)->ACF_WHERE.Line)
gettTree (& ((* yyt)->ACF_WHERE.WHERE_EXP))
gettTree (& ((* yyt)->ACF_WHERE.TRUE_PART))
yyt = & ((* yyt)->ACF_WHERE.FALSE_PART); break;
case kACF_CASE:
getint ((* yyt)->ACF_CASE.Label)
getint ((* yyt)->ACF_CASE.Line)
gettTree (& ((* yyt)->ACF_CASE.CASE_EXP))
yyt = & ((* yyt)->ACF_CASE.CASE_ALTS); break;
case kACF_WHILE:
getint ((* yyt)->ACF_WHILE.Label)
getint ((* yyt)->ACF_WHILE.Line)
gettTree (& ((* yyt)->ACF_WHILE.WHILE_EXP))
yyt = & ((* yyt)->ACF_WHILE.WHILE_BODY); break;
case kACF_REPEAT:
getint ((* yyt)->ACF_REPEAT.Label)
getint ((* yyt)->ACF_REPEAT.Line)
gettTree (& ((* yyt)->ACF_REPEAT.REPEAT_BODY))
yyt = & ((* yyt)->ACF_REPEAT.REPEAT_EXP); break;
case kACF_LOOP:
getint ((* yyt)->ACF_LOOP.Label)
getint ((* yyt)->ACF_LOOP.Line)
yyt = & ((* yyt)->ACF_LOOP.LOOP_BODY); break;
case kACF_DO:
getint ((* yyt)->ACF_DO.Label)
getint ((* yyt)->ACF_DO.Line)
gettTree (& ((* yyt)->ACF_DO.DO_ID))
gettTree (& ((* yyt)->ACF_DO.DO_RANGE))
gettTree (& ((* yyt)->ACF_DO.DO_BODY))
gettTree (& ((* yyt)->ACF_DO.DO_DEP_INFO))
yyt = & ((* yyt)->ACF_DO.DO_HOME_INFO); break;
case kACF_FORALL:
getint ((* yyt)->ACF_FORALL.Label)
getint ((* yyt)->ACF_FORALL.Line)
gettTree (& ((* yyt)->ACF_FORALL.FORALL_ID))
gettTree (& ((* yyt)->ACF_FORALL.FORALL_RANGE))
gettTree (& ((* yyt)->ACF_FORALL.FORALL_BODY))
gettTree (& ((* yyt)->ACF_FORALL.FORALL_DEP_INFO))
yyt = & ((* yyt)->ACF_FORALL.FORALL_HOME_INFO); break;
case kACF_FLOW_GRAPH:
getint ((* yyt)->ACF_FLOW_GRAPH.Label)
getint ((* yyt)->ACF_FLOW_GRAPH.Line)
yyt = & ((* yyt)->ACF_FLOW_GRAPH.FLOW_NODES); break;
case kACF_ENTRY:
getint ((* yyt)->ACF_ENTRY.Label)
getint ((* yyt)->ACF_ENTRY.Line)
yyt = & ((* yyt)->ACF_ENTRY.ENTRY_DECL); break;
case kACF_HOME:
getint ((* yyt)->ACF_HOME.Label)
getint ((* yyt)->ACF_HOME.Line)
gettTree (& ((* yyt)->ACF_HOME.HOME_VAR))
getbool ((* yyt)->ACF_HOME.user_home)
yyt = & ((* yyt)->ACF_HOME.HOME_BODY); break;
case kACF_RESIDENT:
getint ((* yyt)->ACF_RESIDENT.Label)
getint ((* yyt)->ACF_RESIDENT.Line)
gettTree (& ((* yyt)->ACF_RESIDENT.RESIDENT_VAR))
yyt = & ((* yyt)->ACF_RESIDENT.RESIDENT_BODY); break;
case kACF_NEW:
getint ((* yyt)->ACF_NEW.Label)
getint ((* yyt)->ACF_NEW.Line)
gettTree (& ((* yyt)->ACF_NEW.NEW_VAR))
yyt = & ((* yyt)->ACF_NEW.NEW_BODY); break;
case kACF_REDUCTION:
getint ((* yyt)->ACF_REDUCTION.Label)
getint ((* yyt)->ACF_REDUCTION.Line)
getbool ((* yyt)->ACF_REDUCTION.is_zero)
gettTree (& ((* yyt)->ACF_REDUCTION.REDUCTION_VAR))
gettTree (& ((* yyt)->ACF_REDUCTION.REDUCTION_LVAR))
gettTree (& ((* yyt)->ACF_REDUCTION.REDUCTION_FUNC))
yyt = & ((* yyt)->ACF_REDUCTION.REDUCTION_BODY); break;
case kACF_TASK_REGION:
getint ((* yyt)->ACF_TASK_REGION.Label)
getint ((* yyt)->ACF_TASK_REGION.Line)
getint ((* yyt)->ACF_TASK_REGION.task_flag)
yyt = & ((* yyt)->ACF_TASK_REGION.TASK_BODY); break;
case kACF_PARALLEL:
getint ((* yyt)->ACF_PARALLEL.Label)
getint ((* yyt)->ACF_PARALLEL.Line)
gettTree (& ((* yyt)->ACF_PARALLEL.REDUCTION_VARS))
gettTree (& ((* yyt)->ACF_PARALLEL.SHARED_VARS))
gettTree (& ((* yyt)->ACF_PARALLEL.LOCAL_VARS))
yyt = & ((* yyt)->ACF_PARALLEL.PARALLEL_BODY); break;
case kACF_CRITICAL:
getint ((* yyt)->ACF_CRITICAL.Label)
getint ((* yyt)->ACF_CRITICAL.Line)
yyt = & ((* yyt)->ACF_CRITICAL.CRITICAL_BODY); break;
case kACF_ON:
getint ((* yyt)->ACF_ON.Label)
getint ((* yyt)->ACF_ON.Line)
gettTree (& ((* yyt)->ACF_ON.ON_HOME))
gettTree (& ((* yyt)->ACF_ON.ON_SPECS))
yyt = & ((* yyt)->ACF_ON.ON_BODY); break;
case kACF_RM_READ:
getint ((* yyt)->ACF_RM_READ.Label)
getint ((* yyt)->ACF_RM_READ.Line)
gettTree (& ((* yyt)->ACF_RM_READ.LOCAL_VAR))
yyt = & ((* yyt)->ACF_RM_READ.REMOTE_VAR); break;
case kACF_RM_WRITE:
getint ((* yyt)->ACF_RM_WRITE.Label)
getint ((* yyt)->ACF_RM_WRITE.Line)
getint ((* yyt)->ACF_RM_WRITE.op)
gettTree (& ((* yyt)->ACF_RM_WRITE.REMOTE_VAR))
yyt = & ((* yyt)->ACF_RM_WRITE.LOCAL_DATA); break;
case kACF_MOVE:
getint ((* yyt)->ACF_MOVE.Label)
getint ((* yyt)->ACF_MOVE.Line)
yyt = & ((* yyt)->ACF_MOVE.MOVE_STMT); break;
case kACF_BARRIER:
getint ((* yyt)->ACF_BARRIER.Label)
getint ((* yyt)->ACF_BARRIER.Line)
yyt = & ((* yyt)->ACF_BARRIER.BARRIER_VARS); break;
case kACF_UPDATE:
getint ((* yyt)->ACF_UPDATE.Label)
getint ((* yyt)->ACF_UPDATE.Line)
yyt = & ((* yyt)->ACF_UPDATE.OVERLAP_VAR); break;
case kON_VAR_CLAUSE:
yyt = & ((* yyt)->ON_VAR_CLAUSE.ON_VAR); break;
case kON_PROC_CLAUSE:
yyt = & ((* yyt)->ON_PROC_CLAUSE.ON_PROC); break;
case kON_PROCS:
gettObject ((* yyt)->ON_PROCS.topid)
return;
case kON_VAR:
gettObject ((* yyt)->ON_VAR.topid)
gettObject ((* yyt)->ON_VAR.tempid)
gettObject ((* yyt)->ON_VAR.arrayid)
return;
case kON_LIST:
gettTree (& ((* yyt)->ON_LIST.Elem))
yyt = & ((* yyt)->ON_LIST.Next); break;
case kON_SPEC:
getint ((* yyt)->ON_SPEC.top_dim)
getint ((* yyt)->ON_SPEC.temp_dim)
getint ((* yyt)->ON_SPEC.array_dim)
yyt = & ((* yyt)->ON_SPEC.ON_VAL); break;
case kSELECTED_ACF_LIST:
gettTree (& ((* yyt)->SELECTED_ACF_LIST.Elem))
yyt = & ((* yyt)->SELECTED_ACF_LIST.Next); break;
case kSELECTED_ACF_NODE:
gettTree (& ((* yyt)->SELECTED_ACF_NODE.SELECT_LIST))
yyt = & ((* yyt)->SELECTED_ACF_NODE.SELECT_ACFS); break;
case kASSIGN_STMT:
gettTree (& ((* yyt)->ASSIGN_STMT.ASSIGN_VAR))
yyt = & ((* yyt)->ASSIGN_STMT.ASSIGN_EXP); break;
case kPTR_ASSIGN_STMT:
gettTree (& ((* yyt)->PTR_ASSIGN_STMT.ASSIGN_VAR))
yyt = & ((* yyt)->PTR_ASSIGN_STMT.ASSIGN_EXP); break;
case kLABEL_ASSIGN_STMT:
getint ((* yyt)->LABEL_ASSIGN_STMT.assign_label)
yyt = & ((* yyt)->LABEL_ASSIGN_STMT.LABEL_VAR); break;
case kCALL_STMT:
gettTree (& ((* yyt)->CALL_STMT.CALL_ID))
yyt = & ((* yyt)->CALL_STMT.CALL_PARAMS); break;
case kIO_STMT:
gettTree (& ((* yyt)->IO_STMT.ID))
gettTree (& ((* yyt)->IO_STMT.IO_SPECS))
yyt = & ((* yyt)->IO_STMT.IO_ITEMS); break;
case kGOTO_STMT:
getint ((* yyt)->GOTO_STMT.GOTO_LABEL)
return;
case kASS_GOTO_STMT:
gettTree (& ((* yyt)->ASS_GOTO_STMT.GOTO_VAR))
yyt = & ((* yyt)->ASS_GOTO_STMT.LABELS); break;
case kCOMP_GOTO_STMT:
gettTree (& ((* yyt)->COMP_GOTO_STMT.GOTO_LABELS))
yyt = & ((* yyt)->COMP_GOTO_STMT.GOTO_EXP); break;
case kCOMP_IF_STMT:
getint ((* yyt)->COMP_IF_STMT.IF_LT_LABEL)
getint ((* yyt)->COMP_IF_STMT.IF_EQ_LABEL)
getint ((* yyt)->COMP_IF_STMT.IF_GT_LABEL)
yyt = & ((* yyt)->COMP_IF_STMT.IF_EXP); break;
case kRETURN_STMT:
yyt = & ((* yyt)->RETURN_STMT.RETURN_EXP); break;
case kPAUSE_STMT:
yyt = & ((* yyt)->PAUSE_STMT.PAUSE_CONST); break;
case kEXIT_STMT:
gettIdent ((* yyt)->EXIT_STMT.loopid)
return;
case kCYCLE_STMT:
gettIdent ((* yyt)->CYCLE_STMT.loopid)
return;
case kSTOP_STMT:
yyt = & ((* yyt)->STOP_STMT.STOP_CONST); break;
case kFORMAT_STMT:
yyt = & ((* yyt)->FORMAT_STMT.FORMAT_PARAMS); break;
case kINHERIT_DSP_STMT:
gettTree (& ((* yyt)->INHERIT_DSP_STMT.VAR))
getint ((* yyt)->INHERIT_DSP_STMT.flag)
getint ((* yyt)->INHERIT_DSP_STMT.safety)
yyt = & ((* yyt)->INHERIT_DSP_STMT.SHAPE); break;
case kCREATE_DSP_STMT:
gettTree (& ((* yyt)->CREATE_DSP_STMT.VAR))
getint ((* yyt)->CREATE_DSP_STMT.flag)
yyt = & ((* yyt)->CREATE_DSP_STMT.SHAPE); break;
case kFREE_DSP_STMT:
getint ((* yyt)->FREE_DSP_STMT.flag)
yyt = & ((* yyt)->FREE_DSP_STMT.VAR); break;
case kALLOCATE_STMT:
gettTree (& ((* yyt)->ALLOCATE_STMT.PARAMS))
yyt = & ((* yyt)->ALLOCATE_STMT.STATUS); break;
case kDEALLOCATE_STMT:
gettTree (& ((* yyt)->DEALLOCATE_STMT.PARAMS))
yyt = & ((* yyt)->DEALLOCATE_STMT.STATUS); break;
case kNULLIFY_STMT:
yyt = & ((* yyt)->NULLIFY_STMT.PARAMS); break;
case kREDUCE_STMT:
gettTree (& ((* yyt)->REDUCE_STMT.RED_FUNC))
yyt = & ((* yyt)->REDUCE_STMT.RED_PARAMS); break;
case kMOVE_STMT:
gettTree (& ((* yyt)->MOVE_STMT.TARGET))
gettTree (& ((* yyt)->MOVE_STMT.SOURCE))
yyt = & ((* yyt)->MOVE_STMT.MASK); break;
case kSCATTER_STMT:
gettIdent ((* yyt)->SCATTER_STMT.func)
gettTree (& ((* yyt)->SCATTER_STMT.ARRAY))
gettTree (& ((* yyt)->SCATTER_STMT.BASE))
gettTree (& ((* yyt)->SCATTER_STMT.INDEXES))
yyt = & ((* yyt)->SCATTER_STMT.MASK); break;
case kREALIGN_STMT:
gettTree (& ((* yyt)->REALIGN_STMT.ALIGNEE))
gettTree (& ((* yyt)->REALIGN_STMT.ALIGN_SOURCE))
gettDefinitions ((* yyt)->REALIGN_STMT.distribution)
yyt = & ((* yyt)->REALIGN_STMT.ALIGN_TARGET); break;
case kREDISTRIBUTE_STMT:
gettTree (& ((* yyt)->REDISTRIBUTE_STMT.DISTRIBUTEE))
gettTree (& ((* yyt)->REDISTRIBUTE_STMT.DISTRIBUTION))
gettDefinitions ((* yyt)->REDISTRIBUTE_STMT.distribution)
yyt = & ((* yyt)->REDISTRIBUTE_STMT.TARGET); break;
case kREG_SHADOW_GET:
gettTree (& ((* yyt)->REG_SHADOW_GET.SHADOW_VAR))
yyt = & ((* yyt)->REG_SHADOW_GET.SHADOW_SIZE); break;
case kREG_SHADOW_PUT:
getint ((* yyt)->REG_SHADOW_PUT.op)
gettTree (& ((* yyt)->REG_SHADOW_PUT.SHADOW_VAR))
yyt = & ((* yyt)->REG_SHADOW_PUT.SHADOW_SIZE); break;
case kREG_SHADOW_SET:
gettTree (& ((* yyt)->REG_SHADOW_SET.SHADOW_VAR))
gettTree (& ((* yyt)->REG_SHADOW_SET.SHADOW_SIZE))
yyt = & ((* yyt)->REG_SHADOW_SET.VAL); break;
case kIND_SHADOW_CREATE:
getint ((* yyt)->IND_SHADOW_CREATE.id)
gettTree (& ((* yyt)->IND_SHADOW_CREATE.SHADOW_VAR))
gettTree (& ((* yyt)->IND_SHADOW_CREATE.INDEX_VAR))
gettTree (& ((* yyt)->IND_SHADOW_CREATE.TMP_INDEX_VAR))
yyt = & ((* yyt)->IND_SHADOW_CREATE.MASK); break;
case kIND_SHADOW_FREE:
getint ((* yyt)->IND_SHADOW_FREE.id)
yyt = & ((* yyt)->IND_SHADOW_FREE.INDEX_VAR); break;
case kIND_SHADOW_GET:
getint ((* yyt)->IND_SHADOW_GET.id)
yyt = & ((* yyt)->IND_SHADOW_GET.SHADOW_VAR); break;
case kIND_SHADOW_PUT:
getint ((* yyt)->IND_SHADOW_PUT.id)
getint ((* yyt)->IND_SHADOW_PUT.op)
yyt = & ((* yyt)->IND_SHADOW_PUT.SHADOW_VAR); break;
case kIND_SHADOW_SET:
getint ((* yyt)->IND_SHADOW_SET.id)
gettTree (& ((* yyt)->IND_SHADOW_SET.SHADOW_VAR))
yyt = & ((* yyt)->IND_SHADOW_SET.VAL); break;
case kBROADCAST_STMT:
gettTree (& ((* yyt)->BROADCAST_STMT.ELEMS))
gettTree (& ((* yyt)->BROADCAST_STMT.OWNER_TOP))
yyt = & ((* yyt)->BROADCAST_STMT.OWNER_SPEC); break;
case kREDUCTION_STMT:
gettTree (& ((* yyt)->REDUCTION_STMT.REDFUNC))
gettTree (& ((* yyt)->REDUCTION_STMT.ELEMS))
gettTree (& ((* yyt)->REDUCTION_STMT.RED_TOP))
yyt = & ((* yyt)->REDUCTION_STMT.RED_SPEC); break;
case kDIRTY_STMT:
yyt = & ((* yyt)->DIRTY_STMT.ELEMS); break;
case kLABEL_LIST:
getint ((* yyt)->LABEL_LIST.Elem)
yyt = & ((* yyt)->LABEL_LIST.Next); break;
case kOP_DEFINED:
gettIdent ((* yyt)->OP_DEFINED.opname)
return;
case kCONST_EXP:
yyt = & ((* yyt)->CONST_EXP.C); break;
case kOP_EXP:
gettTree (& ((* yyt)->OP_EXP.EXP_OP))
gettTree (& ((* yyt)->OP_EXP.OPND1))
yyt = & ((* yyt)->OP_EXP.OPND2); break;
case kOP1_EXP:
gettTree (& ((* yyt)->OP1_EXP.EXP_OP1))
yyt = & ((* yyt)->OP1_EXP.OPND); break;
case kSLICE_EXP:
gettTree (& ((* yyt)->SLICE_EXP.START))
gettTree (& ((* yyt)->SLICE_EXP.STOP))
yyt = & ((* yyt)->SLICE_EXP.INC); break;
case kVAR_EXP:
yyt = & ((* yyt)->VAR_EXP.V); break;
case kFUNC_CALL_EXP:
gettTree (& ((* yyt)->FUNC_CALL_EXP.FUNC_ID))
yyt = & ((* yyt)->FUNC_CALL_EXP.FUNC_PARAMS); break;
case kNAMED_EXP:
gettIdent ((* yyt)->NAMED_EXP.Name)
yyt = & ((* yyt)->NAMED_EXP.VAL); break;
case kDO_EXP:
gettTree (& ((* yyt)->DO_EXP.DO_ID))
gettTree (& ((* yyt)->DO_EXP.RANGE))
yyt = & ((* yyt)->DO_EXP.BODY); break;
case kTYPE_EXP:
gettTree (& ((* yyt)->TYPE_EXP.ID))
yyt = & ((* yyt)->TYPE_EXP.ELEMENTS); break;
case kARRAY_EXP:
yyt = & ((* yyt)->ARRAY_EXP.ELEMENTS); break;
case kPERM_EXP:
getPermutation ((* yyt)->PERM_EXP.perm)
yyt = & ((* yyt)->PERM_EXP.VAL); break;
case kBOUND_EXP:
getint ((* yyt)->BOUND_EXP.dim)
getint ((* yyt)->BOUND_EXP.kind)
getint ((* yyt)->BOUND_EXP.local)
yyt = & ((* yyt)->BOUND_EXP.VAR); break;
case kRANK_EXP:
getint ((* yyt)->RANK_EXP.kind)
return;
case kBTE_LIST:
gettTree (& ((* yyt)->BTE_LIST.Elem))
yyt = & ((* yyt)->BTE_LIST.Next); break;
case kUSED_VAR:
yyt = & ((* yyt)->USED_VAR.VARNAME); break;
case kINDEXED_VAR:
gettTree (& ((* yyt)->INDEXED_VAR.IND_VAR))
yyt = & ((* yyt)->INDEXED_VAR.IND_EXPS); break;
case kSUBSTRING_VAR:
gettTree (& ((* yyt)->SUBSTRING_VAR.IND_VAR))
yyt = & ((* yyt)->SUBSTRING_VAR.IND_EXP); break;
case kSELECTED_VAR:
gettTree (& ((* yyt)->SELECTED_VAR.SELEC_VAR))
yyt = & ((* yyt)->SELECTED_VAR.SELECTOR); break;
case kLOOP_VAR:
yyt = & ((* yyt)->LOOP_VAR.LOOP_VARNAME); break;
case kPERM_VAR:
getPermutation ((* yyt)->PERM_VAR.perm)
yyt = & ((* yyt)->PERM_VAR.VAL); break;
case kADDR:
yyt = & ((* yyt)->ADDR.E); break;
case kDO_VAR:
gettTree (& ((* yyt)->DO_VAR.DO_ID))
gettTree (& ((* yyt)->DO_VAR.RANGE))
yyt = & ((* yyt)->DO_VAR.BODY); break;
case kREMOTE_VAR:
yyt = & ((* yyt)->REMOTE_VAR.VAR); break;
case kBTV_LIST:
gettTree (& ((* yyt)->BTV_LIST.Elem))
yyt = & ((* yyt)->BTV_LIST.Next); break;
case kVAR_PARAM:
gettDefinitions ((* yyt)->VAR_PARAM.formal)
getint ((* yyt)->VAR_PARAM.intent)
getint ((* yyt)->VAR_PARAM.layout)
getint ((* yyt)->VAR_PARAM.pass_by)
getint ((* yyt)->VAR_PARAM.trace)
yyt = & ((* yyt)->VAR_PARAM.V); break;
case kVALUE_PARAM:
yyt = & ((* yyt)->VALUE_PARAM.E); break;
case kNAMED_PARAM:
gettIdent ((* yyt)->NAMED_PARAM.Name)
yyt = & ((* yyt)->NAMED_PARAM.VAL); break;
case kPROC_PARAM:
yyt = & ((* yyt)->PROC_PARAM.P); break;
case kFUNC_PARAM:
yyt = & ((* yyt)->FUNC_PARAM.F); break;
case kFORMAT_PARAM:
gettStringRef ((* yyt)->FORMAT_PARAM.S)
return;
case kRETURN_PARAM:
getint ((* yyt)->RETURN_PARAM.label)
return;
case kNO_PARAM:
getint ((* yyt)->NO_PARAM.type)
return;
case kBTP_LIST:
gettTree (& ((* yyt)->BTP_LIST.Elem))
yyt = & ((* yyt)->BTP_LIST.Next); break;
case kBOOL_CONSTANT:
getint ((* yyt)->BOOL_CONSTANT.value)
return;
case kINT_CONSTANT:
getint ((* yyt)->INT_CONSTANT.value)
return;
case kREAL_CONSTANT:
gettStringRef ((* yyt)->REAL_CONSTANT.value)
getint ((* yyt)->REAL_CONSTANT.kind)
return;
case kSTRING_CONSTANT:
gettStringRef ((* yyt)->STRING_CONSTANT.value)
return;
case kCOMPLEX_CONSTANT:
gettStringRef ((* yyt)->COMPLEX_CONSTANT.rvalue)
gettStringRef ((* yyt)->COMPLEX_CONSTANT.ivalue)
return;
case kPROC_OBJ:
gettIdent ((* yyt)->PROC_OBJ.Ident)
gettDefinitions ((* yyt)->PROC_OBJ.Object)
return;
case kVAR_OBJ:
getint ((* yyt)->VAR_OBJ.Pos)
gettIdent ((* yyt)->VAR_OBJ.Ident)
gettDefinitions ((* yyt)->VAR_OBJ.Object)
gettReaching ((* yyt)->VAR_OBJ.Reaching)
return;
case kTYPE_OBJ:
gettIdent ((* yyt)->TYPE_OBJ.Ident)
gettDefinitions ((* yyt)->TYPE_OBJ.Object)
return;
case kREC_COMP:
gettIdent ((* yyt)->REC_COMP.Ident)
gettDefinitions ((* yyt)->REC_COMP.Object)
return;
case kTOP_OBJ:
gettIdent ((* yyt)->TOP_OBJ.Ident)
gettDefinitions ((* yyt)->TOP_OBJ.Object)
return;
case kTEMP_OBJ:
gettIdent ((* yyt)->TEMP_OBJ.Ident)
gettDefinitions ((* yyt)->TEMP_OBJ.Object)
return;
case kOVERLAP_SPEC:
getint ((* yyt)->OVERLAP_SPEC.left_size)
getint ((* yyt)->OVERLAP_SPEC.right_size)
getint ((* yyt)->OVERLAP_SPEC.left_update)
getint ((* yyt)->OVERLAP_SPEC.right_update)
return;
case kINDEP_INFO:
getbool ((* yyt)->INDEP_INFO.user_independent)
getint ((* yyt)->INDEP_INFO.selection)
getInfoTree ((* yyt)->INDEP_INFO.red_vars)
getInfoTree ((* yyt)->INDEP_INFO.new_vars)
return;
case kVECTOR_INFO:
getbool ((* yyt)->VECTOR_INFO.is_independent)
getint ((* yyt)->VECTOR_INFO.selection)
getint ((* yyt)->VECTOR_INFO.loop_cnt)
getInfoTree ((* yyt)->VECTOR_INFO.independent_vars)
return;
case kCOMM_INFO:
getInfoTree ((* yyt)->COMM_INFO.home_var)
getbool ((* yyt)->COMM_INFO.is_local)
return;
case kON_INFO:
gettObject ((* yyt)->ON_INFO.on_id)
getint ((* yyt)->ON_INFO.on_dim)
getint ((* yyt)->ON_INFO.base)
getint ((* yyt)->ON_INFO.stride)
return;
case kPARALLEL_INFO:
getInfoTree ((* yyt)->PARALLEL_INFO.reduction_vars)
getInfoTree ((* yyt)->PARALLEL_INFO.home_var)
getInfoTree ((* yyt)->PARALLEL_INFO.new_vars)
getInfoTree ((* yyt)->PARALLEL_INFO.resident_vars)
return;
case kREACHING_INFO:
getInfoTree ((* yyt)->REACHING_INFO.var_allocate)
getInfoTree ((* yyt)->REACHING_INFO.temp_allocate)
gettDefinitions ((* yyt)->REACHING_INFO.var_distribution)
gettDefinitions ((* yyt)->REACHING_INFO.temp_distribution)
return;
  default: return;
  }
 }
}

tTree GetTree
# if defined __STDC__ | defined __cplusplus
 (FILE * yyyf)
# else
 (yyyf) FILE * yyyf;
# endif
{
 tTree yyt;
 yyf = yyyf;
 if (yyRecursionLevel ++ == 0) yyLabelCount = 0;
 yyGetTree (& yyt);
 yyRecursionLevel --;
 return yyt;
}

tTree ReverseTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyOld)
# else
 (yyOld) tTree yyOld;
# endif
{
 register tTree yyNew, yyNext, yyTail;
 yyNew = yyOld;
 yyTail = yyOld;
 for (;;) {
  switch (yyOld->Kind) {
  default: goto yyExit;
  }
  yyNew = yyOld;
  yyOld = yyNext;
 }
yyExit:
 switch (yyTail->Kind) {
 default: ;
 }
 return yyNew;
}

# define yyInitOldToNewStoreSize 32

typedef struct { tTree yyOld, yyNew; } yytOldToNew;
static unsigned long yyOldToNewStoreSize = yyInitOldToNewStoreSize;
static yytOldToNew yyOldToNewStore [yyInitOldToNewStoreSize];
static yytOldToNew * yyOldToNewStorePtr = yyOldToNewStore;
static int yyOldToNewCount;

static void yyStoreOldToNew
# if defined __STDC__ | defined __cplusplus
 (tTree yyOld, tTree yyNew)
# else
 (yyOld, yyNew) tTree yyOld, yyNew;
# endif
{
 if (++ yyOldToNewCount == yyOldToNewStoreSize)
  ExtendArray ((char * *) & yyOldToNewStorePtr, & yyOldToNewStoreSize, sizeof (yytOldToNew));
 yyOldToNewStorePtr [yyOldToNewCount].yyOld = yyOld;
 yyOldToNewStorePtr [yyOldToNewCount].yyNew = yyNew;
}

static tTree yyMapOldToNew
# if defined __STDC__ | defined __cplusplus
 (tTree yyOld)
# else
 (yyOld) tTree yyOld;
# endif
{
 register int yyi;
 for (yyi = 1; yyi <= yyOldToNewCount; yyi ++)
  if (yyOldToNewStorePtr [yyi].yyOld == yyOld) return yyOldToNewStorePtr [yyi].yyNew;
}

static tTree yyCopyTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyt, yyPtrtTree yyNew)
# else
 (yyt, yyNew) tTree yyt; yyPtrtTree yyNew;
# endif
{
 for (;;) {
  if (yyt == NoTree) { * yyNew = NoTree; return; }
  if (yyt->yyHead.yyMark == 0) { * yyNew = yyMapOldToNew (yyt); return; }
  yyALLOC (* yyNew, Tree_NodeSize [yyt->Kind])
  if (yyt->yyHead.yyMark > 1) { yyStoreOldToNew (yyt, * yyNew); }
  yyt->yyHead.yyMark = 0;

  switch (yyt->Kind) {
case kCOMP_UNIT: (* yyNew)->COMP_UNIT = yyt->COMP_UNIT;
copytTree ((* yyNew)->COMP_UNIT.COMP_ELEMENTS, yyt->COMP_UNIT.COMP_ELEMENTS)
yyt = yyt->COMP_UNIT.USED_ELEMENTS;
yyNew = & (* yyNew)->COMP_UNIT.USED_ELEMENTS; break;
case kUNIT_NODE_LIST: (* yyNew)->UNIT_NODE_LIST = yyt->UNIT_NODE_LIST;
return;
case kUNIT_EMPTY: (* yyNew)->UNIT_EMPTY = yyt->UNIT_EMPTY;
return;
case kUNIT_LIST: (* yyNew)->UNIT_LIST = yyt->UNIT_LIST;
copytTree ((* yyNew)->UNIT_LIST.Elem, yyt->UNIT_LIST.Elem)
yyt = yyt->UNIT_LIST.Next;
yyNew = & (* yyNew)->UNIT_LIST.Next; break;
case kUNIT_NODE: (* yyNew)->UNIT_NODE = yyt->UNIT_NODE;
copytIdent ((* yyNew)->UNIT_NODE.Ident, yyt->UNIT_NODE.Ident)
copyint ((* yyNew)->UNIT_NODE.Line, yyt->UNIT_NODE.Line)
copytDefinitions ((* yyNew)->UNIT_NODE.Object, yyt->UNIT_NODE.Object)
return;
case kMODULE_DECL: (* yyNew)->MODULE_DECL = yyt->MODULE_DECL;
copytIdent ((* yyNew)->MODULE_DECL.Ident, yyt->MODULE_DECL.Ident)
copyint ((* yyNew)->MODULE_DECL.Line, yyt->MODULE_DECL.Line)
copytDefinitions ((* yyNew)->MODULE_DECL.Object, yyt->MODULE_DECL.Object)
yyt = yyt->MODULE_DECL.MODULE_BODY;
yyNew = & (* yyNew)->MODULE_DECL.MODULE_BODY; break;
case kPROGRAM_DECL: (* yyNew)->PROGRAM_DECL = yyt->PROGRAM_DECL;
copytIdent ((* yyNew)->PROGRAM_DECL.Ident, yyt->PROGRAM_DECL.Ident)
copyint ((* yyNew)->PROGRAM_DECL.Line, yyt->PROGRAM_DECL.Line)
copytDefinitions ((* yyNew)->PROGRAM_DECL.Object, yyt->PROGRAM_DECL.Object)
copytTree ((* yyNew)->PROGRAM_DECL.FORMALS, yyt->PROGRAM_DECL.FORMALS)
yyt = yyt->PROGRAM_DECL.PROGRAM_BODY;
yyNew = & (* yyNew)->PROGRAM_DECL.PROGRAM_BODY; break;
case kPROC_DECL: (* yyNew)->PROC_DECL = yyt->PROC_DECL;
copytIdent ((* yyNew)->PROC_DECL.Ident, yyt->PROC_DECL.Ident)
copyint ((* yyNew)->PROC_DECL.Line, yyt->PROC_DECL.Line)
copytDefinitions ((* yyNew)->PROC_DECL.Object, yyt->PROC_DECL.Object)
copytTree ((* yyNew)->PROC_DECL.FORMALS, yyt->PROC_DECL.FORMALS)
copybool ((* yyNew)->PROC_DECL.IsRecursive, yyt->PROC_DECL.IsRecursive)
copybool ((* yyNew)->PROC_DECL.IsPure, yyt->PROC_DECL.IsPure)
copytIdent ((* yyNew)->PROC_DECL.HPFExtrinsic, yyt->PROC_DECL.HPFExtrinsic)
yyt = yyt->PROC_DECL.PROC_BODY;
yyNew = & (* yyNew)->PROC_DECL.PROC_BODY; break;
case kFUNC_DECL: (* yyNew)->FUNC_DECL = yyt->FUNC_DECL;
copytIdent ((* yyNew)->FUNC_DECL.Ident, yyt->FUNC_DECL.Ident)
copyint ((* yyNew)->FUNC_DECL.Line, yyt->FUNC_DECL.Line)
copytDefinitions ((* yyNew)->FUNC_DECL.Object, yyt->FUNC_DECL.Object)
copytTree ((* yyNew)->FUNC_DECL.FORMALS, yyt->FUNC_DECL.FORMALS)
copytTree ((* yyNew)->FUNC_DECL.FUNC_BODY, yyt->FUNC_DECL.FUNC_BODY)
copytIdent ((* yyNew)->FUNC_DECL.RESULT_ID, yyt->FUNC_DECL.RESULT_ID)
copybool ((* yyNew)->FUNC_DECL.IsRecursive, yyt->FUNC_DECL.IsRecursive)
copybool ((* yyNew)->FUNC_DECL.IsPure, yyt->FUNC_DECL.IsPure)
copytIdent ((* yyNew)->FUNC_DECL.HPFExtrinsic, yyt->FUNC_DECL.HPFExtrinsic)
yyt = yyt->FUNC_DECL.RESULT_TYPE;
yyNew = & (* yyNew)->FUNC_DECL.RESULT_TYPE; break;
case kMODULE_PROC_DECL: (* yyNew)->MODULE_PROC_DECL = yyt->MODULE_PROC_DECL;
copytIdent ((* yyNew)->MODULE_PROC_DECL.Ident, yyt->MODULE_PROC_DECL.Ident)
copyint ((* yyNew)->MODULE_PROC_DECL.Line, yyt->MODULE_PROC_DECL.Line)
copytDefinitions ((* yyNew)->MODULE_PROC_DECL.Object, yyt->MODULE_PROC_DECL.Object)
return;
case kBLOCK_DATA_DECL: (* yyNew)->BLOCK_DATA_DECL = yyt->BLOCK_DATA_DECL;
copytIdent ((* yyNew)->BLOCK_DATA_DECL.Ident, yyt->BLOCK_DATA_DECL.Ident)
copyint ((* yyNew)->BLOCK_DATA_DECL.Line, yyt->BLOCK_DATA_DECL.Line)
copytDefinitions ((* yyNew)->BLOCK_DATA_DECL.Object, yyt->BLOCK_DATA_DECL.Object)
yyt = yyt->BLOCK_DATA_DECL.DATA_BODY;
yyNew = & (* yyNew)->BLOCK_DATA_DECL.DATA_BODY; break;
case kINTERFACE: (* yyNew)->INTERFACE = yyt->INTERFACE;
copytIdent ((* yyNew)->INTERFACE.Ident, yyt->INTERFACE.Ident)
copyint ((* yyNew)->INTERFACE.Line, yyt->INTERFACE.Line)
copytDefinitions ((* yyNew)->INTERFACE.Object, yyt->INTERFACE.Object)
copytTree ((* yyNew)->INTERFACE.SPEC, yyt->INTERFACE.SPEC)
yyt = yyt->INTERFACE.ITEMS;
yyNew = & (* yyNew)->INTERFACE.ITEMS; break;
case kBODY_NODE: (* yyNew)->BODY_NODE = yyt->BODY_NODE;
copytTree ((* yyNew)->BODY_NODE.DECLS, yyt->BODY_NODE.DECLS)
copytTree ((* yyNew)->BODY_NODE.STATS, yyt->BODY_NODE.STATS)
yyt = yyt->BODY_NODE.INTERNALS;
yyNew = & (* yyNew)->BODY_NODE.INTERNALS; break;
case kTYPE_NODE: (* yyNew)->TYPE_NODE = yyt->TYPE_NODE;
return;
case kDUMMY_TYPE: (* yyNew)->DUMMY_TYPE = yyt->DUMMY_TYPE;
return;
case kALIGN_TYPE: (* yyNew)->ALIGN_TYPE = yyt->ALIGN_TYPE;
return;
case kINTEGER_TYPE: (* yyNew)->INTEGER_TYPE = yyt->INTEGER_TYPE;
yyt = yyt->INTEGER_TYPE.SIZE;
yyNew = & (* yyNew)->INTEGER_TYPE.SIZE; break;
case kREAL_TYPE: (* yyNew)->REAL_TYPE = yyt->REAL_TYPE;
yyt = yyt->REAL_TYPE.SIZE;
yyNew = & (* yyNew)->REAL_TYPE.SIZE; break;
case kBOOLEAN_TYPE: (* yyNew)->BOOLEAN_TYPE = yyt->BOOLEAN_TYPE;
yyt = yyt->BOOLEAN_TYPE.SIZE;
yyNew = & (* yyNew)->BOOLEAN_TYPE.SIZE; break;
case kCOMPLEX_TYPE: (* yyNew)->COMPLEX_TYPE = yyt->COMPLEX_TYPE;
yyt = yyt->COMPLEX_TYPE.SIZE;
yyNew = & (* yyNew)->COMPLEX_TYPE.SIZE; break;
case kSTRING_TYPE: (* yyNew)->STRING_TYPE = yyt->STRING_TYPE;
yyt = yyt->STRING_TYPE.LENGTH;
yyNew = & (* yyNew)->STRING_TYPE.LENGTH; break;
case kTYPE_ID: (* yyNew)->TYPE_ID = yyt->TYPE_ID;
yyt = yyt->TYPE_ID.ID;
yyNew = & (* yyNew)->TYPE_ID.ID; break;
case kARRAY_TYPE: (* yyNew)->ARRAY_TYPE = yyt->ARRAY_TYPE;
copytTree ((* yyNew)->ARRAY_TYPE.ARRAY_INDEX_TYPES, yyt->ARRAY_TYPE.ARRAY_INDEX_TYPES)
yyt = yyt->ARRAY_TYPE.ARRAY_COMP_TYPE;
yyNew = & (* yyNew)->ARRAY_TYPE.ARRAY_COMP_TYPE; break;
case kRECORD_TYPE: (* yyNew)->RECORD_TYPE = yyt->RECORD_TYPE;
copytIdent ((* yyNew)->RECORD_TYPE.Name, yyt->RECORD_TYPE.Name)
yyt = yyt->RECORD_TYPE.COMPONENTS;
yyNew = & (* yyNew)->RECORD_TYPE.COMPONENTS; break;
case kPOINTER_TYPE: (* yyNew)->POINTER_TYPE = yyt->POINTER_TYPE;
yyt = yyt->POINTER_TYPE.PTR_COMP;
yyNew = & (* yyNew)->POINTER_TYPE.PTR_COMP; break;
case kSHAPE_SPEC: (* yyNew)->SHAPE_SPEC = yyt->SHAPE_SPEC;
yyt = yyt->SHAPE_SPEC.Overlap;
yyNew = & (* yyNew)->SHAPE_SPEC.Overlap; break;
case kEXPLICIT_SHAPE: (* yyNew)->EXPLICIT_SHAPE = yyt->EXPLICIT_SHAPE;
copytTree ((* yyNew)->EXPLICIT_SHAPE.Overlap, yyt->EXPLICIT_SHAPE.Overlap)
copytTree ((* yyNew)->EXPLICIT_SHAPE.LOWER, yyt->EXPLICIT_SHAPE.LOWER)
yyt = yyt->EXPLICIT_SHAPE.UPPER;
yyNew = & (* yyNew)->EXPLICIT_SHAPE.UPPER; break;
case kASSUMED_SHAPE: (* yyNew)->ASSUMED_SHAPE = yyt->ASSUMED_SHAPE;
copytTree ((* yyNew)->ASSUMED_SHAPE.Overlap, yyt->ASSUMED_SHAPE.Overlap)
yyt = yyt->ASSUMED_SHAPE.LOWER;
yyNew = & (* yyNew)->ASSUMED_SHAPE.LOWER; break;
case kDEFERRED_SHAPE: (* yyNew)->DEFERRED_SHAPE = yyt->DEFERRED_SHAPE;
yyt = yyt->DEFERRED_SHAPE.Overlap;
yyNew = & (* yyNew)->DEFERRED_SHAPE.Overlap; break;
case kASSUMED_SIZE: (* yyNew)->ASSUMED_SIZE = yyt->ASSUMED_SIZE;
copytTree ((* yyNew)->ASSUMED_SIZE.Overlap, yyt->ASSUMED_SIZE.Overlap)
yyt = yyt->ASSUMED_SIZE.LOWER;
yyNew = & (* yyNew)->ASSUMED_SIZE.LOWER; break;
case kSHAPE_SPEC_LIST: (* yyNew)->SHAPE_SPEC_LIST = yyt->SHAPE_SPEC_LIST;
return;
case kSHAPE_EMPTY: (* yyNew)->SHAPE_EMPTY = yyt->SHAPE_EMPTY;
return;
case kSHAPE_LIST: (* yyNew)->SHAPE_LIST = yyt->SHAPE_LIST;
copytTree ((* yyNew)->SHAPE_LIST.Elem, yyt->SHAPE_LIST.Elem)
yyt = yyt->SHAPE_LIST.Next;
yyNew = & (* yyNew)->SHAPE_LIST.Next; break;
case kDECL_NODE: (* yyNew)->DECL_NODE = yyt->DECL_NODE;
copytIdent ((* yyNew)->DECL_NODE.Ident, yyt->DECL_NODE.Ident)
copyint ((* yyNew)->DECL_NODE.Line, yyt->DECL_NODE.Line)
return;
case kTYPE_DECL: (* yyNew)->TYPE_DECL = yyt->TYPE_DECL;
copytIdent ((* yyNew)->TYPE_DECL.Ident, yyt->TYPE_DECL.Ident)
copyint ((* yyNew)->TYPE_DECL.Line, yyt->TYPE_DECL.Line)
yyt = yyt->TYPE_DECL.VAL;
yyNew = & (* yyNew)->TYPE_DECL.VAL; break;
case kNAME_DECL: (* yyNew)->NAME_DECL = yyt->NAME_DECL;
copytIdent ((* yyNew)->NAME_DECL.Ident, yyt->NAME_DECL.Ident)
copyint ((* yyNew)->NAME_DECL.Line, yyt->NAME_DECL.Line)
return;
case kRENAME_DECL: (* yyNew)->RENAME_DECL = yyt->RENAME_DECL;
copytIdent ((* yyNew)->RENAME_DECL.Ident, yyt->RENAME_DECL.Ident)
copyint ((* yyNew)->RENAME_DECL.Line, yyt->RENAME_DECL.Line)
copytIdent ((* yyNew)->RENAME_DECL.oldname, yyt->RENAME_DECL.oldname)
return;
case kENTITY_DECL: (* yyNew)->ENTITY_DECL = yyt->ENTITY_DECL;
copytIdent ((* yyNew)->ENTITY_DECL.Ident, yyt->ENTITY_DECL.Ident)
copyint ((* yyNew)->ENTITY_DECL.Line, yyt->ENTITY_DECL.Line)
yyt = yyt->ENTITY_DECL.ATTRIBUTES;
yyNew = & (* yyNew)->ENTITY_DECL.ATTRIBUTES; break;
case kVAR_DECL: (* yyNew)->VAR_DECL = yyt->VAR_DECL;
copytIdent ((* yyNew)->VAR_DECL.Ident, yyt->VAR_DECL.Ident)
copyint ((* yyNew)->VAR_DECL.Line, yyt->VAR_DECL.Line)
yyt = yyt->VAR_DECL.VAL;
yyNew = & (* yyNew)->VAR_DECL.VAL; break;
case kCOMMON_DECL: (* yyNew)->COMMON_DECL = yyt->COMMON_DECL;
copytIdent ((* yyNew)->COMMON_DECL.Ident, yyt->COMMON_DECL.Ident)
copyint ((* yyNew)->COMMON_DECL.Line, yyt->COMMON_DECL.Line)
yyt = yyt->COMMON_DECL.IDS;
yyNew = & (* yyNew)->COMMON_DECL.IDS; break;
case kNAMELIST_DECL: (* yyNew)->NAMELIST_DECL = yyt->NAMELIST_DECL;
copytIdent ((* yyNew)->NAMELIST_DECL.Ident, yyt->NAMELIST_DECL.Ident)
copyint ((* yyNew)->NAMELIST_DECL.Line, yyt->NAMELIST_DECL.Line)
yyt = yyt->NAMELIST_DECL.IDS;
yyNew = & (* yyNew)->NAMELIST_DECL.IDS; break;
case kTEMPLATE_DECL: (* yyNew)->TEMPLATE_DECL = yyt->TEMPLATE_DECL;
copytIdent ((* yyNew)->TEMPLATE_DECL.Ident, yyt->TEMPLATE_DECL.Ident)
copyint ((* yyNew)->TEMPLATE_DECL.Line, yyt->TEMPLATE_DECL.Line)
yyt = yyt->TEMPLATE_DECL.DIMENSIONS;
yyNew = & (* yyNew)->TEMPLATE_DECL.DIMENSIONS; break;
case kRAGGED_DECL: (* yyNew)->RAGGED_DECL = yyt->RAGGED_DECL;
copytIdent ((* yyNew)->RAGGED_DECL.Ident, yyt->RAGGED_DECL.Ident)
copyint ((* yyNew)->RAGGED_DECL.Line, yyt->RAGGED_DECL.Line)
yyt = yyt->RAGGED_DECL.DIMENSIONS;
yyNew = & (* yyNew)->RAGGED_DECL.DIMENSIONS; break;
case kPROCESSORS_DECL: (* yyNew)->PROCESSORS_DECL = yyt->PROCESSORS_DECL;
copytIdent ((* yyNew)->PROCESSORS_DECL.Ident, yyt->PROCESSORS_DECL.Ident)
copyint ((* yyNew)->PROCESSORS_DECL.Line, yyt->PROCESSORS_DECL.Line)
yyt = yyt->PROCESSORS_DECL.DIMENSIONS;
yyNew = & (* yyNew)->PROCESSORS_DECL.DIMENSIONS; break;
case kSUB_PROCS_DECL: (* yyNew)->SUB_PROCS_DECL = yyt->SUB_PROCS_DECL;
copytIdent ((* yyNew)->SUB_PROCS_DECL.Ident, yyt->SUB_PROCS_DECL.Ident)
copyint ((* yyNew)->SUB_PROCS_DECL.Line, yyt->SUB_PROCS_DECL.Line)
copytTree ((* yyNew)->SUB_PROCS_DECL.TOP, yyt->SUB_PROCS_DECL.TOP)
yyt = yyt->SUB_PROCS_DECL.DIMENSIONS;
yyNew = & (* yyNew)->SUB_PROCS_DECL.DIMENSIONS; break;
case kINHERIT_DECL: (* yyNew)->INHERIT_DECL = yyt->INHERIT_DECL;
copytIdent ((* yyNew)->INHERIT_DECL.Ident, yyt->INHERIT_DECL.Ident)
copyint ((* yyNew)->INHERIT_DECL.Line, yyt->INHERIT_DECL.Line)
return;
case kRANGE_DECL: (* yyNew)->RANGE_DECL = yyt->RANGE_DECL;
copytIdent ((* yyNew)->RANGE_DECL.Ident, yyt->RANGE_DECL.Ident)
copyint ((* yyNew)->RANGE_DECL.Line, yyt->RANGE_DECL.Line)
yyt = yyt->RANGE_DECL.RANGES;
yyNew = & (* yyNew)->RANGE_DECL.RANGES; break;
case kSTMT_FUNC_DECL: (* yyNew)->STMT_FUNC_DECL = yyt->STMT_FUNC_DECL;
copytIdent ((* yyNew)->STMT_FUNC_DECL.Ident, yyt->STMT_FUNC_DECL.Ident)
copyint ((* yyNew)->STMT_FUNC_DECL.Line, yyt->STMT_FUNC_DECL.Line)
copytTree ((* yyNew)->STMT_FUNC_DECL.FORMALS, yyt->STMT_FUNC_DECL.FORMALS)
copytTree ((* yyNew)->STMT_FUNC_DECL.RESULT_TYPE, yyt->STMT_FUNC_DECL.RESULT_TYPE)
yyt = yyt->STMT_FUNC_DECL.FFUNC_BODY;
yyNew = & (* yyNew)->STMT_FUNC_DECL.FFUNC_BODY; break;
case kINTERFACE_DECL: (* yyNew)->INTERFACE_DECL = yyt->INTERFACE_DECL;
copytIdent ((* yyNew)->INTERFACE_DECL.Ident, yyt->INTERFACE_DECL.Ident)
copyint ((* yyNew)->INTERFACE_DECL.Line, yyt->INTERFACE_DECL.Line)
copytTree ((* yyNew)->INTERFACE_DECL.SPEC, yyt->INTERFACE_DECL.SPEC)
yyt = yyt->INTERFACE_DECL.ITEMS;
yyNew = & (* yyNew)->INTERFACE_DECL.ITEMS; break;
case kVAR_PARAM_DECL: (* yyNew)->VAR_PARAM_DECL = yyt->VAR_PARAM_DECL;
copytIdent ((* yyNew)->VAR_PARAM_DECL.Ident, yyt->VAR_PARAM_DECL.Ident)
copyint ((* yyNew)->VAR_PARAM_DECL.Line, yyt->VAR_PARAM_DECL.Line)
yyt = yyt->VAR_PARAM_DECL.VAL;
yyNew = & (* yyNew)->VAR_PARAM_DECL.VAL; break;
case kRET_PARAM_DECL: (* yyNew)->RET_PARAM_DECL = yyt->RET_PARAM_DECL;
copytIdent ((* yyNew)->RET_PARAM_DECL.Ident, yyt->RET_PARAM_DECL.Ident)
copyint ((* yyNew)->RET_PARAM_DECL.Line, yyt->RET_PARAM_DECL.Line)
return;
case kIMPLICIT_DECL: (* yyNew)->IMPLICIT_DECL = yyt->IMPLICIT_DECL;
copytIdent ((* yyNew)->IMPLICIT_DECL.Ident, yyt->IMPLICIT_DECL.Ident)
copyint ((* yyNew)->IMPLICIT_DECL.Line, yyt->IMPLICIT_DECL.Line)
copytIdent ((* yyNew)->IMPLICIT_DECL.first, yyt->IMPLICIT_DECL.first)
copytIdent ((* yyNew)->IMPLICIT_DECL.last, yyt->IMPLICIT_DECL.last)
yyt = yyt->IMPLICIT_DECL.TYPE;
yyNew = & (* yyNew)->IMPLICIT_DECL.TYPE; break;
case kUSE_DECL: (* yyNew)->USE_DECL = yyt->USE_DECL;
copytIdent ((* yyNew)->USE_DECL.Ident, yyt->USE_DECL.Ident)
copyint ((* yyNew)->USE_DECL.Line, yyt->USE_DECL.Line)
copytIdent ((* yyNew)->USE_DECL.use, yyt->USE_DECL.use)
yyt = yyt->USE_DECL.RENAMINGS;
yyNew = & (* yyNew)->USE_DECL.RENAMINGS; break;
case kONLY_USE_DECL: (* yyNew)->ONLY_USE_DECL = yyt->ONLY_USE_DECL;
copytIdent ((* yyNew)->ONLY_USE_DECL.Ident, yyt->ONLY_USE_DECL.Ident)
copyint ((* yyNew)->ONLY_USE_DECL.Line, yyt->ONLY_USE_DECL.Line)
copytIdent ((* yyNew)->ONLY_USE_DECL.use, yyt->ONLY_USE_DECL.use)
yyt = yyt->ONLY_USE_DECL.NAMES;
yyNew = & (* yyNew)->ONLY_USE_DECL.NAMES; break;
case kEQV_DECL: (* yyNew)->EQV_DECL = yyt->EQV_DECL;
copytIdent ((* yyNew)->EQV_DECL.Ident, yyt->EQV_DECL.Ident)
copyint ((* yyNew)->EQV_DECL.Line, yyt->EQV_DECL.Line)
yyt = yyt->EQV_DECL.VARS;
yyNew = & (* yyNew)->EQV_DECL.VARS; break;
case kDATA_DECL: (* yyNew)->DATA_DECL = yyt->DATA_DECL;
copytIdent ((* yyNew)->DATA_DECL.Ident, yyt->DATA_DECL.Ident)
copyint ((* yyNew)->DATA_DECL.Line, yyt->DATA_DECL.Line)
copytTree ((* yyNew)->DATA_DECL.VARS, yyt->DATA_DECL.VARS)
yyt = yyt->DATA_DECL.VALS;
yyNew = & (* yyNew)->DATA_DECL.VALS; break;
case kFORMAT_DECL: (* yyNew)->FORMAT_DECL = yyt->FORMAT_DECL;
copytIdent ((* yyNew)->FORMAT_DECL.Ident, yyt->FORMAT_DECL.Ident)
copyint ((* yyNew)->FORMAT_DECL.Line, yyt->FORMAT_DECL.Line)
copyint ((* yyNew)->FORMAT_DECL.label, yyt->FORMAT_DECL.label)
yyt = yyt->FORMAT_DECL.FORMAT_PARAMS;
yyNew = & (* yyNew)->FORMAT_DECL.FORMAT_PARAMS; break;
case kDIMENSION_DECL: (* yyNew)->DIMENSION_DECL = yyt->DIMENSION_DECL;
copytIdent ((* yyNew)->DIMENSION_DECL.Ident, yyt->DIMENSION_DECL.Ident)
copyint ((* yyNew)->DIMENSION_DECL.Line, yyt->DIMENSION_DECL.Line)
yyt = yyt->DIMENSION_DECL.INDEXES;
yyNew = & (* yyNew)->DIMENSION_DECL.INDEXES; break;
case kINIT_DATA_DECL: (* yyNew)->INIT_DATA_DECL = yyt->INIT_DATA_DECL;
copytIdent ((* yyNew)->INIT_DATA_DECL.Ident, yyt->INIT_DATA_DECL.Ident)
copyint ((* yyNew)->INIT_DATA_DECL.Line, yyt->INIT_DATA_DECL.Line)
yyt = yyt->INIT_DATA_DECL.VAL;
yyNew = & (* yyNew)->INIT_DATA_DECL.VAL; break;
case kTYPESPEC_DECL: (* yyNew)->TYPESPEC_DECL = yyt->TYPESPEC_DECL;
copytIdent ((* yyNew)->TYPESPEC_DECL.Ident, yyt->TYPESPEC_DECL.Ident)
copyint ((* yyNew)->TYPESPEC_DECL.Line, yyt->TYPESPEC_DECL.Line)
yyt = yyt->TYPESPEC_DECL.VAL;
yyNew = & (* yyNew)->TYPESPEC_DECL.VAL; break;
case kALLOCATABLE_DECL: (* yyNew)->ALLOCATABLE_DECL = yyt->ALLOCATABLE_DECL;
copytIdent ((* yyNew)->ALLOCATABLE_DECL.Ident, yyt->ALLOCATABLE_DECL.Ident)
copyint ((* yyNew)->ALLOCATABLE_DECL.Line, yyt->ALLOCATABLE_DECL.Line)
return;
case kEXTERNAL_DECL: (* yyNew)->EXTERNAL_DECL = yyt->EXTERNAL_DECL;
copytIdent ((* yyNew)->EXTERNAL_DECL.Ident, yyt->EXTERNAL_DECL.Ident)
copyint ((* yyNew)->EXTERNAL_DECL.Line, yyt->EXTERNAL_DECL.Line)
return;
case kINTENT_DECL: (* yyNew)->INTENT_DECL = yyt->INTENT_DECL;
copytIdent ((* yyNew)->INTENT_DECL.Ident, yyt->INTENT_DECL.Ident)
copyint ((* yyNew)->INTENT_DECL.Line, yyt->INTENT_DECL.Line)
copyint ((* yyNew)->INTENT_DECL.intent, yyt->INTENT_DECL.intent)
return;
case kINTRINSIC_DECL: (* yyNew)->INTRINSIC_DECL = yyt->INTRINSIC_DECL;
copytIdent ((* yyNew)->INTRINSIC_DECL.Ident, yyt->INTRINSIC_DECL.Ident)
copyint ((* yyNew)->INTRINSIC_DECL.Line, yyt->INTRINSIC_DECL.Line)
copytIdent ((* yyNew)->INTRINSIC_DECL.generic, yyt->INTRINSIC_DECL.generic)
copyint ((* yyNew)->INTRINSIC_DECL.class, yyt->INTRINSIC_DECL.class)
copyint ((* yyNew)->INTRINSIC_DECL.kind, yyt->INTRINSIC_DECL.kind)
return;
case kOPTIONAL_DECL: (* yyNew)->OPTIONAL_DECL = yyt->OPTIONAL_DECL;
copytIdent ((* yyNew)->OPTIONAL_DECL.Ident, yyt->OPTIONAL_DECL.Ident)
copyint ((* yyNew)->OPTIONAL_DECL.Line, yyt->OPTIONAL_DECL.Line)
return;
case kPOINTER_DECL: (* yyNew)->POINTER_DECL = yyt->POINTER_DECL;
copytIdent ((* yyNew)->POINTER_DECL.Ident, yyt->POINTER_DECL.Ident)
copyint ((* yyNew)->POINTER_DECL.Line, yyt->POINTER_DECL.Line)
return;
case kSAVE_DECL: (* yyNew)->SAVE_DECL = yyt->SAVE_DECL;
copytIdent ((* yyNew)->SAVE_DECL.Ident, yyt->SAVE_DECL.Ident)
copyint ((* yyNew)->SAVE_DECL.Line, yyt->SAVE_DECL.Line)
return;
case kTARGET_DECL: (* yyNew)->TARGET_DECL = yyt->TARGET_DECL;
copytIdent ((* yyNew)->TARGET_DECL.Ident, yyt->TARGET_DECL.Ident)
copyint ((* yyNew)->TARGET_DECL.Line, yyt->TARGET_DECL.Line)
return;
case kPARAMETER_DECL: (* yyNew)->PARAMETER_DECL = yyt->PARAMETER_DECL;
copytIdent ((* yyNew)->PARAMETER_DECL.Ident, yyt->PARAMETER_DECL.Ident)
copyint ((* yyNew)->PARAMETER_DECL.Line, yyt->PARAMETER_DECL.Line)
yyt = yyt->PARAMETER_DECL.VAL;
yyNew = & (* yyNew)->PARAMETER_DECL.VAL; break;
case kPUBLIC_DECL: (* yyNew)->PUBLIC_DECL = yyt->PUBLIC_DECL;
copytIdent ((* yyNew)->PUBLIC_DECL.Ident, yyt->PUBLIC_DECL.Ident)
copyint ((* yyNew)->PUBLIC_DECL.Line, yyt->PUBLIC_DECL.Line)
return;
case kPRIVATE_DECL: (* yyNew)->PRIVATE_DECL = yyt->PRIVATE_DECL;
copytIdent ((* yyNew)->PRIVATE_DECL.Ident, yyt->PRIVATE_DECL.Ident)
copyint ((* yyNew)->PRIVATE_DECL.Line, yyt->PRIVATE_DECL.Line)
return;
case kSEQUENCE_DECL: (* yyNew)->SEQUENCE_DECL = yyt->SEQUENCE_DECL;
copytIdent ((* yyNew)->SEQUENCE_DECL.Ident, yyt->SEQUENCE_DECL.Ident)
copyint ((* yyNew)->SEQUENCE_DECL.Line, yyt->SEQUENCE_DECL.Line)
return;
case kDISTRIBUTE_DECL: (* yyNew)->DISTRIBUTE_DECL = yyt->DISTRIBUTE_DECL;
copytIdent ((* yyNew)->DISTRIBUTE_DECL.Ident, yyt->DISTRIBUTE_DECL.Ident)
copyint ((* yyNew)->DISTRIBUTE_DECL.Line, yyt->DISTRIBUTE_DECL.Line)
copytTree ((* yyNew)->DISTRIBUTE_DECL.DISTRIBUTION, yyt->DISTRIBUTE_DECL.DISTRIBUTION)
yyt = yyt->DISTRIBUTE_DECL.TARGET;
yyNew = & (* yyNew)->DISTRIBUTE_DECL.TARGET; break;
case kALIGN_DECL: (* yyNew)->ALIGN_DECL = yyt->ALIGN_DECL;
copytIdent ((* yyNew)->ALIGN_DECL.Ident, yyt->ALIGN_DECL.Ident)
copyint ((* yyNew)->ALIGN_DECL.Line, yyt->ALIGN_DECL.Line)
copytTree ((* yyNew)->ALIGN_DECL.ALIGN_SOURCE, yyt->ALIGN_DECL.ALIGN_SOURCE)
yyt = yyt->ALIGN_DECL.ALIGN_TARGET;
yyNew = & (* yyNew)->ALIGN_DECL.ALIGN_TARGET; break;
case kMAP_DECL: (* yyNew)->MAP_DECL = yyt->MAP_DECL;
copytIdent ((* yyNew)->MAP_DECL.Ident, yyt->MAP_DECL.Ident)
copyint ((* yyNew)->MAP_DECL.Line, yyt->MAP_DECL.Line)
yyt = yyt->MAP_DECL.MAP_TARGET;
yyNew = & (* yyNew)->MAP_DECL.MAP_TARGET; break;
case kOVERLAP_DECL: (* yyNew)->OVERLAP_DECL = yyt->OVERLAP_DECL;
copytIdent ((* yyNew)->OVERLAP_DECL.Ident, yyt->OVERLAP_DECL.Ident)
copyint ((* yyNew)->OVERLAP_DECL.Line, yyt->OVERLAP_DECL.Line)
yyt = yyt->OVERLAP_DECL.OVERLAPS;
yyNew = & (* yyNew)->OVERLAP_DECL.OVERLAPS; break;
case kSHARED_DECL: (* yyNew)->SHARED_DECL = yyt->SHARED_DECL;
copytIdent ((* yyNew)->SHARED_DECL.Ident, yyt->SHARED_DECL.Ident)
copyint ((* yyNew)->SHARED_DECL.Line, yyt->SHARED_DECL.Line)
copyint ((* yyNew)->SHARED_DECL.kind, yyt->SHARED_DECL.kind)
return;
case kDYNAMIC_DECL: (* yyNew)->DYNAMIC_DECL = yyt->DYNAMIC_DECL;
copytIdent ((* yyNew)->DYNAMIC_DECL.Ident, yyt->DYNAMIC_DECL.Ident)
copyint ((* yyNew)->DYNAMIC_DECL.Line, yyt->DYNAMIC_DECL.Line)
return;
case kTRACE_DECL: (* yyNew)->TRACE_DECL = yyt->TRACE_DECL;
copytIdent ((* yyNew)->TRACE_DECL.Ident, yyt->TRACE_DECL.Ident)
copyint ((* yyNew)->TRACE_DECL.Line, yyt->TRACE_DECL.Line)
return;
case kTREE_DECL: (* yyNew)->TREE_DECL = yyt->TREE_DECL;
copytIdent ((* yyNew)->TREE_DECL.Ident, yyt->TREE_DECL.Ident)
copyint ((* yyNew)->TREE_DECL.Line, yyt->TREE_DECL.Line)
return;
case kSELECT_DECL: (* yyNew)->SELECT_DECL = yyt->SELECT_DECL;
copytIdent ((* yyNew)->SELECT_DECL.Ident, yyt->SELECT_DECL.Ident)
copyint ((* yyNew)->SELECT_DECL.Line, yyt->SELECT_DECL.Line)
yyt = yyt->SELECT_DECL.SELECTIONS;
yyNew = & (* yyNew)->SELECT_DECL.SELECTIONS; break;
case kHPF_SEQUENCE_DECL: (* yyNew)->HPF_SEQUENCE_DECL = yyt->HPF_SEQUENCE_DECL;
copytIdent ((* yyNew)->HPF_SEQUENCE_DECL.Ident, yyt->HPF_SEQUENCE_DECL.Ident)
copyint ((* yyNew)->HPF_SEQUENCE_DECL.Line, yyt->HPF_SEQUENCE_DECL.Line)
return;
case kNOSEQUENCE_DECL: (* yyNew)->NOSEQUENCE_DECL = yyt->NOSEQUENCE_DECL;
copytIdent ((* yyNew)->NOSEQUENCE_DECL.Ident, yyt->NOSEQUENCE_DECL.Ident)
copyint ((* yyNew)->NOSEQUENCE_DECL.Line, yyt->NOSEQUENCE_DECL.Line)
return;
case kLAYOUT_DECL: (* yyNew)->LAYOUT_DECL = yyt->LAYOUT_DECL;
copytIdent ((* yyNew)->LAYOUT_DECL.Ident, yyt->LAYOUT_DECL.Ident)
copyint ((* yyNew)->LAYOUT_DECL.Line, yyt->LAYOUT_DECL.Line)
copytStringRef ((* yyNew)->LAYOUT_DECL.kind, yyt->LAYOUT_DECL.kind)
return;
case kPASS_BY_DECL: (* yyNew)->PASS_BY_DECL = yyt->PASS_BY_DECL;
copytIdent ((* yyNew)->PASS_BY_DECL.Ident, yyt->PASS_BY_DECL.Ident)
copyint ((* yyNew)->PASS_BY_DECL.Line, yyt->PASS_BY_DECL.Line)
copytStringRef ((* yyNew)->PASS_BY_DECL.kind, yyt->PASS_BY_DECL.kind)
return;
case kMAP_TO_DECL: (* yyNew)->MAP_TO_DECL = yyt->MAP_TO_DECL;
copytIdent ((* yyNew)->MAP_TO_DECL.Ident, yyt->MAP_TO_DECL.Ident)
copyint ((* yyNew)->MAP_TO_DECL.Line, yyt->MAP_TO_DECL.Line)
copytStringRef ((* yyNew)->MAP_TO_DECL.kind, yyt->MAP_TO_DECL.kind)
return;
case kDECL_NODE_LIST: (* yyNew)->DECL_NODE_LIST = yyt->DECL_NODE_LIST;
return;
case kDECL_EMPTY: (* yyNew)->DECL_EMPTY = yyt->DECL_EMPTY;
return;
case kDECL_LIST: (* yyNew)->DECL_LIST = yyt->DECL_LIST;
copytTree ((* yyNew)->DECL_LIST.Elem, yyt->DECL_LIST.Elem)
yyt = yyt->DECL_LIST.Next;
yyNew = & (* yyNew)->DECL_LIST.Next; break;
case kRANGE_SPEC_LIST: (* yyNew)->RANGE_SPEC_LIST = yyt->RANGE_SPEC_LIST;
return;
case kRANGE_EMPTY: (* yyNew)->RANGE_EMPTY = yyt->RANGE_EMPTY;
return;
case kRANGE_LIST: (* yyNew)->RANGE_LIST = yyt->RANGE_LIST;
copytTree ((* yyNew)->RANGE_LIST.Elem, yyt->RANGE_LIST.Elem)
yyt = yyt->RANGE_LIST.Next;
yyNew = & (* yyNew)->RANGE_LIST.Next; break;
case kSELECT_SPEC_LIST: (* yyNew)->SELECT_SPEC_LIST = yyt->SELECT_SPEC_LIST;
return;
case kSELECT_EMPTY: (* yyNew)->SELECT_EMPTY = yyt->SELECT_EMPTY;
return;
case kSELECT_LIST: (* yyNew)->SELECT_LIST = yyt->SELECT_LIST;
copytTree ((* yyNew)->SELECT_LIST.Elem, yyt->SELECT_LIST.Elem)
yyt = yyt->SELECT_LIST.Next;
yyNew = & (* yyNew)->SELECT_LIST.Next; break;
case kSELECT_SPEC: (* yyNew)->SELECT_SPEC = yyt->SELECT_SPEC;
copyint ((* yyNew)->SELECT_SPEC.selector, yyt->SELECT_SPEC.selector)
return;
case kDISTRIBUTION_SPEC: (* yyNew)->DISTRIBUTION_SPEC = yyt->DISTRIBUTION_SPEC;
copyint ((* yyNew)->DISTRIBUTION_SPEC.is_descriptive, yyt->DISTRIBUTION_SPEC.is_descriptive)
return;
case kANY_DISTRIBUTION: (* yyNew)->ANY_DISTRIBUTION = yyt->ANY_DISTRIBUTION;
copyint ((* yyNew)->ANY_DISTRIBUTION.is_descriptive, yyt->ANY_DISTRIBUTION.is_descriptive)
return;
case kDEFAULT_DISTRIBUTION: (* yyNew)->DEFAULT_DISTRIBUTION = yyt->DEFAULT_DISTRIBUTION;
copyint ((* yyNew)->DEFAULT_DISTRIBUTION.is_descriptive, yyt->DEFAULT_DISTRIBUTION.is_descriptive)
return;
case kREPL_DISTRIBUTION: (* yyNew)->REPL_DISTRIBUTION = yyt->REPL_DISTRIBUTION;
copyint ((* yyNew)->REPL_DISTRIBUTION.is_descriptive, yyt->REPL_DISTRIBUTION.is_descriptive)
return;
case kNODE_DISTRIBUTION: (* yyNew)->NODE_DISTRIBUTION = yyt->NODE_DISTRIBUTION;
copyint ((* yyNew)->NODE_DISTRIBUTION.is_descriptive, yyt->NODE_DISTRIBUTION.is_descriptive)
yyt = yyt->NODE_DISTRIBUTION.MAPPING;
yyNew = & (* yyNew)->NODE_DISTRIBUTION.MAPPING; break;
case kDIST_FORMAT_LIST: (* yyNew)->DIST_FORMAT_LIST = yyt->DIST_FORMAT_LIST;
return;
case kDIST_EMPTY: (* yyNew)->DIST_EMPTY = yyt->DIST_EMPTY;
return;
case kDIST_LIST: (* yyNew)->DIST_LIST = yyt->DIST_LIST;
copytTree ((* yyNew)->DIST_LIST.Elem, yyt->DIST_LIST.Elem)
yyt = yyt->DIST_LIST.Next;
yyNew = & (* yyNew)->DIST_LIST.Next; break;
case kDISTRIBUTION_FORMAT: (* yyNew)->DISTRIBUTION_FORMAT = yyt->DISTRIBUTION_FORMAT;
return;
case kBLOCK_DISTRIBUTION: (* yyNew)->BLOCK_DISTRIBUTION = yyt->BLOCK_DISTRIBUTION;
yyt = yyt->BLOCK_DISTRIBUTION.SIZE;
yyNew = & (* yyNew)->BLOCK_DISTRIBUTION.SIZE; break;
case kCYCLIC_DISTRIBUTION: (* yyNew)->CYCLIC_DISTRIBUTION = yyt->CYCLIC_DISTRIBUTION;
yyt = yyt->CYCLIC_DISTRIBUTION.SIZE;
yyNew = & (* yyNew)->CYCLIC_DISTRIBUTION.SIZE; break;
case kGEN_BLOCK_DISTRIBUTION: (* yyNew)->GEN_BLOCK_DISTRIBUTION = yyt->GEN_BLOCK_DISTRIBUTION;
yyt = yyt->GEN_BLOCK_DISTRIBUTION.ARRAY;
yyNew = & (* yyNew)->GEN_BLOCK_DISTRIBUTION.ARRAY; break;
case kINDIRECT_DISTRIBUTION: (* yyNew)->INDIRECT_DISTRIBUTION = yyt->INDIRECT_DISTRIBUTION;
yyt = yyt->INDIRECT_DISTRIBUTION.ARRAY;
yyNew = & (* yyNew)->INDIRECT_DISTRIBUTION.ARRAY; break;
case kARBITRARY_DISTRIBUTION: (* yyNew)->ARBITRARY_DISTRIBUTION = yyt->ARBITRARY_DISTRIBUTION;
copytTree ((* yyNew)->ARBITRARY_DISTRIBUTION.SIZE, yyt->ARBITRARY_DISTRIBUTION.SIZE)
copytTree ((* yyNew)->ARBITRARY_DISTRIBUTION.LENGTH, yyt->ARBITRARY_DISTRIBUTION.LENGTH)
yyt = yyt->ARBITRARY_DISTRIBUTION.MAP;
yyNew = & (* yyNew)->ARBITRARY_DISTRIBUTION.MAP; break;
case kSERIAL_DISTRIBUTION: (* yyNew)->SERIAL_DISTRIBUTION = yyt->SERIAL_DISTRIBUTION;
return;
case kANY_BLOCK_DISTRIBUTION: (* yyNew)->ANY_BLOCK_DISTRIBUTION = yyt->ANY_BLOCK_DISTRIBUTION;
return;
case kANY_CYCLIC_DISTRIBUTION: (* yyNew)->ANY_CYCLIC_DISTRIBUTION = yyt->ANY_CYCLIC_DISTRIBUTION;
return;
case kANY_GEN_BLOCK_DISTRIBUTION: (* yyNew)->ANY_GEN_BLOCK_DISTRIBUTION = yyt->ANY_GEN_BLOCK_DISTRIBUTION;
return;
case kANY_INDIRECT_DISTRIBUTION: (* yyNew)->ANY_INDIRECT_DISTRIBUTION = yyt->ANY_INDIRECT_DISTRIBUTION;
return;
case kANY_ARBITRARY_DISTRIBUTION: (* yyNew)->ANY_ARBITRARY_DISTRIBUTION = yyt->ANY_ARBITRARY_DISTRIBUTION;
return;
case kANY_DIM_DISTRIBUTION: (* yyNew)->ANY_DIM_DISTRIBUTION = yyt->ANY_DIM_DISTRIBUTION;
return;
case kPROCESSOR_SPEC: (* yyNew)->PROCESSOR_SPEC = yyt->PROCESSOR_SPEC;
return;
case kANY_PROCESSOR_ARRAY: (* yyNew)->ANY_PROCESSOR_ARRAY = yyt->ANY_PROCESSOR_ARRAY;
return;
case kDEFAULT_PROCESSOR_ARRAY: (* yyNew)->DEFAULT_PROCESSOR_ARRAY = yyt->DEFAULT_PROCESSOR_ARRAY;
return;
case kPROCESSOR_ARRAY: (* yyNew)->PROCESSOR_ARRAY = yyt->PROCESSOR_ARRAY;
yyt = yyt->PROCESSOR_ARRAY.TOPNAME;
yyNew = & (* yyNew)->PROCESSOR_ARRAY.TOPNAME; break;
case kPROCESSOR_SUBSET: (* yyNew)->PROCESSOR_SUBSET = yyt->PROCESSOR_SUBSET;
copytTree ((* yyNew)->PROCESSOR_SUBSET.SUBTOP, yyt->PROCESSOR_SUBSET.SUBTOP)
copytTree ((* yyNew)->PROCESSOR_SUBSET.FULLTOP, yyt->PROCESSOR_SUBSET.FULLTOP)
yyt = yyt->PROCESSOR_SUBSET.SUBSCRIPTS;
yyNew = & (* yyNew)->PROCESSOR_SUBSET.SUBSCRIPTS; break;
case kALIGN_SPEC: (* yyNew)->ALIGN_SPEC = yyt->ALIGN_SPEC;
copyint ((* yyNew)->ALIGN_SPEC.is_descriptive, yyt->ALIGN_SPEC.is_descriptive)
copytTree ((* yyNew)->ALIGN_SPEC.TEMPNAME, yyt->ALIGN_SPEC.TEMPNAME)
yyt = yyt->ALIGN_SPEC.SUBSCRIPTS;
yyNew = & (* yyNew)->ALIGN_SPEC.SUBSCRIPTS; break;
case kGENERIC_SPEC: (* yyNew)->GENERIC_SPEC = yyt->GENERIC_SPEC;
return;
case kNO_GENERIC_SPEC: (* yyNew)->NO_GENERIC_SPEC = yyt->NO_GENERIC_SPEC;
return;
case kID_GENERIC_SPEC: (* yyNew)->ID_GENERIC_SPEC = yyt->ID_GENERIC_SPEC;
copytIdent ((* yyNew)->ID_GENERIC_SPEC.name, yyt->ID_GENERIC_SPEC.name)
return;
case kOP_GENERIC_SPEC: (* yyNew)->OP_GENERIC_SPEC = yyt->OP_GENERIC_SPEC;
yyt = yyt->OP_GENERIC_SPEC.OPERATOR;
yyNew = & (* yyNew)->OP_GENERIC_SPEC.OPERATOR; break;
case kASSIGN_GENERIC_SPEC: (* yyNew)->ASSIGN_GENERIC_SPEC = yyt->ASSIGN_GENERIC_SPEC;
return;
case kACF_NODE_LIST: (* yyNew)->ACF_NODE_LIST = yyt->ACF_NODE_LIST;
return;
case kACF_EMPTY: (* yyNew)->ACF_EMPTY = yyt->ACF_EMPTY;
return;
case kACF_LIST: (* yyNew)->ACF_LIST = yyt->ACF_LIST;
copytTree ((* yyNew)->ACF_LIST.Elem, yyt->ACF_LIST.Elem)
yyt = yyt->ACF_LIST.Next;
yyNew = & (* yyNew)->ACF_LIST.Next; break;
case kACF_NODE: (* yyNew)->ACF_NODE = yyt->ACF_NODE;
copyint ((* yyNew)->ACF_NODE.Label, yyt->ACF_NODE.Label)
copyint ((* yyNew)->ACF_NODE.Line, yyt->ACF_NODE.Line)
return;
case kACF_DUMMY: (* yyNew)->ACF_DUMMY = yyt->ACF_DUMMY;
copyint ((* yyNew)->ACF_DUMMY.Label, yyt->ACF_DUMMY.Label)
copyint ((* yyNew)->ACF_DUMMY.Line, yyt->ACF_DUMMY.Line)
return;
case kACF_BASIC: (* yyNew)->ACF_BASIC = yyt->ACF_BASIC;
copyint ((* yyNew)->ACF_BASIC.Label, yyt->ACF_BASIC.Label)
copyint ((* yyNew)->ACF_BASIC.Line, yyt->ACF_BASIC.Line)
yyt = yyt->ACF_BASIC.BASIC_STMT;
yyNew = & (* yyNew)->ACF_BASIC.BASIC_STMT; break;
case kACF_IF: (* yyNew)->ACF_IF = yyt->ACF_IF;
copyint ((* yyNew)->ACF_IF.Label, yyt->ACF_IF.Label)
copyint ((* yyNew)->ACF_IF.Line, yyt->ACF_IF.Line)
copytTree ((* yyNew)->ACF_IF.IF_EXP, yyt->ACF_IF.IF_EXP)
copytTree ((* yyNew)->ACF_IF.THEN_PART, yyt->ACF_IF.THEN_PART)
yyt = yyt->ACF_IF.ELSE_PART;
yyNew = & (* yyNew)->ACF_IF.ELSE_PART; break;
case kACF_WHERE: (* yyNew)->ACF_WHERE = yyt->ACF_WHERE;
copyint ((* yyNew)->ACF_WHERE.Label, yyt->ACF_WHERE.Label)
copyint ((* yyNew)->ACF_WHERE.Line, yyt->ACF_WHERE.Line)
copytTree ((* yyNew)->ACF_WHERE.WHERE_EXP, yyt->ACF_WHERE.WHERE_EXP)
copytTree ((* yyNew)->ACF_WHERE.TRUE_PART, yyt->ACF_WHERE.TRUE_PART)
yyt = yyt->ACF_WHERE.FALSE_PART;
yyNew = & (* yyNew)->ACF_WHERE.FALSE_PART; break;
case kACF_CASE: (* yyNew)->ACF_CASE = yyt->ACF_CASE;
copyint ((* yyNew)->ACF_CASE.Label, yyt->ACF_CASE.Label)
copyint ((* yyNew)->ACF_CASE.Line, yyt->ACF_CASE.Line)
copytTree ((* yyNew)->ACF_CASE.CASE_EXP, yyt->ACF_CASE.CASE_EXP)
yyt = yyt->ACF_CASE.CASE_ALTS;
yyNew = & (* yyNew)->ACF_CASE.CASE_ALTS; break;
case kACF_WHILE: (* yyNew)->ACF_WHILE = yyt->ACF_WHILE;
copyint ((* yyNew)->ACF_WHILE.Label, yyt->ACF_WHILE.Label)
copyint ((* yyNew)->ACF_WHILE.Line, yyt->ACF_WHILE.Line)
copytTree ((* yyNew)->ACF_WHILE.WHILE_EXP, yyt->ACF_WHILE.WHILE_EXP)
yyt = yyt->ACF_WHILE.WHILE_BODY;
yyNew = & (* yyNew)->ACF_WHILE.WHILE_BODY; break;
case kACF_REPEAT: (* yyNew)->ACF_REPEAT = yyt->ACF_REPEAT;
copyint ((* yyNew)->ACF_REPEAT.Label, yyt->ACF_REPEAT.Label)
copyint ((* yyNew)->ACF_REPEAT.Line, yyt->ACF_REPEAT.Line)
copytTree ((* yyNew)->ACF_REPEAT.REPEAT_BODY, yyt->ACF_REPEAT.REPEAT_BODY)
yyt = yyt->ACF_REPEAT.REPEAT_EXP;
yyNew = & (* yyNew)->ACF_REPEAT.REPEAT_EXP; break;
case kACF_LOOP: (* yyNew)->ACF_LOOP = yyt->ACF_LOOP;
copyint ((* yyNew)->ACF_LOOP.Label, yyt->ACF_LOOP.Label)
copyint ((* yyNew)->ACF_LOOP.Line, yyt->ACF_LOOP.Line)
yyt = yyt->ACF_LOOP.LOOP_BODY;
yyNew = & (* yyNew)->ACF_LOOP.LOOP_BODY; break;
case kACF_DO: (* yyNew)->ACF_DO = yyt->ACF_DO;
copyint ((* yyNew)->ACF_DO.Label, yyt->ACF_DO.Label)
copyint ((* yyNew)->ACF_DO.Line, yyt->ACF_DO.Line)
copytTree ((* yyNew)->ACF_DO.DO_ID, yyt->ACF_DO.DO_ID)
copytTree ((* yyNew)->ACF_DO.DO_RANGE, yyt->ACF_DO.DO_RANGE)
copytTree ((* yyNew)->ACF_DO.DO_BODY, yyt->ACF_DO.DO_BODY)
copytTree ((* yyNew)->ACF_DO.DO_DEP_INFO, yyt->ACF_DO.DO_DEP_INFO)
yyt = yyt->ACF_DO.DO_HOME_INFO;
yyNew = & (* yyNew)->ACF_DO.DO_HOME_INFO; break;
case kACF_FORALL: (* yyNew)->ACF_FORALL = yyt->ACF_FORALL;
copyint ((* yyNew)->ACF_FORALL.Label, yyt->ACF_FORALL.Label)
copyint ((* yyNew)->ACF_FORALL.Line, yyt->ACF_FORALL.Line)
copytTree ((* yyNew)->ACF_FORALL.FORALL_ID, yyt->ACF_FORALL.FORALL_ID)
copytTree ((* yyNew)->ACF_FORALL.FORALL_RANGE, yyt->ACF_FORALL.FORALL_RANGE)
copytTree ((* yyNew)->ACF_FORALL.FORALL_BODY, yyt->ACF_FORALL.FORALL_BODY)
copytTree ((* yyNew)->ACF_FORALL.FORALL_DEP_INFO, yyt->ACF_FORALL.FORALL_DEP_INFO)
yyt = yyt->ACF_FORALL.FORALL_HOME_INFO;
yyNew = & (* yyNew)->ACF_FORALL.FORALL_HOME_INFO; break;
case kACF_FLOW_GRAPH: (* yyNew)->ACF_FLOW_GRAPH = yyt->ACF_FLOW_GRAPH;
copyint ((* yyNew)->ACF_FLOW_GRAPH.Label, yyt->ACF_FLOW_GRAPH.Label)
copyint ((* yyNew)->ACF_FLOW_GRAPH.Line, yyt->ACF_FLOW_GRAPH.Line)
yyt = yyt->ACF_FLOW_GRAPH.FLOW_NODES;
yyNew = & (* yyNew)->ACF_FLOW_GRAPH.FLOW_NODES; break;
case kACF_ENTRY: (* yyNew)->ACF_ENTRY = yyt->ACF_ENTRY;
copyint ((* yyNew)->ACF_ENTRY.Label, yyt->ACF_ENTRY.Label)
copyint ((* yyNew)->ACF_ENTRY.Line, yyt->ACF_ENTRY.Line)
yyt = yyt->ACF_ENTRY.ENTRY_DECL;
yyNew = & (* yyNew)->ACF_ENTRY.ENTRY_DECL; break;
case kACF_HOME: (* yyNew)->ACF_HOME = yyt->ACF_HOME;
copyint ((* yyNew)->ACF_HOME.Label, yyt->ACF_HOME.Label)
copyint ((* yyNew)->ACF_HOME.Line, yyt->ACF_HOME.Line)
copytTree ((* yyNew)->ACF_HOME.HOME_VAR, yyt->ACF_HOME.HOME_VAR)
copybool ((* yyNew)->ACF_HOME.user_home, yyt->ACF_HOME.user_home)
yyt = yyt->ACF_HOME.HOME_BODY;
yyNew = & (* yyNew)->ACF_HOME.HOME_BODY; break;
case kACF_RESIDENT: (* yyNew)->ACF_RESIDENT = yyt->ACF_RESIDENT;
copyint ((* yyNew)->ACF_RESIDENT.Label, yyt->ACF_RESIDENT.Label)
copyint ((* yyNew)->ACF_RESIDENT.Line, yyt->ACF_RESIDENT.Line)
copytTree ((* yyNew)->ACF_RESIDENT.RESIDENT_VAR, yyt->ACF_RESIDENT.RESIDENT_VAR)
yyt = yyt->ACF_RESIDENT.RESIDENT_BODY;
yyNew = & (* yyNew)->ACF_RESIDENT.RESIDENT_BODY; break;
case kACF_NEW: (* yyNew)->ACF_NEW = yyt->ACF_NEW;
copyint ((* yyNew)->ACF_NEW.Label, yyt->ACF_NEW.Label)
copyint ((* yyNew)->ACF_NEW.Line, yyt->ACF_NEW.Line)
copytTree ((* yyNew)->ACF_NEW.NEW_VAR, yyt->ACF_NEW.NEW_VAR)
yyt = yyt->ACF_NEW.NEW_BODY;
yyNew = & (* yyNew)->ACF_NEW.NEW_BODY; break;
case kACF_REDUCTION: (* yyNew)->ACF_REDUCTION = yyt->ACF_REDUCTION;
copyint ((* yyNew)->ACF_REDUCTION.Label, yyt->ACF_REDUCTION.Label)
copyint ((* yyNew)->ACF_REDUCTION.Line, yyt->ACF_REDUCTION.Line)
copybool ((* yyNew)->ACF_REDUCTION.is_zero, yyt->ACF_REDUCTION.is_zero)
copytTree ((* yyNew)->ACF_REDUCTION.REDUCTION_VAR, yyt->ACF_REDUCTION.REDUCTION_VAR)
copytTree ((* yyNew)->ACF_REDUCTION.REDUCTION_LVAR, yyt->ACF_REDUCTION.REDUCTION_LVAR)
copytTree ((* yyNew)->ACF_REDUCTION.REDUCTION_FUNC, yyt->ACF_REDUCTION.REDUCTION_FUNC)
yyt = yyt->ACF_REDUCTION.REDUCTION_BODY;
yyNew = & (* yyNew)->ACF_REDUCTION.REDUCTION_BODY; break;
case kACF_TASK_REGION: (* yyNew)->ACF_TASK_REGION = yyt->ACF_TASK_REGION;
copyint ((* yyNew)->ACF_TASK_REGION.Label, yyt->ACF_TASK_REGION.Label)
copyint ((* yyNew)->ACF_TASK_REGION.Line, yyt->ACF_TASK_REGION.Line)
copyint ((* yyNew)->ACF_TASK_REGION.task_flag, yyt->ACF_TASK_REGION.task_flag)
yyt = yyt->ACF_TASK_REGION.TASK_BODY;
yyNew = & (* yyNew)->ACF_TASK_REGION.TASK_BODY; break;
case kACF_PARALLEL: (* yyNew)->ACF_PARALLEL = yyt->ACF_PARALLEL;
copyint ((* yyNew)->ACF_PARALLEL.Label, yyt->ACF_PARALLEL.Label)
copyint ((* yyNew)->ACF_PARALLEL.Line, yyt->ACF_PARALLEL.Line)
copytTree ((* yyNew)->ACF_PARALLEL.REDUCTION_VARS, yyt->ACF_PARALLEL.REDUCTION_VARS)
copytTree ((* yyNew)->ACF_PARALLEL.SHARED_VARS, yyt->ACF_PARALLEL.SHARED_VARS)
copytTree ((* yyNew)->ACF_PARALLEL.LOCAL_VARS, yyt->ACF_PARALLEL.LOCAL_VARS)
yyt = yyt->ACF_PARALLEL.PARALLEL_BODY;
yyNew = & (* yyNew)->ACF_PARALLEL.PARALLEL_BODY; break;
case kACF_CRITICAL: (* yyNew)->ACF_CRITICAL = yyt->ACF_CRITICAL;
copyint ((* yyNew)->ACF_CRITICAL.Label, yyt->ACF_CRITICAL.Label)
copyint ((* yyNew)->ACF_CRITICAL.Line, yyt->ACF_CRITICAL.Line)
yyt = yyt->ACF_CRITICAL.CRITICAL_BODY;
yyNew = & (* yyNew)->ACF_CRITICAL.CRITICAL_BODY; break;
case kACF_ON: (* yyNew)->ACF_ON = yyt->ACF_ON;
copyint ((* yyNew)->ACF_ON.Label, yyt->ACF_ON.Label)
copyint ((* yyNew)->ACF_ON.Line, yyt->ACF_ON.Line)
copytTree ((* yyNew)->ACF_ON.ON_HOME, yyt->ACF_ON.ON_HOME)
copytTree ((* yyNew)->ACF_ON.ON_SPECS, yyt->ACF_ON.ON_SPECS)
yyt = yyt->ACF_ON.ON_BODY;
yyNew = & (* yyNew)->ACF_ON.ON_BODY; break;
case kACF_RM_READ: (* yyNew)->ACF_RM_READ = yyt->ACF_RM_READ;
copyint ((* yyNew)->ACF_RM_READ.Label, yyt->ACF_RM_READ.Label)
copyint ((* yyNew)->ACF_RM_READ.Line, yyt->ACF_RM_READ.Line)
copytTree ((* yyNew)->ACF_RM_READ.LOCAL_VAR, yyt->ACF_RM_READ.LOCAL_VAR)
yyt = yyt->ACF_RM_READ.REMOTE_VAR;
yyNew = & (* yyNew)->ACF_RM_READ.REMOTE_VAR; break;
case kACF_RM_WRITE: (* yyNew)->ACF_RM_WRITE = yyt->ACF_RM_WRITE;
copyint ((* yyNew)->ACF_RM_WRITE.Label, yyt->ACF_RM_WRITE.Label)
copyint ((* yyNew)->ACF_RM_WRITE.Line, yyt->ACF_RM_WRITE.Line)
copyint ((* yyNew)->ACF_RM_WRITE.op, yyt->ACF_RM_WRITE.op)
copytTree ((* yyNew)->ACF_RM_WRITE.REMOTE_VAR, yyt->ACF_RM_WRITE.REMOTE_VAR)
yyt = yyt->ACF_RM_WRITE.LOCAL_DATA;
yyNew = & (* yyNew)->ACF_RM_WRITE.LOCAL_DATA; break;
case kACF_MOVE: (* yyNew)->ACF_MOVE = yyt->ACF_MOVE;
copyint ((* yyNew)->ACF_MOVE.Label, yyt->ACF_MOVE.Label)
copyint ((* yyNew)->ACF_MOVE.Line, yyt->ACF_MOVE.Line)
yyt = yyt->ACF_MOVE.MOVE_STMT;
yyNew = & (* yyNew)->ACF_MOVE.MOVE_STMT; break;
case kACF_BARRIER: (* yyNew)->ACF_BARRIER = yyt->ACF_BARRIER;
copyint ((* yyNew)->ACF_BARRIER.Label, yyt->ACF_BARRIER.Label)
copyint ((* yyNew)->ACF_BARRIER.Line, yyt->ACF_BARRIER.Line)
yyt = yyt->ACF_BARRIER.BARRIER_VARS;
yyNew = & (* yyNew)->ACF_BARRIER.BARRIER_VARS; break;
case kACF_UPDATE: (* yyNew)->ACF_UPDATE = yyt->ACF_UPDATE;
copyint ((* yyNew)->ACF_UPDATE.Label, yyt->ACF_UPDATE.Label)
copyint ((* yyNew)->ACF_UPDATE.Line, yyt->ACF_UPDATE.Line)
yyt = yyt->ACF_UPDATE.OVERLAP_VAR;
yyNew = & (* yyNew)->ACF_UPDATE.OVERLAP_VAR; break;
case kACF_FLOW_KIND: (* yyNew)->ACF_FLOW_KIND = yyt->ACF_FLOW_KIND;
return;
case kACF_COMPLEX_FLOW: (* yyNew)->ACF_COMPLEX_FLOW = yyt->ACF_COMPLEX_FLOW;
return;
case kACF_INTERVAL_FLOW: (* yyNew)->ACF_INTERVAL_FLOW = yyt->ACF_INTERVAL_FLOW;
return;
case kACF_SIMPLE_FLOW: (* yyNew)->ACF_SIMPLE_FLOW = yyt->ACF_SIMPLE_FLOW;
return;
case kON_CLAUSE: (* yyNew)->ON_CLAUSE = yyt->ON_CLAUSE;
return;
case kON_HOST_CLAUSE: (* yyNew)->ON_HOST_CLAUSE = yyt->ON_HOST_CLAUSE;
return;
case kON_ALL_CLAUSE: (* yyNew)->ON_ALL_CLAUSE = yyt->ON_ALL_CLAUSE;
return;
case kON_VAR_CLAUSE: (* yyNew)->ON_VAR_CLAUSE = yyt->ON_VAR_CLAUSE;
yyt = yyt->ON_VAR_CLAUSE.ON_VAR;
yyNew = & (* yyNew)->ON_VAR_CLAUSE.ON_VAR; break;
case kON_PROC_CLAUSE: (* yyNew)->ON_PROC_CLAUSE = yyt->ON_PROC_CLAUSE;
yyt = yyt->ON_PROC_CLAUSE.ON_PROC;
yyNew = & (* yyNew)->ON_PROC_CLAUSE.ON_PROC; break;
case kON_TOPOLOGY: (* yyNew)->ON_TOPOLOGY = yyt->ON_TOPOLOGY;
return;
case kON_ALL: (* yyNew)->ON_ALL = yyt->ON_ALL;
return;
case kON_HOST: (* yyNew)->ON_HOST = yyt->ON_HOST;
return;
case kON_PROCS: (* yyNew)->ON_PROCS = yyt->ON_PROCS;
copytObject ((* yyNew)->ON_PROCS.topid, yyt->ON_PROCS.topid)
return;
case kON_VAR: (* yyNew)->ON_VAR = yyt->ON_VAR;
copytObject ((* yyNew)->ON_VAR.topid, yyt->ON_VAR.topid)
copytObject ((* yyNew)->ON_VAR.tempid, yyt->ON_VAR.tempid)
copytObject ((* yyNew)->ON_VAR.arrayid, yyt->ON_VAR.arrayid)
return;
case kON_SPEC_LIST: (* yyNew)->ON_SPEC_LIST = yyt->ON_SPEC_LIST;
return;
case kON_EMPTY: (* yyNew)->ON_EMPTY = yyt->ON_EMPTY;
return;
case kON_LIST: (* yyNew)->ON_LIST = yyt->ON_LIST;
copytTree ((* yyNew)->ON_LIST.Elem, yyt->ON_LIST.Elem)
yyt = yyt->ON_LIST.Next;
yyNew = & (* yyNew)->ON_LIST.Next; break;
case kON_SPEC: (* yyNew)->ON_SPEC = yyt->ON_SPEC;
copyint ((* yyNew)->ON_SPEC.top_dim, yyt->ON_SPEC.top_dim)
copyint ((* yyNew)->ON_SPEC.temp_dim, yyt->ON_SPEC.temp_dim)
copyint ((* yyNew)->ON_SPEC.array_dim, yyt->ON_SPEC.array_dim)
yyt = yyt->ON_SPEC.ON_VAL;
yyNew = & (* yyNew)->ON_SPEC.ON_VAL; break;
case kSELECTED_ACF_NODE_LIST: (* yyNew)->SELECTED_ACF_NODE_LIST = yyt->SELECTED_ACF_NODE_LIST;
return;
case kSELECTED_ACF_EMPTY: (* yyNew)->SELECTED_ACF_EMPTY = yyt->SELECTED_ACF_EMPTY;
return;
case kSELECTED_ACF_LIST: (* yyNew)->SELECTED_ACF_LIST = yyt->SELECTED_ACF_LIST;
copytTree ((* yyNew)->SELECTED_ACF_LIST.Elem, yyt->SELECTED_ACF_LIST.Elem)
yyt = yyt->SELECTED_ACF_LIST.Next;
yyNew = & (* yyNew)->SELECTED_ACF_LIST.Next; break;
case kSELECTED_ACF_NODE: (* yyNew)->SELECTED_ACF_NODE = yyt->SELECTED_ACF_NODE;
copytTree ((* yyNew)->SELECTED_ACF_NODE.SELECT_LIST, yyt->SELECTED_ACF_NODE.SELECT_LIST)
yyt = yyt->SELECTED_ACF_NODE.SELECT_ACFS;
yyNew = & (* yyNew)->SELECTED_ACF_NODE.SELECT_ACFS; break;
case kBT_STMT: (* yyNew)->BT_STMT = yyt->BT_STMT;
return;
case kASSIGN_STMT: (* yyNew)->ASSIGN_STMT = yyt->ASSIGN_STMT;
copytTree ((* yyNew)->ASSIGN_STMT.ASSIGN_VAR, yyt->ASSIGN_STMT.ASSIGN_VAR)
yyt = yyt->ASSIGN_STMT.ASSIGN_EXP;
yyNew = & (* yyNew)->ASSIGN_STMT.ASSIGN_EXP; break;
case kPTR_ASSIGN_STMT: (* yyNew)->PTR_ASSIGN_STMT = yyt->PTR_ASSIGN_STMT;
copytTree ((* yyNew)->PTR_ASSIGN_STMT.ASSIGN_VAR, yyt->PTR_ASSIGN_STMT.ASSIGN_VAR)
yyt = yyt->PTR_ASSIGN_STMT.ASSIGN_EXP;
yyNew = & (* yyNew)->PTR_ASSIGN_STMT.ASSIGN_EXP; break;
case kLABEL_ASSIGN_STMT: (* yyNew)->LABEL_ASSIGN_STMT = yyt->LABEL_ASSIGN_STMT;
copyint ((* yyNew)->LABEL_ASSIGN_STMT.assign_label, yyt->LABEL_ASSIGN_STMT.assign_label)
yyt = yyt->LABEL_ASSIGN_STMT.LABEL_VAR;
yyNew = & (* yyNew)->LABEL_ASSIGN_STMT.LABEL_VAR; break;
case kCALL_STMT: (* yyNew)->CALL_STMT = yyt->CALL_STMT;
copytTree ((* yyNew)->CALL_STMT.CALL_ID, yyt->CALL_STMT.CALL_ID)
yyt = yyt->CALL_STMT.CALL_PARAMS;
yyNew = & (* yyNew)->CALL_STMT.CALL_PARAMS; break;
case kIO_STMT: (* yyNew)->IO_STMT = yyt->IO_STMT;
copytTree ((* yyNew)->IO_STMT.ID, yyt->IO_STMT.ID)
copytTree ((* yyNew)->IO_STMT.IO_SPECS, yyt->IO_STMT.IO_SPECS)
yyt = yyt->IO_STMT.IO_ITEMS;
yyNew = & (* yyNew)->IO_STMT.IO_ITEMS; break;
case kGOTO_STMT: (* yyNew)->GOTO_STMT = yyt->GOTO_STMT;
copyint ((* yyNew)->GOTO_STMT.GOTO_LABEL, yyt->GOTO_STMT.GOTO_LABEL)
return;
case kASS_GOTO_STMT: (* yyNew)->ASS_GOTO_STMT = yyt->ASS_GOTO_STMT;
copytTree ((* yyNew)->ASS_GOTO_STMT.GOTO_VAR, yyt->ASS_GOTO_STMT.GOTO_VAR)
yyt = yyt->ASS_GOTO_STMT.LABELS;
yyNew = & (* yyNew)->ASS_GOTO_STMT.LABELS; break;
case kCOMP_GOTO_STMT: (* yyNew)->COMP_GOTO_STMT = yyt->COMP_GOTO_STMT;
copytTree ((* yyNew)->COMP_GOTO_STMT.GOTO_LABELS, yyt->COMP_GOTO_STMT.GOTO_LABELS)
yyt = yyt->COMP_GOTO_STMT.GOTO_EXP;
yyNew = & (* yyNew)->COMP_GOTO_STMT.GOTO_EXP; break;
case kCOMP_IF_STMT: (* yyNew)->COMP_IF_STMT = yyt->COMP_IF_STMT;
copyint ((* yyNew)->COMP_IF_STMT.IF_LT_LABEL, yyt->COMP_IF_STMT.IF_LT_LABEL)
copyint ((* yyNew)->COMP_IF_STMT.IF_EQ_LABEL, yyt->COMP_IF_STMT.IF_EQ_LABEL)
copyint ((* yyNew)->COMP_IF_STMT.IF_GT_LABEL, yyt->COMP_IF_STMT.IF_GT_LABEL)
yyt = yyt->COMP_IF_STMT.IF_EXP;
yyNew = & (* yyNew)->COMP_IF_STMT.IF_EXP; break;
case kRETURN_STMT: (* yyNew)->RETURN_STMT = yyt->RETURN_STMT;
yyt = yyt->RETURN_STMT.RETURN_EXP;
yyNew = & (* yyNew)->RETURN_STMT.RETURN_EXP; break;
case kPAUSE_STMT: (* yyNew)->PAUSE_STMT = yyt->PAUSE_STMT;
yyt = yyt->PAUSE_STMT.PAUSE_CONST;
yyNew = & (* yyNew)->PAUSE_STMT.PAUSE_CONST; break;
case kEXIT_STMT: (* yyNew)->EXIT_STMT = yyt->EXIT_STMT;
copytIdent ((* yyNew)->EXIT_STMT.loopid, yyt->EXIT_STMT.loopid)
return;
case kCYCLE_STMT: (* yyNew)->CYCLE_STMT = yyt->CYCLE_STMT;
copytIdent ((* yyNew)->CYCLE_STMT.loopid, yyt->CYCLE_STMT.loopid)
return;
case kSTOP_STMT: (* yyNew)->STOP_STMT = yyt->STOP_STMT;
yyt = yyt->STOP_STMT.STOP_CONST;
yyNew = & (* yyNew)->STOP_STMT.STOP_CONST; break;
case kFORMAT_STMT: (* yyNew)->FORMAT_STMT = yyt->FORMAT_STMT;
yyt = yyt->FORMAT_STMT.FORMAT_PARAMS;
yyNew = & (* yyNew)->FORMAT_STMT.FORMAT_PARAMS; break;
case kINHERIT_DSP_STMT: (* yyNew)->INHERIT_DSP_STMT = yyt->INHERIT_DSP_STMT;
copytTree ((* yyNew)->INHERIT_DSP_STMT.VAR, yyt->INHERIT_DSP_STMT.VAR)
copyint ((* yyNew)->INHERIT_DSP_STMT.flag, yyt->INHERIT_DSP_STMT.flag)
copyint ((* yyNew)->INHERIT_DSP_STMT.safety, yyt->INHERIT_DSP_STMT.safety)
yyt = yyt->INHERIT_DSP_STMT.SHAPE;
yyNew = & (* yyNew)->INHERIT_DSP_STMT.SHAPE; break;
case kCREATE_DSP_STMT: (* yyNew)->CREATE_DSP_STMT = yyt->CREATE_DSP_STMT;
copytTree ((* yyNew)->CREATE_DSP_STMT.VAR, yyt->CREATE_DSP_STMT.VAR)
copyint ((* yyNew)->CREATE_DSP_STMT.flag, yyt->CREATE_DSP_STMT.flag)
yyt = yyt->CREATE_DSP_STMT.SHAPE;
yyNew = & (* yyNew)->CREATE_DSP_STMT.SHAPE; break;
case kFREE_DSP_STMT: (* yyNew)->FREE_DSP_STMT = yyt->FREE_DSP_STMT;
copyint ((* yyNew)->FREE_DSP_STMT.flag, yyt->FREE_DSP_STMT.flag)
yyt = yyt->FREE_DSP_STMT.VAR;
yyNew = & (* yyNew)->FREE_DSP_STMT.VAR; break;
case kALLOCATE_STMT: (* yyNew)->ALLOCATE_STMT = yyt->ALLOCATE_STMT;
copytTree ((* yyNew)->ALLOCATE_STMT.PARAMS, yyt->ALLOCATE_STMT.PARAMS)
yyt = yyt->ALLOCATE_STMT.STATUS;
yyNew = & (* yyNew)->ALLOCATE_STMT.STATUS; break;
case kDEALLOCATE_STMT: (* yyNew)->DEALLOCATE_STMT = yyt->DEALLOCATE_STMT;
copytTree ((* yyNew)->DEALLOCATE_STMT.PARAMS, yyt->DEALLOCATE_STMT.PARAMS)
yyt = yyt->DEALLOCATE_STMT.STATUS;
yyNew = & (* yyNew)->DEALLOCATE_STMT.STATUS; break;
case kNULLIFY_STMT: (* yyNew)->NULLIFY_STMT = yyt->NULLIFY_STMT;
yyt = yyt->NULLIFY_STMT.PARAMS;
yyNew = & (* yyNew)->NULLIFY_STMT.PARAMS; break;
case kREDUCE_STMT: (* yyNew)->REDUCE_STMT = yyt->REDUCE_STMT;
copytTree ((* yyNew)->REDUCE_STMT.RED_FUNC, yyt->REDUCE_STMT.RED_FUNC)
yyt = yyt->REDUCE_STMT.RED_PARAMS;
yyNew = & (* yyNew)->REDUCE_STMT.RED_PARAMS; break;
case kMOVE_STMT: (* yyNew)->MOVE_STMT = yyt->MOVE_STMT;
copytTree ((* yyNew)->MOVE_STMT.TARGET, yyt->MOVE_STMT.TARGET)
copytTree ((* yyNew)->MOVE_STMT.SOURCE, yyt->MOVE_STMT.SOURCE)
yyt = yyt->MOVE_STMT.MASK;
yyNew = & (* yyNew)->MOVE_STMT.MASK; break;
case kSCATTER_STMT: (* yyNew)->SCATTER_STMT = yyt->SCATTER_STMT;
copytIdent ((* yyNew)->SCATTER_STMT.func, yyt->SCATTER_STMT.func)
copytTree ((* yyNew)->SCATTER_STMT.ARRAY, yyt->SCATTER_STMT.ARRAY)
copytTree ((* yyNew)->SCATTER_STMT.BASE, yyt->SCATTER_STMT.BASE)
copytTree ((* yyNew)->SCATTER_STMT.INDEXES, yyt->SCATTER_STMT.INDEXES)
yyt = yyt->SCATTER_STMT.MASK;
yyNew = & (* yyNew)->SCATTER_STMT.MASK; break;
case kREALIGN_STMT: (* yyNew)->REALIGN_STMT = yyt->REALIGN_STMT;
copytTree ((* yyNew)->REALIGN_STMT.ALIGNEE, yyt->REALIGN_STMT.ALIGNEE)
copytTree ((* yyNew)->REALIGN_STMT.ALIGN_SOURCE, yyt->REALIGN_STMT.ALIGN_SOURCE)
copytDefinitions ((* yyNew)->REALIGN_STMT.distribution, yyt->REALIGN_STMT.distribution)
yyt = yyt->REALIGN_STMT.ALIGN_TARGET;
yyNew = & (* yyNew)->REALIGN_STMT.ALIGN_TARGET; break;
case kREDISTRIBUTE_STMT: (* yyNew)->REDISTRIBUTE_STMT = yyt->REDISTRIBUTE_STMT;
copytTree ((* yyNew)->REDISTRIBUTE_STMT.DISTRIBUTEE, yyt->REDISTRIBUTE_STMT.DISTRIBUTEE)
copytTree ((* yyNew)->REDISTRIBUTE_STMT.DISTRIBUTION, yyt->REDISTRIBUTE_STMT.DISTRIBUTION)
copytDefinitions ((* yyNew)->REDISTRIBUTE_STMT.distribution, yyt->REDISTRIBUTE_STMT.distribution)
yyt = yyt->REDISTRIBUTE_STMT.TARGET;
yyNew = & (* yyNew)->REDISTRIBUTE_STMT.TARGET; break;
case kREG_SHADOW_GET: (* yyNew)->REG_SHADOW_GET = yyt->REG_SHADOW_GET;
copytTree ((* yyNew)->REG_SHADOW_GET.SHADOW_VAR, yyt->REG_SHADOW_GET.SHADOW_VAR)
yyt = yyt->REG_SHADOW_GET.SHADOW_SIZE;
yyNew = & (* yyNew)->REG_SHADOW_GET.SHADOW_SIZE; break;
case kREG_SHADOW_PUT: (* yyNew)->REG_SHADOW_PUT = yyt->REG_SHADOW_PUT;
copyint ((* yyNew)->REG_SHADOW_PUT.op, yyt->REG_SHADOW_PUT.op)
copytTree ((* yyNew)->REG_SHADOW_PUT.SHADOW_VAR, yyt->REG_SHADOW_PUT.SHADOW_VAR)
yyt = yyt->REG_SHADOW_PUT.SHADOW_SIZE;
yyNew = & (* yyNew)->REG_SHADOW_PUT.SHADOW_SIZE; break;
case kREG_SHADOW_SET: (* yyNew)->REG_SHADOW_SET = yyt->REG_SHADOW_SET;
copytTree ((* yyNew)->REG_SHADOW_SET.SHADOW_VAR, yyt->REG_SHADOW_SET.SHADOW_VAR)
copytTree ((* yyNew)->REG_SHADOW_SET.SHADOW_SIZE, yyt->REG_SHADOW_SET.SHADOW_SIZE)
yyt = yyt->REG_SHADOW_SET.VAL;
yyNew = & (* yyNew)->REG_SHADOW_SET.VAL; break;
case kIND_SHADOW_CREATE: (* yyNew)->IND_SHADOW_CREATE = yyt->IND_SHADOW_CREATE;
copyint ((* yyNew)->IND_SHADOW_CREATE.id, yyt->IND_SHADOW_CREATE.id)
copytTree ((* yyNew)->IND_SHADOW_CREATE.SHADOW_VAR, yyt->IND_SHADOW_CREATE.SHADOW_VAR)
copytTree ((* yyNew)->IND_SHADOW_CREATE.INDEX_VAR, yyt->IND_SHADOW_CREATE.INDEX_VAR)
copytTree ((* yyNew)->IND_SHADOW_CREATE.TMP_INDEX_VAR, yyt->IND_SHADOW_CREATE.TMP_INDEX_VAR)
yyt = yyt->IND_SHADOW_CREATE.MASK;
yyNew = & (* yyNew)->IND_SHADOW_CREATE.MASK; break;
case kIND_SHADOW_FREE: (* yyNew)->IND_SHADOW_FREE = yyt->IND_SHADOW_FREE;
copyint ((* yyNew)->IND_SHADOW_FREE.id, yyt->IND_SHADOW_FREE.id)
yyt = yyt->IND_SHADOW_FREE.INDEX_VAR;
yyNew = & (* yyNew)->IND_SHADOW_FREE.INDEX_VAR; break;
case kIND_SHADOW_GET: (* yyNew)->IND_SHADOW_GET = yyt->IND_SHADOW_GET;
copyint ((* yyNew)->IND_SHADOW_GET.id, yyt->IND_SHADOW_GET.id)
yyt = yyt->IND_SHADOW_GET.SHADOW_VAR;
yyNew = & (* yyNew)->IND_SHADOW_GET.SHADOW_VAR; break;
case kIND_SHADOW_PUT: (* yyNew)->IND_SHADOW_PUT = yyt->IND_SHADOW_PUT;
copyint ((* yyNew)->IND_SHADOW_PUT.id, yyt->IND_SHADOW_PUT.id)
copyint ((* yyNew)->IND_SHADOW_PUT.op, yyt->IND_SHADOW_PUT.op)
yyt = yyt->IND_SHADOW_PUT.SHADOW_VAR;
yyNew = & (* yyNew)->IND_SHADOW_PUT.SHADOW_VAR; break;
case kIND_SHADOW_SET: (* yyNew)->IND_SHADOW_SET = yyt->IND_SHADOW_SET;
copyint ((* yyNew)->IND_SHADOW_SET.id, yyt->IND_SHADOW_SET.id)
copytTree ((* yyNew)->IND_SHADOW_SET.SHADOW_VAR, yyt->IND_SHADOW_SET.SHADOW_VAR)
yyt = yyt->IND_SHADOW_SET.VAL;
yyNew = & (* yyNew)->IND_SHADOW_SET.VAL; break;
case kBROADCAST_STMT: (* yyNew)->BROADCAST_STMT = yyt->BROADCAST_STMT;
copytTree ((* yyNew)->BROADCAST_STMT.ELEMS, yyt->BROADCAST_STMT.ELEMS)
copytTree ((* yyNew)->BROADCAST_STMT.OWNER_TOP, yyt->BROADCAST_STMT.OWNER_TOP)
yyt = yyt->BROADCAST_STMT.OWNER_SPEC;
yyNew = & (* yyNew)->BROADCAST_STMT.OWNER_SPEC; break;
case kREDUCTION_STMT: (* yyNew)->REDUCTION_STMT = yyt->REDUCTION_STMT;
copytTree ((* yyNew)->REDUCTION_STMT.REDFUNC, yyt->REDUCTION_STMT.REDFUNC)
copytTree ((* yyNew)->REDUCTION_STMT.ELEMS, yyt->REDUCTION_STMT.ELEMS)
copytTree ((* yyNew)->REDUCTION_STMT.RED_TOP, yyt->REDUCTION_STMT.RED_TOP)
yyt = yyt->REDUCTION_STMT.RED_SPEC;
yyNew = & (* yyNew)->REDUCTION_STMT.RED_SPEC; break;
case kDIRTY_STMT: (* yyNew)->DIRTY_STMT = yyt->DIRTY_STMT;
yyt = yyt->DIRTY_STMT.ELEMS;
yyNew = & (* yyNew)->DIRTY_STMT.ELEMS; break;
case kLABEL_USE_LIST: (* yyNew)->LABEL_USE_LIST = yyt->LABEL_USE_LIST;
return;
case kLABEL_EMPTY: (* yyNew)->LABEL_EMPTY = yyt->LABEL_EMPTY;
return;
case kLABEL_LIST: (* yyNew)->LABEL_LIST = yyt->LABEL_LIST;
copyint ((* yyNew)->LABEL_LIST.Elem, yyt->LABEL_LIST.Elem)
yyt = yyt->LABEL_LIST.Next;
yyNew = & (* yyNew)->LABEL_LIST.Next; break;
case kOP: (* yyNew)->OP = yyt->OP;
return;
case kOP_EQ: (* yyNew)->OP_EQ = yyt->OP_EQ;
return;
case kOP_NE: (* yyNew)->OP_NE = yyt->OP_NE;
return;
case kOP_LT: (* yyNew)->OP_LT = yyt->OP_LT;
return;
case kOP_GT: (* yyNew)->OP_GT = yyt->OP_GT;
return;
case kOP_GE: (* yyNew)->OP_GE = yyt->OP_GE;
return;
case kOP_LE: (* yyNew)->OP_LE = yyt->OP_LE;
return;
case kOP_PLUS: (* yyNew)->OP_PLUS = yyt->OP_PLUS;
return;
case kOP_MINUS: (* yyNew)->OP_MINUS = yyt->OP_MINUS;
return;
case kOP_XOR: (* yyNew)->OP_XOR = yyt->OP_XOR;
return;
case kOP_OR: (* yyNew)->OP_OR = yyt->OP_OR;
return;
case kOP_CONCAT: (* yyNew)->OP_CONCAT = yyt->OP_CONCAT;
return;
case kOP_TIMES: (* yyNew)->OP_TIMES = yyt->OP_TIMES;
return;
case kOP_DIVIDE: (* yyNew)->OP_DIVIDE = yyt->OP_DIVIDE;
return;
case kOP_AND: (* yyNew)->OP_AND = yyt->OP_AND;
return;
case kOP_EQV: (* yyNew)->OP_EQV = yyt->OP_EQV;
return;
case kOP_NEQV: (* yyNew)->OP_NEQV = yyt->OP_NEQV;
return;
case kOP_EXPO: (* yyNew)->OP_EXPO = yyt->OP_EXPO;
return;
case kOP_DEFINED: (* yyNew)->OP_DEFINED = yyt->OP_DEFINED;
copytIdent ((* yyNew)->OP_DEFINED.opname, yyt->OP_DEFINED.opname)
return;
case kOP_ASSIGN: (* yyNew)->OP_ASSIGN = yyt->OP_ASSIGN;
return;
case kOP1_NOT: (* yyNew)->OP1_NOT = yyt->OP1_NOT;
return;
case kOP1_SIGN: (* yyNew)->OP1_SIGN = yyt->OP1_SIGN;
return;
case kOP1_PSIGN: (* yyNew)->OP1_PSIGN = yyt->OP1_PSIGN;
return;
case kBT_EXP: (* yyNew)->BT_EXP = yyt->BT_EXP;
return;
case kDUMMY_EXP: (* yyNew)->DUMMY_EXP = yyt->DUMMY_EXP;
return;
case kCONST_EXP: (* yyNew)->CONST_EXP = yyt->CONST_EXP;
yyt = yyt->CONST_EXP.C;
yyNew = & (* yyNew)->CONST_EXP.C; break;
case kOP_EXP: (* yyNew)->OP_EXP = yyt->OP_EXP;
copytTree ((* yyNew)->OP_EXP.EXP_OP, yyt->OP_EXP.EXP_OP)
copytTree ((* yyNew)->OP_EXP.OPND1, yyt->OP_EXP.OPND1)
yyt = yyt->OP_EXP.OPND2;
yyNew = & (* yyNew)->OP_EXP.OPND2; break;
case kOP1_EXP: (* yyNew)->OP1_EXP = yyt->OP1_EXP;
copytTree ((* yyNew)->OP1_EXP.EXP_OP1, yyt->OP1_EXP.EXP_OP1)
yyt = yyt->OP1_EXP.OPND;
yyNew = & (* yyNew)->OP1_EXP.OPND; break;
case kSLICE_EXP: (* yyNew)->SLICE_EXP = yyt->SLICE_EXP;
copytTree ((* yyNew)->SLICE_EXP.START, yyt->SLICE_EXP.START)
copytTree ((* yyNew)->SLICE_EXP.STOP, yyt->SLICE_EXP.STOP)
yyt = yyt->SLICE_EXP.INC;
yyNew = & (* yyNew)->SLICE_EXP.INC; break;
case kVAR_EXP: (* yyNew)->VAR_EXP = yyt->VAR_EXP;
yyt = yyt->VAR_EXP.V;
yyNew = & (* yyNew)->VAR_EXP.V; break;
case kFUNC_CALL_EXP: (* yyNew)->FUNC_CALL_EXP = yyt->FUNC_CALL_EXP;
copytTree ((* yyNew)->FUNC_CALL_EXP.FUNC_ID, yyt->FUNC_CALL_EXP.FUNC_ID)
yyt = yyt->FUNC_CALL_EXP.FUNC_PARAMS;
yyNew = & (* yyNew)->FUNC_CALL_EXP.FUNC_PARAMS; break;
case kNAMED_EXP: (* yyNew)->NAMED_EXP = yyt->NAMED_EXP;
copytIdent ((* yyNew)->NAMED_EXP.Name, yyt->NAMED_EXP.Name)
yyt = yyt->NAMED_EXP.VAL;
yyNew = & (* yyNew)->NAMED_EXP.VAL; break;
case kDO_EXP: (* yyNew)->DO_EXP = yyt->DO_EXP;
copytTree ((* yyNew)->DO_EXP.DO_ID, yyt->DO_EXP.DO_ID)
copytTree ((* yyNew)->DO_EXP.RANGE, yyt->DO_EXP.RANGE)
yyt = yyt->DO_EXP.BODY;
yyNew = & (* yyNew)->DO_EXP.BODY; break;
case kTYPE_EXP: (* yyNew)->TYPE_EXP = yyt->TYPE_EXP;
copytTree ((* yyNew)->TYPE_EXP.ID, yyt->TYPE_EXP.ID)
yyt = yyt->TYPE_EXP.ELEMENTS;
yyNew = & (* yyNew)->TYPE_EXP.ELEMENTS; break;
case kARRAY_EXP: (* yyNew)->ARRAY_EXP = yyt->ARRAY_EXP;
yyt = yyt->ARRAY_EXP.ELEMENTS;
yyNew = & (* yyNew)->ARRAY_EXP.ELEMENTS; break;
case kPERM_EXP: (* yyNew)->PERM_EXP = yyt->PERM_EXP;
copyPermutation ((* yyNew)->PERM_EXP.perm, yyt->PERM_EXP.perm)
yyt = yyt->PERM_EXP.VAL;
yyNew = & (* yyNew)->PERM_EXP.VAL; break;
case kBOUND_EXP: (* yyNew)->BOUND_EXP = yyt->BOUND_EXP;
copyint ((* yyNew)->BOUND_EXP.dim, yyt->BOUND_EXP.dim)
copyint ((* yyNew)->BOUND_EXP.kind, yyt->BOUND_EXP.kind)
copyint ((* yyNew)->BOUND_EXP.local, yyt->BOUND_EXP.local)
yyt = yyt->BOUND_EXP.VAR;
yyNew = & (* yyNew)->BOUND_EXP.VAR; break;
case kRANK_EXP: (* yyNew)->RANK_EXP = yyt->RANK_EXP;
copyint ((* yyNew)->RANK_EXP.kind, yyt->RANK_EXP.kind)
return;
case kBT_EXP_LIST: (* yyNew)->BT_EXP_LIST = yyt->BT_EXP_LIST;
return;
case kBTE_EMPTY: (* yyNew)->BTE_EMPTY = yyt->BTE_EMPTY;
return;
case kBTE_LIST: (* yyNew)->BTE_LIST = yyt->BTE_LIST;
copytTree ((* yyNew)->BTE_LIST.Elem, yyt->BTE_LIST.Elem)
yyt = yyt->BTE_LIST.Next;
yyNew = & (* yyNew)->BTE_LIST.Next; break;
case kBT_VAR: (* yyNew)->BT_VAR = yyt->BT_VAR;
return;
case kDUMMY_VAR: (* yyNew)->DUMMY_VAR = yyt->DUMMY_VAR;
return;
case kUSED_VAR: (* yyNew)->USED_VAR = yyt->USED_VAR;
yyt = yyt->USED_VAR.VARNAME;
yyNew = & (* yyNew)->USED_VAR.VARNAME; break;
case kINDEXED_VAR: (* yyNew)->INDEXED_VAR = yyt->INDEXED_VAR;
copytTree ((* yyNew)->INDEXED_VAR.IND_VAR, yyt->INDEXED_VAR.IND_VAR)
yyt = yyt->INDEXED_VAR.IND_EXPS;
yyNew = & (* yyNew)->INDEXED_VAR.IND_EXPS; break;
case kSUBSTRING_VAR: (* yyNew)->SUBSTRING_VAR = yyt->SUBSTRING_VAR;
copytTree ((* yyNew)->SUBSTRING_VAR.IND_VAR, yyt->SUBSTRING_VAR.IND_VAR)
yyt = yyt->SUBSTRING_VAR.IND_EXP;
yyNew = & (* yyNew)->SUBSTRING_VAR.IND_EXP; break;
case kSELECTED_VAR: (* yyNew)->SELECTED_VAR = yyt->SELECTED_VAR;
copytTree ((* yyNew)->SELECTED_VAR.SELEC_VAR, yyt->SELECTED_VAR.SELEC_VAR)
yyt = yyt->SELECTED_VAR.SELECTOR;
yyNew = & (* yyNew)->SELECTED_VAR.SELECTOR; break;
case kLOOP_VAR: (* yyNew)->LOOP_VAR = yyt->LOOP_VAR;
yyt = yyt->LOOP_VAR.LOOP_VARNAME;
yyNew = & (* yyNew)->LOOP_VAR.LOOP_VARNAME; break;
case kPERM_VAR: (* yyNew)->PERM_VAR = yyt->PERM_VAR;
copyPermutation ((* yyNew)->PERM_VAR.perm, yyt->PERM_VAR.perm)
yyt = yyt->PERM_VAR.VAL;
yyNew = & (* yyNew)->PERM_VAR.VAL; break;
case kADDR: (* yyNew)->ADDR = yyt->ADDR;
yyt = yyt->ADDR.E;
yyNew = & (* yyNew)->ADDR.E; break;
case kDO_VAR: (* yyNew)->DO_VAR = yyt->DO_VAR;
copytTree ((* yyNew)->DO_VAR.DO_ID, yyt->DO_VAR.DO_ID)
copytTree ((* yyNew)->DO_VAR.RANGE, yyt->DO_VAR.RANGE)
yyt = yyt->DO_VAR.BODY;
yyNew = & (* yyNew)->DO_VAR.BODY; break;
case kREMOTE_VAR: (* yyNew)->REMOTE_VAR = yyt->REMOTE_VAR;
yyt = yyt->REMOTE_VAR.VAR;
yyNew = & (* yyNew)->REMOTE_VAR.VAR; break;
case kBT_VAR_LIST: (* yyNew)->BT_VAR_LIST = yyt->BT_VAR_LIST;
return;
case kBTV_EMPTY: (* yyNew)->BTV_EMPTY = yyt->BTV_EMPTY;
return;
case kBTV_LIST: (* yyNew)->BTV_LIST = yyt->BTV_LIST;
copytTree ((* yyNew)->BTV_LIST.Elem, yyt->BTV_LIST.Elem)
yyt = yyt->BTV_LIST.Next;
yyNew = & (* yyNew)->BTV_LIST.Next; break;
case kBT_PARAM: (* yyNew)->BT_PARAM = yyt->BT_PARAM;
return;
case kVAR_PARAM: (* yyNew)->VAR_PARAM = yyt->VAR_PARAM;
copytDefinitions ((* yyNew)->VAR_PARAM.formal, yyt->VAR_PARAM.formal)
copyint ((* yyNew)->VAR_PARAM.intent, yyt->VAR_PARAM.intent)
copyint ((* yyNew)->VAR_PARAM.layout, yyt->VAR_PARAM.layout)
copyint ((* yyNew)->VAR_PARAM.pass_by, yyt->VAR_PARAM.pass_by)
copyint ((* yyNew)->VAR_PARAM.trace, yyt->VAR_PARAM.trace)
yyt = yyt->VAR_PARAM.V;
yyNew = & (* yyNew)->VAR_PARAM.V; break;
case kVALUE_PARAM: (* yyNew)->VALUE_PARAM = yyt->VALUE_PARAM;
yyt = yyt->VALUE_PARAM.E;
yyNew = & (* yyNew)->VALUE_PARAM.E; break;
case kNAMED_PARAM: (* yyNew)->NAMED_PARAM = yyt->NAMED_PARAM;
copytIdent ((* yyNew)->NAMED_PARAM.Name, yyt->NAMED_PARAM.Name)
yyt = yyt->NAMED_PARAM.VAL;
yyNew = & (* yyNew)->NAMED_PARAM.VAL; break;
case kPROC_PARAM: (* yyNew)->PROC_PARAM = yyt->PROC_PARAM;
yyt = yyt->PROC_PARAM.P;
yyNew = & (* yyNew)->PROC_PARAM.P; break;
case kFUNC_PARAM: (* yyNew)->FUNC_PARAM = yyt->FUNC_PARAM;
yyt = yyt->FUNC_PARAM.F;
yyNew = & (* yyNew)->FUNC_PARAM.F; break;
case kFORMAT_PARAM: (* yyNew)->FORMAT_PARAM = yyt->FORMAT_PARAM;
copytStringRef ((* yyNew)->FORMAT_PARAM.S, yyt->FORMAT_PARAM.S)
return;
case kRETURN_PARAM: (* yyNew)->RETURN_PARAM = yyt->RETURN_PARAM;
copyint ((* yyNew)->RETURN_PARAM.label, yyt->RETURN_PARAM.label)
return;
case kNO_PARAM: (* yyNew)->NO_PARAM = yyt->NO_PARAM;
copyint ((* yyNew)->NO_PARAM.type, yyt->NO_PARAM.type)
return;
case kBT_PARAM_LIST: (* yyNew)->BT_PARAM_LIST = yyt->BT_PARAM_LIST;
return;
case kBTP_EMPTY: (* yyNew)->BTP_EMPTY = yyt->BTP_EMPTY;
return;
case kBTP_LIST: (* yyNew)->BTP_LIST = yyt->BTP_LIST;
copytTree ((* yyNew)->BTP_LIST.Elem, yyt->BTP_LIST.Elem)
yyt = yyt->BTP_LIST.Next;
yyNew = & (* yyNew)->BTP_LIST.Next; break;
case kCONSTANT: (* yyNew)->CONSTANT = yyt->CONSTANT;
return;
case kBOOL_CONSTANT: (* yyNew)->BOOL_CONSTANT = yyt->BOOL_CONSTANT;
copyint ((* yyNew)->BOOL_CONSTANT.value, yyt->BOOL_CONSTANT.value)
return;
case kINT_CONSTANT: (* yyNew)->INT_CONSTANT = yyt->INT_CONSTANT;
copyint ((* yyNew)->INT_CONSTANT.value, yyt->INT_CONSTANT.value)
return;
case kREAL_CONSTANT: (* yyNew)->REAL_CONSTANT = yyt->REAL_CONSTANT;
copytStringRef ((* yyNew)->REAL_CONSTANT.value, yyt->REAL_CONSTANT.value)
copyint ((* yyNew)->REAL_CONSTANT.kind, yyt->REAL_CONSTANT.kind)
return;
case kSTRING_CONSTANT: (* yyNew)->STRING_CONSTANT = yyt->STRING_CONSTANT;
copytStringRef ((* yyNew)->STRING_CONSTANT.value, yyt->STRING_CONSTANT.value)
return;
case kCOMPLEX_CONSTANT: (* yyNew)->COMPLEX_CONSTANT = yyt->COMPLEX_CONSTANT;
copytStringRef ((* yyNew)->COMPLEX_CONSTANT.rvalue, yyt->COMPLEX_CONSTANT.rvalue)
copytStringRef ((* yyNew)->COMPLEX_CONSTANT.ivalue, yyt->COMPLEX_CONSTANT.ivalue)
return;
case kPROC_OBJ: (* yyNew)->PROC_OBJ = yyt->PROC_OBJ;
copytIdent ((* yyNew)->PROC_OBJ.Ident, yyt->PROC_OBJ.Ident)
copytDefinitions ((* yyNew)->PROC_OBJ.Object, yyt->PROC_OBJ.Object)
return;
case kVAR_OBJ: (* yyNew)->VAR_OBJ = yyt->VAR_OBJ;
copyint ((* yyNew)->VAR_OBJ.Pos, yyt->VAR_OBJ.Pos)
copytIdent ((* yyNew)->VAR_OBJ.Ident, yyt->VAR_OBJ.Ident)
copytDefinitions ((* yyNew)->VAR_OBJ.Object, yyt->VAR_OBJ.Object)
copytReaching ((* yyNew)->VAR_OBJ.Reaching, yyt->VAR_OBJ.Reaching)
return;
case kTYPE_OBJ: (* yyNew)->TYPE_OBJ = yyt->TYPE_OBJ;
copytIdent ((* yyNew)->TYPE_OBJ.Ident, yyt->TYPE_OBJ.Ident)
copytDefinitions ((* yyNew)->TYPE_OBJ.Object, yyt->TYPE_OBJ.Object)
return;
case kREC_COMP: (* yyNew)->REC_COMP = yyt->REC_COMP;
copytIdent ((* yyNew)->REC_COMP.Ident, yyt->REC_COMP.Ident)
copytDefinitions ((* yyNew)->REC_COMP.Object, yyt->REC_COMP.Object)
return;
case kTOP_OBJ: (* yyNew)->TOP_OBJ = yyt->TOP_OBJ;
copytIdent ((* yyNew)->TOP_OBJ.Ident, yyt->TOP_OBJ.Ident)
copytDefinitions ((* yyNew)->TOP_OBJ.Object, yyt->TOP_OBJ.Object)
return;
case kTEMP_OBJ: (* yyNew)->TEMP_OBJ = yyt->TEMP_OBJ;
copytIdent ((* yyNew)->TEMP_OBJ.Ident, yyt->TEMP_OBJ.Ident)
copytDefinitions ((* yyNew)->TEMP_OBJ.Object, yyt->TEMP_OBJ.Object)
return;
case kOVERLAP_SPEC: (* yyNew)->OVERLAP_SPEC = yyt->OVERLAP_SPEC;
copyint ((* yyNew)->OVERLAP_SPEC.left_size, yyt->OVERLAP_SPEC.left_size)
copyint ((* yyNew)->OVERLAP_SPEC.right_size, yyt->OVERLAP_SPEC.right_size)
copyint ((* yyNew)->OVERLAP_SPEC.left_update, yyt->OVERLAP_SPEC.left_update)
copyint ((* yyNew)->OVERLAP_SPEC.right_update, yyt->OVERLAP_SPEC.right_update)
return;
case kDEPENDENCE_INFO: (* yyNew)->DEPENDENCE_INFO = yyt->DEPENDENCE_INFO;
return;
case kSERIAL_INFO: (* yyNew)->SERIAL_INFO = yyt->SERIAL_INFO;
return;
case kINDEP_INFO: (* yyNew)->INDEP_INFO = yyt->INDEP_INFO;
copybool ((* yyNew)->INDEP_INFO.user_independent, yyt->INDEP_INFO.user_independent)
copyint ((* yyNew)->INDEP_INFO.selection, yyt->INDEP_INFO.selection)
copyInfoTree ((* yyNew)->INDEP_INFO.red_vars, yyt->INDEP_INFO.red_vars)
copyInfoTree ((* yyNew)->INDEP_INFO.new_vars, yyt->INDEP_INFO.new_vars)
return;
case kVECTOR_INFO: (* yyNew)->VECTOR_INFO = yyt->VECTOR_INFO;
copybool ((* yyNew)->VECTOR_INFO.is_independent, yyt->VECTOR_INFO.is_independent)
copyint ((* yyNew)->VECTOR_INFO.selection, yyt->VECTOR_INFO.selection)
copyint ((* yyNew)->VECTOR_INFO.loop_cnt, yyt->VECTOR_INFO.loop_cnt)
copyInfoTree ((* yyNew)->VECTOR_INFO.independent_vars, yyt->VECTOR_INFO.independent_vars)
return;
case kPARDO_INFO: (* yyNew)->PARDO_INFO = yyt->PARDO_INFO;
return;
case kHOME_INFO: (* yyNew)->HOME_INFO = yyt->HOME_INFO;
return;
case kNO_HOME_INFO: (* yyNew)->NO_HOME_INFO = yyt->NO_HOME_INFO;
return;
case kCOMM_INFO: (* yyNew)->COMM_INFO = yyt->COMM_INFO;
copyInfoTree ((* yyNew)->COMM_INFO.home_var, yyt->COMM_INFO.home_var)
copybool ((* yyNew)->COMM_INFO.is_local, yyt->COMM_INFO.is_local)
return;
case kON_INFO: (* yyNew)->ON_INFO = yyt->ON_INFO;
copytObject ((* yyNew)->ON_INFO.on_id, yyt->ON_INFO.on_id)
copyint ((* yyNew)->ON_INFO.on_dim, yyt->ON_INFO.on_dim)
copyint ((* yyNew)->ON_INFO.base, yyt->ON_INFO.base)
copyint ((* yyNew)->ON_INFO.stride, yyt->ON_INFO.stride)
return;
case kPARALLEL_INFO: (* yyNew)->PARALLEL_INFO = yyt->PARALLEL_INFO;
copyInfoTree ((* yyNew)->PARALLEL_INFO.reduction_vars, yyt->PARALLEL_INFO.reduction_vars)
copyInfoTree ((* yyNew)->PARALLEL_INFO.home_var, yyt->PARALLEL_INFO.home_var)
copyInfoTree ((* yyNew)->PARALLEL_INFO.new_vars, yyt->PARALLEL_INFO.new_vars)
copyInfoTree ((* yyNew)->PARALLEL_INFO.resident_vars, yyt->PARALLEL_INFO.resident_vars)
return;
case kREACHING_INFO: (* yyNew)->REACHING_INFO = yyt->REACHING_INFO;
copyInfoTree ((* yyNew)->REACHING_INFO.var_allocate, yyt->REACHING_INFO.var_allocate)
copyInfoTree ((* yyNew)->REACHING_INFO.temp_allocate, yyt->REACHING_INFO.temp_allocate)
copytDefinitions ((* yyNew)->REACHING_INFO.var_distribution, yyt->REACHING_INFO.var_distribution)
copytDefinitions ((* yyNew)->REACHING_INFO.temp_distribution, yyt->REACHING_INFO.temp_distribution)
return;
  default: ;
  }
 }
}

tTree CopyTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyt)
# else
 (yyt) tTree yyt;
# endif
{
 tTree yyNew;
 yyMark (yyt);
 yyOldToNewCount = 0;
 yyCopyTree (yyt, & yyNew);
 return yyNew;
}

static bool yyCheckTree ARGS((tTree yyt));

bool CheckTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyt)
# else
 (yyt) tTree yyt;
# endif
{
 yyMark (yyt);
 return yyCheckTree (yyt);
}

static bool yyCheckChild
# if defined __STDC__ | defined __cplusplus
 (tTree yyParent, tTree yyyChild, Tree_tKind yyType, char * yySelector)
# else
 (yyParent, yyyChild, yyType, yySelector)
 tTree yyParent, yyyChild;
 Tree_tKind yyType;
 char * yySelector;
# endif
{
 bool yySuccess = Tree_IsType (yyyChild, yyType);
 if (! yySuccess) {
  (void) fputs ("CheckTree: parent = ", stderr);
  WriteTreeNode (stderr, yyParent);
  (void) fprintf (stderr, "\nselector: %s child = ", yySelector);
  WriteTreeNode (stderr, yyyChild);
  (void) fputc ('\n', stderr);
 }
 return yyCheckTree (yyyChild) && yySuccess;
}

static bool yyCheckTree
# if defined __STDC__ | defined __cplusplus
 (tTree yyt)
# else
 (yyt) tTree yyt;
# endif
{
 bool yyResult;
 if (yyt == NoTree) return false;
 else if (yyt->yyHead.yyMark == 0) return true;
 yyt->yyHead.yyMark = 0;

 yyResult = true;
 switch (yyt->Kind) {
case kCOMP_UNIT:
yyResult = yyCheckChild (yyt, yyt->COMP_UNIT.COMP_ELEMENTS, kUNIT_NODE_LIST, "COMP_ELEMENTS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->COMP_UNIT.USED_ELEMENTS, kUNIT_NODE_LIST, "USED_ELEMENTS") && yyResult;
break;
case kUNIT_LIST:
yyResult = yyCheckChild (yyt, yyt->UNIT_LIST.Elem, kUNIT_NODE, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->UNIT_LIST.Next, kUNIT_NODE_LIST, "Next") && yyResult;
break;
case kMODULE_DECL:
yyResult = yyCheckChild (yyt, yyt->MODULE_DECL.MODULE_BODY, kBODY_NODE, "MODULE_BODY") && yyResult;
break;
case kPROGRAM_DECL:
yyResult = yyCheckChild (yyt, yyt->PROGRAM_DECL.FORMALS, kDECL_NODE_LIST, "FORMALS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->PROGRAM_DECL.PROGRAM_BODY, kBODY_NODE, "PROGRAM_BODY") && yyResult;
break;
case kPROC_DECL:
yyResult = yyCheckChild (yyt, yyt->PROC_DECL.FORMALS, kDECL_NODE_LIST, "FORMALS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->PROC_DECL.PROC_BODY, kBODY_NODE, "PROC_BODY") && yyResult;
break;
case kFUNC_DECL:
yyResult = yyCheckChild (yyt, yyt->FUNC_DECL.FORMALS, kDECL_NODE_LIST, "FORMALS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->FUNC_DECL.FUNC_BODY, kBODY_NODE, "FUNC_BODY") && yyResult;
yyResult = yyCheckChild (yyt, yyt->FUNC_DECL.RESULT_TYPE, kTYPE_NODE, "RESULT_TYPE") && yyResult;
break;
case kBLOCK_DATA_DECL:
yyResult = yyCheckChild (yyt, yyt->BLOCK_DATA_DECL.DATA_BODY, kBODY_NODE, "DATA_BODY") && yyResult;
break;
case kINTERFACE:
yyResult = yyCheckChild (yyt, yyt->INTERFACE.SPEC, kGENERIC_SPEC, "SPEC") && yyResult;
yyResult = yyCheckChild (yyt, yyt->INTERFACE.ITEMS, kUNIT_NODE_LIST, "ITEMS") && yyResult;
break;
case kBODY_NODE:
yyResult = yyCheckChild (yyt, yyt->BODY_NODE.DECLS, kDECL_NODE_LIST, "DECLS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BODY_NODE.STATS, kACF_NODE_LIST, "STATS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BODY_NODE.INTERNALS, kUNIT_NODE_LIST, "INTERNALS") && yyResult;
break;
case kINTEGER_TYPE:
yyResult = yyCheckChild (yyt, yyt->INTEGER_TYPE.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kREAL_TYPE:
yyResult = yyCheckChild (yyt, yyt->REAL_TYPE.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kBOOLEAN_TYPE:
yyResult = yyCheckChild (yyt, yyt->BOOLEAN_TYPE.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kCOMPLEX_TYPE:
yyResult = yyCheckChild (yyt, yyt->COMPLEX_TYPE.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kSTRING_TYPE:
yyResult = yyCheckChild (yyt, yyt->STRING_TYPE.LENGTH, kBT_EXP, "LENGTH") && yyResult;
break;
case kTYPE_ID:
yyResult = yyCheckChild (yyt, yyt->TYPE_ID.ID, kTYPE_OBJ, "ID") && yyResult;
break;
case kARRAY_TYPE:
yyResult = yyCheckChild (yyt, yyt->ARRAY_TYPE.ARRAY_INDEX_TYPES, kSHAPE_SPEC_LIST, "ARRAY_INDEX_TYPES") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ARRAY_TYPE.ARRAY_COMP_TYPE, kTYPE_NODE, "ARRAY_COMP_TYPE") && yyResult;
break;
case kRECORD_TYPE:
yyResult = yyCheckChild (yyt, yyt->RECORD_TYPE.COMPONENTS, kDECL_NODE_LIST, "COMPONENTS") && yyResult;
break;
case kPOINTER_TYPE:
yyResult = yyCheckChild (yyt, yyt->POINTER_TYPE.PTR_COMP, kTYPE_NODE, "PTR_COMP") && yyResult;
break;
case kSHAPE_SPEC:
yyResult = yyCheckChild (yyt, yyt->SHAPE_SPEC.Overlap, kOVERLAP_SPEC, "Overlap") && yyResult;
break;
case kEXPLICIT_SHAPE:
yyResult = yyCheckChild (yyt, yyt->EXPLICIT_SHAPE.Overlap, kOVERLAP_SPEC, "Overlap") && yyResult;
yyResult = yyCheckChild (yyt, yyt->EXPLICIT_SHAPE.LOWER, kBT_EXP, "LOWER") && yyResult;
yyResult = yyCheckChild (yyt, yyt->EXPLICIT_SHAPE.UPPER, kBT_EXP, "UPPER") && yyResult;
break;
case kASSUMED_SHAPE:
yyResult = yyCheckChild (yyt, yyt->ASSUMED_SHAPE.Overlap, kOVERLAP_SPEC, "Overlap") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ASSUMED_SHAPE.LOWER, kBT_EXP, "LOWER") && yyResult;
break;
case kDEFERRED_SHAPE:
yyResult = yyCheckChild (yyt, yyt->DEFERRED_SHAPE.Overlap, kOVERLAP_SPEC, "Overlap") && yyResult;
break;
case kASSUMED_SIZE:
yyResult = yyCheckChild (yyt, yyt->ASSUMED_SIZE.Overlap, kOVERLAP_SPEC, "Overlap") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ASSUMED_SIZE.LOWER, kBT_EXP, "LOWER") && yyResult;
break;
case kSHAPE_LIST:
yyResult = yyCheckChild (yyt, yyt->SHAPE_LIST.Elem, kSHAPE_SPEC, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SHAPE_LIST.Next, kSHAPE_SPEC_LIST, "Next") && yyResult;
break;
case kTYPE_DECL:
yyResult = yyCheckChild (yyt, yyt->TYPE_DECL.VAL, kTYPE_NODE, "VAL") && yyResult;
break;
case kENTITY_DECL:
yyResult = yyCheckChild (yyt, yyt->ENTITY_DECL.ATTRIBUTES, kDECL_LIST, "ATTRIBUTES") && yyResult;
break;
case kVAR_DECL:
yyResult = yyCheckChild (yyt, yyt->VAR_DECL.VAL, kTYPE_NODE, "VAL") && yyResult;
break;
case kCOMMON_DECL:
yyResult = yyCheckChild (yyt, yyt->COMMON_DECL.IDS, kDECL_NODE_LIST, "IDS") && yyResult;
break;
case kNAMELIST_DECL:
yyResult = yyCheckChild (yyt, yyt->NAMELIST_DECL.IDS, kDECL_NODE_LIST, "IDS") && yyResult;
break;
case kTEMPLATE_DECL:
yyResult = yyCheckChild (yyt, yyt->TEMPLATE_DECL.DIMENSIONS, kSHAPE_SPEC_LIST, "DIMENSIONS") && yyResult;
break;
case kRAGGED_DECL:
yyResult = yyCheckChild (yyt, yyt->RAGGED_DECL.DIMENSIONS, kSHAPE_SPEC_LIST, "DIMENSIONS") && yyResult;
break;
case kPROCESSORS_DECL:
yyResult = yyCheckChild (yyt, yyt->PROCESSORS_DECL.DIMENSIONS, kSHAPE_SPEC_LIST, "DIMENSIONS") && yyResult;
break;
case kSUB_PROCS_DECL:
yyResult = yyCheckChild (yyt, yyt->SUB_PROCS_DECL.TOP, kTOP_OBJ, "TOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SUB_PROCS_DECL.DIMENSIONS, kBTE_LIST, "DIMENSIONS") && yyResult;
break;
case kRANGE_DECL:
yyResult = yyCheckChild (yyt, yyt->RANGE_DECL.RANGES, kRANGE_SPEC_LIST, "RANGES") && yyResult;
break;
case kSTMT_FUNC_DECL:
yyResult = yyCheckChild (yyt, yyt->STMT_FUNC_DECL.FORMALS, kDECL_NODE_LIST, "FORMALS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->STMT_FUNC_DECL.RESULT_TYPE, kTYPE_NODE, "RESULT_TYPE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->STMT_FUNC_DECL.FFUNC_BODY, kBT_EXP, "FFUNC_BODY") && yyResult;
break;
case kINTERFACE_DECL:
yyResult = yyCheckChild (yyt, yyt->INTERFACE_DECL.SPEC, kGENERIC_SPEC, "SPEC") && yyResult;
yyResult = yyCheckChild (yyt, yyt->INTERFACE_DECL.ITEMS, kUNIT_NODE_LIST, "ITEMS") && yyResult;
break;
case kVAR_PARAM_DECL:
yyResult = yyCheckChild (yyt, yyt->VAR_PARAM_DECL.VAL, kTYPE_NODE, "VAL") && yyResult;
break;
case kIMPLICIT_DECL:
yyResult = yyCheckChild (yyt, yyt->IMPLICIT_DECL.TYPE, kTYPE_NODE, "TYPE") && yyResult;
break;
case kUSE_DECL:
yyResult = yyCheckChild (yyt, yyt->USE_DECL.RENAMINGS, kDECL_NODE_LIST, "RENAMINGS") && yyResult;
break;
case kONLY_USE_DECL:
yyResult = yyCheckChild (yyt, yyt->ONLY_USE_DECL.NAMES, kDECL_NODE_LIST, "NAMES") && yyResult;
break;
case kEQV_DECL:
yyResult = yyCheckChild (yyt, yyt->EQV_DECL.VARS, kBT_VAR_LIST, "VARS") && yyResult;
break;
case kDATA_DECL:
yyResult = yyCheckChild (yyt, yyt->DATA_DECL.VARS, kBT_VAR_LIST, "VARS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DATA_DECL.VALS, kBT_EXP_LIST, "VALS") && yyResult;
break;
case kFORMAT_DECL:
yyResult = yyCheckChild (yyt, yyt->FORMAT_DECL.FORMAT_PARAMS, kBT_PARAM_LIST, "FORMAT_PARAMS") && yyResult;
break;
case kDIMENSION_DECL:
yyResult = yyCheckChild (yyt, yyt->DIMENSION_DECL.INDEXES, kSHAPE_SPEC_LIST, "INDEXES") && yyResult;
break;
case kINIT_DATA_DECL:
yyResult = yyCheckChild (yyt, yyt->INIT_DATA_DECL.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kTYPESPEC_DECL:
yyResult = yyCheckChild (yyt, yyt->TYPESPEC_DECL.VAL, kTYPE_NODE, "VAL") && yyResult;
break;
case kPARAMETER_DECL:
yyResult = yyCheckChild (yyt, yyt->PARAMETER_DECL.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kDISTRIBUTE_DECL:
yyResult = yyCheckChild (yyt, yyt->DISTRIBUTE_DECL.DISTRIBUTION, kDISTRIBUTION_SPEC, "DISTRIBUTION") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DISTRIBUTE_DECL.TARGET, kPROCESSOR_SPEC, "TARGET") && yyResult;
break;
case kALIGN_DECL:
yyResult = yyCheckChild (yyt, yyt->ALIGN_DECL.ALIGN_SOURCE, kBT_EXP_LIST, "ALIGN_SOURCE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ALIGN_DECL.ALIGN_TARGET, kALIGN_SPEC, "ALIGN_TARGET") && yyResult;
break;
case kMAP_DECL:
yyResult = yyCheckChild (yyt, yyt->MAP_DECL.MAP_TARGET, kBT_EXP, "MAP_TARGET") && yyResult;
break;
case kOVERLAP_DECL:
yyResult = yyCheckChild (yyt, yyt->OVERLAP_DECL.OVERLAPS, kSHAPE_SPEC_LIST, "OVERLAPS") && yyResult;
break;
case kSELECT_DECL:
yyResult = yyCheckChild (yyt, yyt->SELECT_DECL.SELECTIONS, kSELECT_SPEC_LIST, "SELECTIONS") && yyResult;
break;
case kDECL_LIST:
yyResult = yyCheckChild (yyt, yyt->DECL_LIST.Elem, kDECL_NODE, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DECL_LIST.Next, kDECL_NODE_LIST, "Next") && yyResult;
break;
case kRANGE_LIST:
yyResult = yyCheckChild (yyt, yyt->RANGE_LIST.Elem, kDISTRIBUTION_SPEC, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->RANGE_LIST.Next, kRANGE_SPEC_LIST, "Next") && yyResult;
break;
case kSELECT_LIST:
yyResult = yyCheckChild (yyt, yyt->SELECT_LIST.Elem, kSELECT_SPEC, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SELECT_LIST.Next, kSELECT_SPEC_LIST, "Next") && yyResult;
break;
case kNODE_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->NODE_DISTRIBUTION.MAPPING, kDIST_FORMAT_LIST, "MAPPING") && yyResult;
break;
case kDIST_LIST:
yyResult = yyCheckChild (yyt, yyt->DIST_LIST.Elem, kDISTRIBUTION_FORMAT, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DIST_LIST.Next, kDIST_FORMAT_LIST, "Next") && yyResult;
break;
case kBLOCK_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->BLOCK_DISTRIBUTION.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kCYCLIC_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->CYCLIC_DISTRIBUTION.SIZE, kBT_EXP, "SIZE") && yyResult;
break;
case kGEN_BLOCK_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->GEN_BLOCK_DISTRIBUTION.ARRAY, kBT_VAR, "ARRAY") && yyResult;
break;
case kINDIRECT_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->INDIRECT_DISTRIBUTION.ARRAY, kBT_VAR, "ARRAY") && yyResult;
break;
case kARBITRARY_DISTRIBUTION:
yyResult = yyCheckChild (yyt, yyt->ARBITRARY_DISTRIBUTION.SIZE, kBT_EXP, "SIZE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ARBITRARY_DISTRIBUTION.LENGTH, kBT_VAR, "LENGTH") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ARBITRARY_DISTRIBUTION.MAP, kBT_VAR, "MAP") && yyResult;
break;
case kPROCESSOR_ARRAY:
yyResult = yyCheckChild (yyt, yyt->PROCESSOR_ARRAY.TOPNAME, kTOP_OBJ, "TOPNAME") && yyResult;
break;
case kPROCESSOR_SUBSET:
yyResult = yyCheckChild (yyt, yyt->PROCESSOR_SUBSET.SUBTOP, kTOP_OBJ, "SUBTOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->PROCESSOR_SUBSET.FULLTOP, kTOP_OBJ, "FULLTOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->PROCESSOR_SUBSET.SUBSCRIPTS, kBT_EXP_LIST, "SUBSCRIPTS") && yyResult;
break;
case kALIGN_SPEC:
yyResult = yyCheckChild (yyt, yyt->ALIGN_SPEC.TEMPNAME, kTEMP_OBJ, "TEMPNAME") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ALIGN_SPEC.SUBSCRIPTS, kBT_EXP_LIST, "SUBSCRIPTS") && yyResult;
break;
case kOP_GENERIC_SPEC:
yyResult = yyCheckChild (yyt, yyt->OP_GENERIC_SPEC.OPERATOR, kOP, "OPERATOR") && yyResult;
break;
case kACF_LIST:
yyResult = yyCheckChild (yyt, yyt->ACF_LIST.Elem, kACF_NODE, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_LIST.Next, kACF_NODE_LIST, "Next") && yyResult;
break;
case kACF_BASIC:
yyResult = yyCheckChild (yyt, yyt->ACF_BASIC.BASIC_STMT, kBT_STMT, "BASIC_STMT") && yyResult;
break;
case kACF_IF:
yyResult = yyCheckChild (yyt, yyt->ACF_IF.IF_EXP, kBT_EXP, "IF_EXP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_IF.THEN_PART, kACF_NODE_LIST, "THEN_PART") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_IF.ELSE_PART, kACF_NODE_LIST, "ELSE_PART") && yyResult;
break;
case kACF_WHERE:
yyResult = yyCheckChild (yyt, yyt->ACF_WHERE.WHERE_EXP, kBT_EXP, "WHERE_EXP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_WHERE.TRUE_PART, kACF_NODE_LIST, "TRUE_PART") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_WHERE.FALSE_PART, kACF_NODE_LIST, "FALSE_PART") && yyResult;
break;
case kACF_CASE:
yyResult = yyCheckChild (yyt, yyt->ACF_CASE.CASE_EXP, kBT_EXP, "CASE_EXP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_CASE.CASE_ALTS, kSELECTED_ACF_NODE_LIST, "CASE_ALTS") && yyResult;
break;
case kACF_WHILE:
yyResult = yyCheckChild (yyt, yyt->ACF_WHILE.WHILE_EXP, kBT_EXP, "WHILE_EXP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_WHILE.WHILE_BODY, kACF_NODE_LIST, "WHILE_BODY") && yyResult;
break;
case kACF_REPEAT:
yyResult = yyCheckChild (yyt, yyt->ACF_REPEAT.REPEAT_BODY, kACF_NODE_LIST, "REPEAT_BODY") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_REPEAT.REPEAT_EXP, kBT_EXP, "REPEAT_EXP") && yyResult;
break;
case kACF_LOOP:
yyResult = yyCheckChild (yyt, yyt->ACF_LOOP.LOOP_BODY, kACF_NODE_LIST, "LOOP_BODY") && yyResult;
break;
case kACF_DO:
yyResult = yyCheckChild (yyt, yyt->ACF_DO.DO_ID, kBT_VAR, "DO_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_DO.DO_RANGE, kSLICE_EXP, "DO_RANGE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_DO.DO_BODY, kACF_NODE_LIST, "DO_BODY") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_DO.DO_DEP_INFO, kDEPENDENCE_INFO, "DO_DEP_INFO") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_DO.DO_HOME_INFO, kHOME_INFO, "DO_HOME_INFO") && yyResult;
break;
case kACF_FORALL:
yyResult = yyCheckChild (yyt, yyt->ACF_FORALL.FORALL_ID, kBT_VAR, "FORALL_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_FORALL.FORALL_RANGE, kSLICE_EXP, "FORALL_RANGE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_FORALL.FORALL_BODY, kACF_NODE_LIST, "FORALL_BODY") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_FORALL.FORALL_DEP_INFO, kDEPENDENCE_INFO, "FORALL_DEP_INFO") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_FORALL.FORALL_HOME_INFO, kHOME_INFO, "FORALL_HOME_INFO") && yyResult;
break;
case kACF_FLOW_GRAPH:
yyResult = yyCheckChild (yyt, yyt->ACF_FLOW_GRAPH.FLOW_NODES, kACF_NODE_LIST, "FLOW_NODES") && yyResult;
break;
case kACF_ENTRY:
yyResult = yyCheckChild (yyt, yyt->ACF_ENTRY.ENTRY_DECL, kUNIT_NODE, "ENTRY_DECL") && yyResult;
break;
case kACF_HOME:
yyResult = yyCheckChild (yyt, yyt->ACF_HOME.HOME_VAR, kON_CLAUSE, "HOME_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_HOME.HOME_BODY, kACF_NODE_LIST, "HOME_BODY") && yyResult;
break;
case kACF_RESIDENT:
yyResult = yyCheckChild (yyt, yyt->ACF_RESIDENT.RESIDENT_VAR, kBT_VAR_LIST, "RESIDENT_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_RESIDENT.RESIDENT_BODY, kACF_NODE_LIST, "RESIDENT_BODY") && yyResult;
break;
case kACF_NEW:
yyResult = yyCheckChild (yyt, yyt->ACF_NEW.NEW_VAR, kBT_VAR_LIST, "NEW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_NEW.NEW_BODY, kACF_NODE_LIST, "NEW_BODY") && yyResult;
break;
case kACF_REDUCTION:
yyResult = yyCheckChild (yyt, yyt->ACF_REDUCTION.REDUCTION_VAR, kBT_VAR_LIST, "REDUCTION_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_REDUCTION.REDUCTION_LVAR, kBT_VAR_LIST, "REDUCTION_LVAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_REDUCTION.REDUCTION_FUNC, kPROC_OBJ, "REDUCTION_FUNC") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_REDUCTION.REDUCTION_BODY, kACF_NODE_LIST, "REDUCTION_BODY") && yyResult;
break;
case kACF_TASK_REGION:
yyResult = yyCheckChild (yyt, yyt->ACF_TASK_REGION.TASK_BODY, kACF_NODE_LIST, "TASK_BODY") && yyResult;
break;
case kACF_PARALLEL:
yyResult = yyCheckChild (yyt, yyt->ACF_PARALLEL.REDUCTION_VARS, kBT_VAR_LIST, "REDUCTION_VARS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_PARALLEL.SHARED_VARS, kBT_VAR_LIST, "SHARED_VARS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_PARALLEL.LOCAL_VARS, kBT_VAR_LIST, "LOCAL_VARS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_PARALLEL.PARALLEL_BODY, kACF_NODE_LIST, "PARALLEL_BODY") && yyResult;
break;
case kACF_CRITICAL:
yyResult = yyCheckChild (yyt, yyt->ACF_CRITICAL.CRITICAL_BODY, kACF_NODE_LIST, "CRITICAL_BODY") && yyResult;
break;
case kACF_ON:
yyResult = yyCheckChild (yyt, yyt->ACF_ON.ON_HOME, kON_TOPOLOGY, "ON_HOME") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_ON.ON_SPECS, kON_SPEC_LIST, "ON_SPECS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_ON.ON_BODY, kACF_NODE_LIST, "ON_BODY") && yyResult;
break;
case kACF_RM_READ:
yyResult = yyCheckChild (yyt, yyt->ACF_RM_READ.LOCAL_VAR, kBT_VAR, "LOCAL_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_RM_READ.REMOTE_VAR, kBT_VAR, "REMOTE_VAR") && yyResult;
break;
case kACF_RM_WRITE:
yyResult = yyCheckChild (yyt, yyt->ACF_RM_WRITE.REMOTE_VAR, kBT_VAR, "REMOTE_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ACF_RM_WRITE.LOCAL_DATA, kBT_EXP, "LOCAL_DATA") && yyResult;
break;
case kACF_MOVE:
yyResult = yyCheckChild (yyt, yyt->ACF_MOVE.MOVE_STMT, kBT_STMT, "MOVE_STMT") && yyResult;
break;
case kACF_BARRIER:
yyResult = yyCheckChild (yyt, yyt->ACF_BARRIER.BARRIER_VARS, kBT_VAR_LIST, "BARRIER_VARS") && yyResult;
break;
case kACF_UPDATE:
yyResult = yyCheckChild (yyt, yyt->ACF_UPDATE.OVERLAP_VAR, kBT_VAR, "OVERLAP_VAR") && yyResult;
break;
case kON_VAR_CLAUSE:
yyResult = yyCheckChild (yyt, yyt->ON_VAR_CLAUSE.ON_VAR, kBT_VAR, "ON_VAR") && yyResult;
break;
case kON_PROC_CLAUSE:
yyResult = yyCheckChild (yyt, yyt->ON_PROC_CLAUSE.ON_PROC, kPROCESSOR_SPEC, "ON_PROC") && yyResult;
break;
case kON_LIST:
yyResult = yyCheckChild (yyt, yyt->ON_LIST.Elem, kON_SPEC, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ON_LIST.Next, kON_SPEC_LIST, "Next") && yyResult;
break;
case kON_SPEC:
yyResult = yyCheckChild (yyt, yyt->ON_SPEC.ON_VAL, kBT_EXP, "ON_VAL") && yyResult;
break;
case kSELECTED_ACF_LIST:
yyResult = yyCheckChild (yyt, yyt->SELECTED_ACF_LIST.Elem, kSELECTED_ACF_NODE, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SELECTED_ACF_LIST.Next, kSELECTED_ACF_NODE_LIST, "Next") && yyResult;
break;
case kSELECTED_ACF_NODE:
yyResult = yyCheckChild (yyt, yyt->SELECTED_ACF_NODE.SELECT_LIST, kBT_EXP_LIST, "SELECT_LIST") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SELECTED_ACF_NODE.SELECT_ACFS, kACF_NODE_LIST, "SELECT_ACFS") && yyResult;
break;
case kASSIGN_STMT:
yyResult = yyCheckChild (yyt, yyt->ASSIGN_STMT.ASSIGN_VAR, kBT_VAR, "ASSIGN_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ASSIGN_STMT.ASSIGN_EXP, kBT_EXP, "ASSIGN_EXP") && yyResult;
break;
case kPTR_ASSIGN_STMT:
yyResult = yyCheckChild (yyt, yyt->PTR_ASSIGN_STMT.ASSIGN_VAR, kBT_VAR, "ASSIGN_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->PTR_ASSIGN_STMT.ASSIGN_EXP, kBT_EXP, "ASSIGN_EXP") && yyResult;
break;
case kLABEL_ASSIGN_STMT:
yyResult = yyCheckChild (yyt, yyt->LABEL_ASSIGN_STMT.LABEL_VAR, kBT_VAR, "LABEL_VAR") && yyResult;
break;
case kCALL_STMT:
yyResult = yyCheckChild (yyt, yyt->CALL_STMT.CALL_ID, kPROC_OBJ, "CALL_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->CALL_STMT.CALL_PARAMS, kBT_PARAM_LIST, "CALL_PARAMS") && yyResult;
break;
case kIO_STMT:
yyResult = yyCheckChild (yyt, yyt->IO_STMT.ID, kPROC_OBJ, "ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IO_STMT.IO_SPECS, kBT_PARAM_LIST, "IO_SPECS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IO_STMT.IO_ITEMS, kBT_PARAM_LIST, "IO_ITEMS") && yyResult;
break;
case kASS_GOTO_STMT:
yyResult = yyCheckChild (yyt, yyt->ASS_GOTO_STMT.GOTO_VAR, kBT_VAR, "GOTO_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ASS_GOTO_STMT.LABELS, kLABEL_USE_LIST, "LABELS") && yyResult;
break;
case kCOMP_GOTO_STMT:
yyResult = yyCheckChild (yyt, yyt->COMP_GOTO_STMT.GOTO_LABELS, kLABEL_USE_LIST, "GOTO_LABELS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->COMP_GOTO_STMT.GOTO_EXP, kBT_EXP, "GOTO_EXP") && yyResult;
break;
case kCOMP_IF_STMT:
yyResult = yyCheckChild (yyt, yyt->COMP_IF_STMT.IF_EXP, kBT_EXP, "IF_EXP") && yyResult;
break;
case kRETURN_STMT:
yyResult = yyCheckChild (yyt, yyt->RETURN_STMT.RETURN_EXP, kBT_EXP, "RETURN_EXP") && yyResult;
break;
case kPAUSE_STMT:
yyResult = yyCheckChild (yyt, yyt->PAUSE_STMT.PAUSE_CONST, kCONSTANT, "PAUSE_CONST") && yyResult;
break;
case kSTOP_STMT:
yyResult = yyCheckChild (yyt, yyt->STOP_STMT.STOP_CONST, kBT_EXP, "STOP_CONST") && yyResult;
break;
case kFORMAT_STMT:
yyResult = yyCheckChild (yyt, yyt->FORMAT_STMT.FORMAT_PARAMS, kBT_PARAM_LIST, "FORMAT_PARAMS") && yyResult;
break;
case kINHERIT_DSP_STMT:
yyResult = yyCheckChild (yyt, yyt->INHERIT_DSP_STMT.VAR, kBT_VAR, "VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->INHERIT_DSP_STMT.SHAPE, kSHAPE_SPEC_LIST, "SHAPE") && yyResult;
break;
case kCREATE_DSP_STMT:
yyResult = yyCheckChild (yyt, yyt->CREATE_DSP_STMT.VAR, kBT_VAR, "VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->CREATE_DSP_STMT.SHAPE, kSHAPE_SPEC_LIST, "SHAPE") && yyResult;
break;
case kFREE_DSP_STMT:
yyResult = yyCheckChild (yyt, yyt->FREE_DSP_STMT.VAR, kBT_VAR, "VAR") && yyResult;
break;
case kALLOCATE_STMT:
yyResult = yyCheckChild (yyt, yyt->ALLOCATE_STMT.PARAMS, kBT_PARAM_LIST, "PARAMS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->ALLOCATE_STMT.STATUS, kBT_VAR, "STATUS") && yyResult;
break;
case kDEALLOCATE_STMT:
yyResult = yyCheckChild (yyt, yyt->DEALLOCATE_STMT.PARAMS, kBT_PARAM_LIST, "PARAMS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DEALLOCATE_STMT.STATUS, kBT_VAR, "STATUS") && yyResult;
break;
case kNULLIFY_STMT:
yyResult = yyCheckChild (yyt, yyt->NULLIFY_STMT.PARAMS, kBT_PARAM_LIST, "PARAMS") && yyResult;
break;
case kREDUCE_STMT:
yyResult = yyCheckChild (yyt, yyt->REDUCE_STMT.RED_FUNC, kPROC_OBJ, "RED_FUNC") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDUCE_STMT.RED_PARAMS, kBT_PARAM_LIST, "RED_PARAMS") && yyResult;
break;
case kMOVE_STMT:
yyResult = yyCheckChild (yyt, yyt->MOVE_STMT.TARGET, kBT_VAR, "TARGET") && yyResult;
yyResult = yyCheckChild (yyt, yyt->MOVE_STMT.SOURCE, kBT_VAR, "SOURCE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->MOVE_STMT.MASK, kBT_PARAM, "MASK") && yyResult;
break;
case kSCATTER_STMT:
yyResult = yyCheckChild (yyt, yyt->SCATTER_STMT.ARRAY, kBT_VAR, "ARRAY") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SCATTER_STMT.BASE, kBT_VAR, "BASE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SCATTER_STMT.INDEXES, kBT_PARAM_LIST, "INDEXES") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SCATTER_STMT.MASK, kBT_PARAM, "MASK") && yyResult;
break;
case kREALIGN_STMT:
yyResult = yyCheckChild (yyt, yyt->REALIGN_STMT.ALIGNEE, kBT_VAR_LIST, "ALIGNEE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REALIGN_STMT.ALIGN_SOURCE, kBT_EXP_LIST, "ALIGN_SOURCE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REALIGN_STMT.ALIGN_TARGET, kALIGN_SPEC, "ALIGN_TARGET") && yyResult;
break;
case kREDISTRIBUTE_STMT:
yyResult = yyCheckChild (yyt, yyt->REDISTRIBUTE_STMT.DISTRIBUTEE, kBT_VAR_LIST, "DISTRIBUTEE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDISTRIBUTE_STMT.DISTRIBUTION, kDISTRIBUTION_SPEC, "DISTRIBUTION") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDISTRIBUTE_STMT.TARGET, kPROCESSOR_SPEC, "TARGET") && yyResult;
break;
case kREG_SHADOW_GET:
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_GET.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_GET.SHADOW_SIZE, kBT_EXP_LIST, "SHADOW_SIZE") && yyResult;
break;
case kREG_SHADOW_PUT:
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_PUT.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_PUT.SHADOW_SIZE, kBT_EXP_LIST, "SHADOW_SIZE") && yyResult;
break;
case kREG_SHADOW_SET:
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_SET.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_SET.SHADOW_SIZE, kBT_EXP_LIST, "SHADOW_SIZE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REG_SHADOW_SET.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kIND_SHADOW_CREATE:
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_CREATE.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_CREATE.INDEX_VAR, kBT_VAR, "INDEX_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_CREATE.TMP_INDEX_VAR, kBT_VAR, "TMP_INDEX_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_CREATE.MASK, kBT_VAR, "MASK") && yyResult;
break;
case kIND_SHADOW_FREE:
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_FREE.INDEX_VAR, kBT_VAR, "INDEX_VAR") && yyResult;
break;
case kIND_SHADOW_GET:
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_GET.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
break;
case kIND_SHADOW_PUT:
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_PUT.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
break;
case kIND_SHADOW_SET:
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_SET.SHADOW_VAR, kBT_VAR, "SHADOW_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->IND_SHADOW_SET.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kBROADCAST_STMT:
yyResult = yyCheckChild (yyt, yyt->BROADCAST_STMT.ELEMS, kBT_VAR_LIST, "ELEMS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BROADCAST_STMT.OWNER_TOP, kON_TOPOLOGY, "OWNER_TOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BROADCAST_STMT.OWNER_SPEC, kON_SPEC_LIST, "OWNER_SPEC") && yyResult;
break;
case kREDUCTION_STMT:
yyResult = yyCheckChild (yyt, yyt->REDUCTION_STMT.REDFUNC, kPROC_OBJ, "REDFUNC") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDUCTION_STMT.ELEMS, kBT_VAR_LIST, "ELEMS") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDUCTION_STMT.RED_TOP, kON_TOPOLOGY, "RED_TOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->REDUCTION_STMT.RED_SPEC, kON_SPEC_LIST, "RED_SPEC") && yyResult;
break;
case kDIRTY_STMT:
yyResult = yyCheckChild (yyt, yyt->DIRTY_STMT.ELEMS, kBT_VAR_LIST, "ELEMS") && yyResult;
break;
case kLABEL_LIST:
yyResult = yyCheckChild (yyt, yyt->LABEL_LIST.Next, kLABEL_USE_LIST, "Next") && yyResult;
break;
case kCONST_EXP:
yyResult = yyCheckChild (yyt, yyt->CONST_EXP.C, kCONSTANT, "C") && yyResult;
break;
case kOP_EXP:
yyResult = yyCheckChild (yyt, yyt->OP_EXP.EXP_OP, kOP, "EXP_OP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->OP_EXP.OPND1, kBT_EXP, "OPND1") && yyResult;
yyResult = yyCheckChild (yyt, yyt->OP_EXP.OPND2, kBT_EXP, "OPND2") && yyResult;
break;
case kOP1_EXP:
yyResult = yyCheckChild (yyt, yyt->OP1_EXP.EXP_OP1, kOP, "EXP_OP1") && yyResult;
yyResult = yyCheckChild (yyt, yyt->OP1_EXP.OPND, kBT_EXP, "OPND") && yyResult;
break;
case kSLICE_EXP:
yyResult = yyCheckChild (yyt, yyt->SLICE_EXP.START, kBT_EXP, "START") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SLICE_EXP.STOP, kBT_EXP, "STOP") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SLICE_EXP.INC, kBT_EXP, "INC") && yyResult;
break;
case kVAR_EXP:
yyResult = yyCheckChild (yyt, yyt->VAR_EXP.V, kBT_VAR, "V") && yyResult;
break;
case kFUNC_CALL_EXP:
yyResult = yyCheckChild (yyt, yyt->FUNC_CALL_EXP.FUNC_ID, kPROC_OBJ, "FUNC_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->FUNC_CALL_EXP.FUNC_PARAMS, kBT_PARAM_LIST, "FUNC_PARAMS") && yyResult;
break;
case kNAMED_EXP:
yyResult = yyCheckChild (yyt, yyt->NAMED_EXP.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kDO_EXP:
yyResult = yyCheckChild (yyt, yyt->DO_EXP.DO_ID, kBT_VAR, "DO_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DO_EXP.RANGE, kSLICE_EXP, "RANGE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DO_EXP.BODY, kBT_EXP_LIST, "BODY") && yyResult;
break;
case kTYPE_EXP:
yyResult = yyCheckChild (yyt, yyt->TYPE_EXP.ID, kTYPE_OBJ, "ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->TYPE_EXP.ELEMENTS, kBT_EXP_LIST, "ELEMENTS") && yyResult;
break;
case kARRAY_EXP:
yyResult = yyCheckChild (yyt, yyt->ARRAY_EXP.ELEMENTS, kBT_EXP_LIST, "ELEMENTS") && yyResult;
break;
case kPERM_EXP:
yyResult = yyCheckChild (yyt, yyt->PERM_EXP.VAL, kBT_EXP, "VAL") && yyResult;
break;
case kBOUND_EXP:
yyResult = yyCheckChild (yyt, yyt->BOUND_EXP.VAR, kBT_VAR, "VAR") && yyResult;
break;
case kBTE_LIST:
yyResult = yyCheckChild (yyt, yyt->BTE_LIST.Elem, kBT_EXP, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BTE_LIST.Next, kBT_EXP_LIST, "Next") && yyResult;
break;
case kUSED_VAR:
yyResult = yyCheckChild (yyt, yyt->USED_VAR.VARNAME, kVAR_OBJ, "VARNAME") && yyResult;
break;
case kINDEXED_VAR:
yyResult = yyCheckChild (yyt, yyt->INDEXED_VAR.IND_VAR, kBT_VAR, "IND_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->INDEXED_VAR.IND_EXPS, kBT_EXP_LIST, "IND_EXPS") && yyResult;
break;
case kSUBSTRING_VAR:
yyResult = yyCheckChild (yyt, yyt->SUBSTRING_VAR.IND_VAR, kBT_VAR, "IND_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SUBSTRING_VAR.IND_EXP, kBT_EXP, "IND_EXP") && yyResult;
break;
case kSELECTED_VAR:
yyResult = yyCheckChild (yyt, yyt->SELECTED_VAR.SELEC_VAR, kBT_VAR, "SELEC_VAR") && yyResult;
yyResult = yyCheckChild (yyt, yyt->SELECTED_VAR.SELECTOR, kREC_COMP, "SELECTOR") && yyResult;
break;
case kLOOP_VAR:
yyResult = yyCheckChild (yyt, yyt->LOOP_VAR.LOOP_VARNAME, kVAR_OBJ, "LOOP_VARNAME") && yyResult;
break;
case kPERM_VAR:
yyResult = yyCheckChild (yyt, yyt->PERM_VAR.VAL, kBT_VAR, "VAL") && yyResult;
break;
case kADDR:
yyResult = yyCheckChild (yyt, yyt->ADDR.E, kBT_EXP, "E") && yyResult;
break;
case kDO_VAR:
yyResult = yyCheckChild (yyt, yyt->DO_VAR.DO_ID, kBT_VAR, "DO_ID") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DO_VAR.RANGE, kSLICE_EXP, "RANGE") && yyResult;
yyResult = yyCheckChild (yyt, yyt->DO_VAR.BODY, kBT_VAR_LIST, "BODY") && yyResult;
break;
case kREMOTE_VAR:
yyResult = yyCheckChild (yyt, yyt->REMOTE_VAR.VAR, kBT_VAR, "VAR") && yyResult;
break;
case kBTV_LIST:
yyResult = yyCheckChild (yyt, yyt->BTV_LIST.Elem, kBT_VAR, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BTV_LIST.Next, kBT_VAR_LIST, "Next") && yyResult;
break;
case kVAR_PARAM:
yyResult = yyCheckChild (yyt, yyt->VAR_PARAM.V, kBT_VAR, "V") && yyResult;
break;
case kVALUE_PARAM:
yyResult = yyCheckChild (yyt, yyt->VALUE_PARAM.E, kBT_EXP, "E") && yyResult;
break;
case kNAMED_PARAM:
yyResult = yyCheckChild (yyt, yyt->NAMED_PARAM.VAL, kBT_PARAM, "VAL") && yyResult;
break;
case kPROC_PARAM:
yyResult = yyCheckChild (yyt, yyt->PROC_PARAM.P, kPROC_OBJ, "P") && yyResult;
break;
case kFUNC_PARAM:
yyResult = yyCheckChild (yyt, yyt->FUNC_PARAM.F, kPROC_OBJ, "F") && yyResult;
break;
case kBTP_LIST:
yyResult = yyCheckChild (yyt, yyt->BTP_LIST.Elem, kBT_PARAM, "Elem") && yyResult;
yyResult = yyCheckChild (yyt, yyt->BTP_LIST.Next, kBT_PARAM_LIST, "Next") && yyResult;
break;
 default: ;
 }
 return yyResult;
}

void InitTree
# if defined __STDC__ | defined __cplusplus
 (register tTree yyt)
# else
 (yyt) register tTree yyt;
# endif
{
 register tTree yyr;
 for (;;) {
  yyt->yyHead.yyIsComp0 = 0;
  yyt->yyHead.yyIsDone0 = 0;
  switch (yyt->Kind) {
case kCOMP_UNIT:
yyr = yyt->COMP_UNIT.COMP_ELEMENTS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->COMP_UNIT.USED_ELEMENTS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->COMP_UNIT.USED_ELEMENTS; break;
case kUNIT_LIST:
yyr = yyt->UNIT_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->UNIT_LIST.Next; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->UNIT_LIST.Next; break;
case kMODULE_DECL:
yyr = yyt->MODULE_DECL.MODULE_BODY; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->MODULE_DECL.MODULE_BODY; break;
case kPROGRAM_DECL:
yyr = yyt->PROGRAM_DECL.FORMALS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->PROGRAM_DECL.PROGRAM_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROGRAM_DECL.PROGRAM_BODY; break;
case kPROC_DECL:
yyr = yyt->PROC_DECL.FORMALS; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->PROC_DECL.PROC_BODY; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROC_DECL.PROC_BODY; break;
case kFUNC_DECL:
yyr = yyt->FUNC_DECL.FORMALS; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->FUNC_DECL.FUNC_BODY; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->FUNC_DECL.RESULT_TYPE; yyr->yyHead.yyOffset = 7; yyr->yyHead.yyParent = yyt;
yyt = yyt->FUNC_DECL.RESULT_TYPE; break;
case kBLOCK_DATA_DECL:
yyr = yyt->BLOCK_DATA_DECL.DATA_BODY; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->BLOCK_DATA_DECL.DATA_BODY; break;
case kINTERFACE:
yyr = yyt->INTERFACE.SPEC; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->INTERFACE.ITEMS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->INTERFACE.ITEMS; break;
case kBODY_NODE:
yyr = yyt->BODY_NODE.DECLS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BODY_NODE.STATS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BODY_NODE.INTERNALS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->BODY_NODE.INTERNALS; break;
case kINTEGER_TYPE:
yyr = yyt->INTEGER_TYPE.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->INTEGER_TYPE.SIZE; break;
case kREAL_TYPE:
yyr = yyt->REAL_TYPE.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->REAL_TYPE.SIZE; break;
case kBOOLEAN_TYPE:
yyr = yyt->BOOLEAN_TYPE.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->BOOLEAN_TYPE.SIZE; break;
case kCOMPLEX_TYPE:
yyr = yyt->COMPLEX_TYPE.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->COMPLEX_TYPE.SIZE; break;
case kSTRING_TYPE:
yyr = yyt->STRING_TYPE.LENGTH; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->STRING_TYPE.LENGTH; break;
case kTYPE_ID:
yyr = yyt->TYPE_ID.ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->TYPE_ID.ID; break;
case kARRAY_TYPE:
yyr = yyt->ARRAY_TYPE.ARRAY_INDEX_TYPES; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ARRAY_TYPE.ARRAY_COMP_TYPE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ARRAY_TYPE.ARRAY_COMP_TYPE; break;
case kRECORD_TYPE:
yyr = yyt->RECORD_TYPE.COMPONENTS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->RECORD_TYPE.COMPONENTS; break;
case kPOINTER_TYPE:
yyr = yyt->POINTER_TYPE.PTR_COMP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->POINTER_TYPE.PTR_COMP; break;
case kSHAPE_SPEC:
return;
case kEXPLICIT_SHAPE:
yyr = yyt->EXPLICIT_SHAPE.LOWER; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->EXPLICIT_SHAPE.UPPER; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->EXPLICIT_SHAPE.UPPER; break;
case kASSUMED_SHAPE:
yyr = yyt->ASSUMED_SHAPE.LOWER; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ASSUMED_SHAPE.LOWER; break;
case kDEFERRED_SHAPE:
return;
case kASSUMED_SIZE:
yyr = yyt->ASSUMED_SIZE.LOWER; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ASSUMED_SIZE.LOWER; break;
case kSHAPE_LIST:
yyr = yyt->SHAPE_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SHAPE_LIST.Next; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->SHAPE_LIST.Next; break;
case kTYPE_DECL:
yyr = yyt->TYPE_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->TYPE_DECL.VAL; break;
case kENTITY_DECL:
yyr = yyt->ENTITY_DECL.ATTRIBUTES; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ENTITY_DECL.ATTRIBUTES; break;
case kVAR_DECL:
yyr = yyt->VAR_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->VAR_DECL.VAL; break;
case kCOMMON_DECL:
yyr = yyt->COMMON_DECL.IDS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->COMMON_DECL.IDS; break;
case kNAMELIST_DECL:
yyr = yyt->NAMELIST_DECL.IDS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->NAMELIST_DECL.IDS; break;
case kTEMPLATE_DECL:
yyr = yyt->TEMPLATE_DECL.DIMENSIONS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->TEMPLATE_DECL.DIMENSIONS; break;
case kRAGGED_DECL:
yyr = yyt->RAGGED_DECL.DIMENSIONS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->RAGGED_DECL.DIMENSIONS; break;
case kPROCESSORS_DECL:
yyr = yyt->PROCESSORS_DECL.DIMENSIONS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROCESSORS_DECL.DIMENSIONS; break;
case kSUB_PROCS_DECL:
yyr = yyt->SUB_PROCS_DECL.TOP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SUB_PROCS_DECL.DIMENSIONS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->SUB_PROCS_DECL.DIMENSIONS; break;
case kRANGE_DECL:
yyr = yyt->RANGE_DECL.RANGES; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->RANGE_DECL.RANGES; break;
case kSTMT_FUNC_DECL:
yyr = yyt->STMT_FUNC_DECL.FORMALS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->STMT_FUNC_DECL.RESULT_TYPE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->STMT_FUNC_DECL.FFUNC_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->STMT_FUNC_DECL.FFUNC_BODY; break;
case kINTERFACE_DECL:
yyr = yyt->INTERFACE_DECL.SPEC; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->INTERFACE_DECL.ITEMS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->INTERFACE_DECL.ITEMS; break;
case kVAR_PARAM_DECL:
yyr = yyt->VAR_PARAM_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->VAR_PARAM_DECL.VAL; break;
case kIMPLICIT_DECL:
yyr = yyt->IMPLICIT_DECL.TYPE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->IMPLICIT_DECL.TYPE; break;
case kUSE_DECL:
yyr = yyt->USE_DECL.RENAMINGS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->USE_DECL.RENAMINGS; break;
case kONLY_USE_DECL:
yyr = yyt->ONLY_USE_DECL.NAMES; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ONLY_USE_DECL.NAMES; break;
case kEQV_DECL:
yyr = yyt->EQV_DECL.VARS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->EQV_DECL.VARS; break;
case kDATA_DECL:
yyr = yyt->DATA_DECL.VARS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DATA_DECL.VALS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->DATA_DECL.VALS; break;
case kFORMAT_DECL:
yyr = yyt->FORMAT_DECL.FORMAT_PARAMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->FORMAT_DECL.FORMAT_PARAMS; break;
case kDIMENSION_DECL:
yyr = yyt->DIMENSION_DECL.INDEXES; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->DIMENSION_DECL.INDEXES; break;
case kINIT_DATA_DECL:
yyr = yyt->INIT_DATA_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->INIT_DATA_DECL.VAL; break;
case kTYPESPEC_DECL:
yyr = yyt->TYPESPEC_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->TYPESPEC_DECL.VAL; break;
case kPARAMETER_DECL:
yyr = yyt->PARAMETER_DECL.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PARAMETER_DECL.VAL; break;
case kDISTRIBUTE_DECL:
yyr = yyt->DISTRIBUTE_DECL.DISTRIBUTION; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DISTRIBUTE_DECL.TARGET; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->DISTRIBUTE_DECL.TARGET; break;
case kALIGN_DECL:
yyr = yyt->ALIGN_DECL.ALIGN_SOURCE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ALIGN_DECL.ALIGN_TARGET; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ALIGN_DECL.ALIGN_TARGET; break;
case kMAP_DECL:
yyr = yyt->MAP_DECL.MAP_TARGET; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->MAP_DECL.MAP_TARGET; break;
case kOVERLAP_DECL:
yyr = yyt->OVERLAP_DECL.OVERLAPS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->OVERLAP_DECL.OVERLAPS; break;
case kSELECT_DECL:
yyr = yyt->SELECT_DECL.SELECTIONS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->SELECT_DECL.SELECTIONS; break;
case kDECL_LIST:
yyr = yyt->DECL_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DECL_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->DECL_LIST.Next; break;
case kRANGE_LIST:
yyr = yyt->RANGE_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->RANGE_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->RANGE_LIST.Next; break;
case kSELECT_LIST:
yyr = yyt->SELECT_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SELECT_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->SELECT_LIST.Next; break;
case kNODE_DISTRIBUTION:
yyr = yyt->NODE_DISTRIBUTION.MAPPING; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->NODE_DISTRIBUTION.MAPPING; break;
case kDIST_LIST:
yyr = yyt->DIST_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DIST_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->DIST_LIST.Next; break;
case kBLOCK_DISTRIBUTION:
yyr = yyt->BLOCK_DISTRIBUTION.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->BLOCK_DISTRIBUTION.SIZE; break;
case kCYCLIC_DISTRIBUTION:
yyr = yyt->CYCLIC_DISTRIBUTION.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->CYCLIC_DISTRIBUTION.SIZE; break;
case kGEN_BLOCK_DISTRIBUTION:
yyr = yyt->GEN_BLOCK_DISTRIBUTION.ARRAY; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->GEN_BLOCK_DISTRIBUTION.ARRAY; break;
case kINDIRECT_DISTRIBUTION:
yyr = yyt->INDIRECT_DISTRIBUTION.ARRAY; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->INDIRECT_DISTRIBUTION.ARRAY; break;
case kARBITRARY_DISTRIBUTION:
yyr = yyt->ARBITRARY_DISTRIBUTION.SIZE; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ARBITRARY_DISTRIBUTION.LENGTH; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ARBITRARY_DISTRIBUTION.MAP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ARBITRARY_DISTRIBUTION.MAP; break;
case kPROCESSOR_ARRAY:
yyr = yyt->PROCESSOR_ARRAY.TOPNAME; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROCESSOR_ARRAY.TOPNAME; break;
case kPROCESSOR_SUBSET:
yyr = yyt->PROCESSOR_SUBSET.SUBTOP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->PROCESSOR_SUBSET.FULLTOP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->PROCESSOR_SUBSET.SUBSCRIPTS; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROCESSOR_SUBSET.SUBSCRIPTS; break;
case kALIGN_SPEC:
yyr = yyt->ALIGN_SPEC.TEMPNAME; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ALIGN_SPEC.SUBSCRIPTS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ALIGN_SPEC.SUBSCRIPTS; break;
case kOP_GENERIC_SPEC:
yyr = yyt->OP_GENERIC_SPEC.OPERATOR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->OP_GENERIC_SPEC.OPERATOR; break;
case kACF_LIST:
yyr = yyt->ACF_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_LIST.Next; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_LIST.Next; break;
case kACF_BASIC:
yyr = yyt->ACF_BASIC.BASIC_STMT; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_BASIC.BASIC_STMT; break;
case kACF_IF:
yyr = yyt->ACF_IF.IF_EXP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_IF.THEN_PART; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_IF.ELSE_PART; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_IF.ELSE_PART; break;
case kACF_WHERE:
yyr = yyt->ACF_WHERE.WHERE_EXP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_WHERE.TRUE_PART; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_WHERE.FALSE_PART; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_WHERE.FALSE_PART; break;
case kACF_CASE:
yyr = yyt->ACF_CASE.CASE_EXP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_CASE.CASE_ALTS; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_CASE.CASE_ALTS; break;
case kACF_WHILE:
yyr = yyt->ACF_WHILE.WHILE_EXP; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_WHILE.WHILE_BODY; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_WHILE.WHILE_BODY; break;
case kACF_REPEAT:
yyr = yyt->ACF_REPEAT.REPEAT_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_REPEAT.REPEAT_EXP; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_REPEAT.REPEAT_EXP; break;
case kACF_LOOP:
yyr = yyt->ACF_LOOP.LOOP_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_LOOP.LOOP_BODY; break;
case kACF_DO:
yyr = yyt->ACF_DO.DO_ID; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_DO.DO_RANGE; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_DO.DO_BODY; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_DO.DO_DEP_INFO; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_DO.DO_HOME_INFO; yyr->yyHead.yyOffset = 7; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_DO.DO_HOME_INFO; break;
case kACF_FORALL:
yyr = yyt->ACF_FORALL.FORALL_ID; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_FORALL.FORALL_RANGE; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_FORALL.FORALL_BODY; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_FORALL.FORALL_DEP_INFO; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_FORALL.FORALL_HOME_INFO; yyr->yyHead.yyOffset = 7; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_FORALL.FORALL_HOME_INFO; break;
case kACF_FLOW_GRAPH:
yyr = yyt->ACF_FLOW_GRAPH.FLOW_NODES; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_FLOW_GRAPH.FLOW_NODES; break;
case kACF_ENTRY:
yyr = yyt->ACF_ENTRY.ENTRY_DECL; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_ENTRY.ENTRY_DECL; break;
case kACF_HOME:
yyr = yyt->ACF_HOME.HOME_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_HOME.HOME_BODY; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_HOME.HOME_BODY; break;
case kACF_RESIDENT:
yyr = yyt->ACF_RESIDENT.RESIDENT_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_RESIDENT.RESIDENT_BODY; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_RESIDENT.RESIDENT_BODY; break;
case kACF_NEW:
yyr = yyt->ACF_NEW.NEW_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_NEW.NEW_BODY; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_NEW.NEW_BODY; break;
case kACF_REDUCTION:
yyr = yyt->ACF_REDUCTION.REDUCTION_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_REDUCTION.REDUCTION_LVAR; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_REDUCTION.REDUCTION_FUNC; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_REDUCTION.REDUCTION_BODY; yyr->yyHead.yyOffset = 7; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_REDUCTION.REDUCTION_BODY; break;
case kACF_TASK_REGION:
yyr = yyt->ACF_TASK_REGION.TASK_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_TASK_REGION.TASK_BODY; break;
case kACF_PARALLEL:
yyr = yyt->ACF_PARALLEL.REDUCTION_VARS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_PARALLEL.SHARED_VARS; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_PARALLEL.LOCAL_VARS; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_PARALLEL.PARALLEL_BODY; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_PARALLEL.PARALLEL_BODY; break;
case kACF_CRITICAL:
yyr = yyt->ACF_CRITICAL.CRITICAL_BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_CRITICAL.CRITICAL_BODY; break;
case kACF_ON:
yyr = yyt->ACF_ON.ON_HOME; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_ON.ON_SPECS; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_ON.ON_BODY; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_ON.ON_BODY; break;
case kACF_RM_READ:
yyr = yyt->ACF_RM_READ.LOCAL_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_RM_READ.REMOTE_VAR; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_RM_READ.REMOTE_VAR; break;
case kACF_RM_WRITE:
yyr = yyt->ACF_RM_WRITE.REMOTE_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ACF_RM_WRITE.LOCAL_DATA; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_RM_WRITE.LOCAL_DATA; break;
case kACF_MOVE:
yyr = yyt->ACF_MOVE.MOVE_STMT; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_MOVE.MOVE_STMT; break;
case kACF_BARRIER:
yyr = yyt->ACF_BARRIER.BARRIER_VARS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_BARRIER.BARRIER_VARS; break;
case kACF_UPDATE:
yyr = yyt->ACF_UPDATE.OVERLAP_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->ACF_UPDATE.OVERLAP_VAR; break;
case kON_VAR_CLAUSE:
yyr = yyt->ON_VAR_CLAUSE.ON_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ON_VAR_CLAUSE.ON_VAR; break;
case kON_PROC_CLAUSE:
yyr = yyt->ON_PROC_CLAUSE.ON_PROC; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ON_PROC_CLAUSE.ON_PROC; break;
case kON_LIST:
yyr = yyt->ON_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ON_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ON_LIST.Next; break;
case kON_SPEC:
yyr = yyt->ON_SPEC.ON_VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ON_SPEC.ON_VAL; break;
case kSELECTED_ACF_LIST:
yyr = yyt->SELECTED_ACF_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SELECTED_ACF_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->SELECTED_ACF_LIST.Next; break;
case kSELECTED_ACF_NODE:
yyr = yyt->SELECTED_ACF_NODE.SELECT_LIST; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SELECTED_ACF_NODE.SELECT_ACFS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->SELECTED_ACF_NODE.SELECT_ACFS; break;
case kASSIGN_STMT:
yyr = yyt->ASSIGN_STMT.ASSIGN_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ASSIGN_STMT.ASSIGN_EXP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ASSIGN_STMT.ASSIGN_EXP; break;
case kPTR_ASSIGN_STMT:
yyr = yyt->PTR_ASSIGN_STMT.ASSIGN_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->PTR_ASSIGN_STMT.ASSIGN_EXP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->PTR_ASSIGN_STMT.ASSIGN_EXP; break;
case kLABEL_ASSIGN_STMT:
yyr = yyt->LABEL_ASSIGN_STMT.LABEL_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->LABEL_ASSIGN_STMT.LABEL_VAR; break;
case kCALL_STMT:
yyr = yyt->CALL_STMT.CALL_ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->CALL_STMT.CALL_PARAMS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->CALL_STMT.CALL_PARAMS; break;
case kIO_STMT:
yyr = yyt->IO_STMT.ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IO_STMT.IO_SPECS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IO_STMT.IO_ITEMS; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->IO_STMT.IO_ITEMS; break;
case kASS_GOTO_STMT:
yyr = yyt->ASS_GOTO_STMT.GOTO_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ASS_GOTO_STMT.LABELS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ASS_GOTO_STMT.LABELS; break;
case kCOMP_GOTO_STMT:
yyr = yyt->COMP_GOTO_STMT.GOTO_LABELS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->COMP_GOTO_STMT.GOTO_EXP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->COMP_GOTO_STMT.GOTO_EXP; break;
case kCOMP_IF_STMT:
yyr = yyt->COMP_IF_STMT.IF_EXP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->COMP_IF_STMT.IF_EXP; break;
case kRETURN_STMT:
yyr = yyt->RETURN_STMT.RETURN_EXP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->RETURN_STMT.RETURN_EXP; break;
case kPAUSE_STMT:
yyr = yyt->PAUSE_STMT.PAUSE_CONST; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PAUSE_STMT.PAUSE_CONST; break;
case kSTOP_STMT:
yyr = yyt->STOP_STMT.STOP_CONST; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->STOP_STMT.STOP_CONST; break;
case kFORMAT_STMT:
yyr = yyt->FORMAT_STMT.FORMAT_PARAMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->FORMAT_STMT.FORMAT_PARAMS; break;
case kINHERIT_DSP_STMT:
yyr = yyt->INHERIT_DSP_STMT.VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->INHERIT_DSP_STMT.SHAPE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->INHERIT_DSP_STMT.SHAPE; break;
case kCREATE_DSP_STMT:
yyr = yyt->CREATE_DSP_STMT.VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->CREATE_DSP_STMT.SHAPE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->CREATE_DSP_STMT.SHAPE; break;
case kFREE_DSP_STMT:
yyr = yyt->FREE_DSP_STMT.VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->FREE_DSP_STMT.VAR; break;
case kALLOCATE_STMT:
yyr = yyt->ALLOCATE_STMT.PARAMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->ALLOCATE_STMT.STATUS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->ALLOCATE_STMT.STATUS; break;
case kDEALLOCATE_STMT:
yyr = yyt->DEALLOCATE_STMT.PARAMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DEALLOCATE_STMT.STATUS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->DEALLOCATE_STMT.STATUS; break;
case kNULLIFY_STMT:
yyr = yyt->NULLIFY_STMT.PARAMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->NULLIFY_STMT.PARAMS; break;
case kREDUCE_STMT:
yyr = yyt->REDUCE_STMT.RED_FUNC; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDUCE_STMT.RED_PARAMS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->REDUCE_STMT.RED_PARAMS; break;
case kMOVE_STMT:
yyr = yyt->MOVE_STMT.TARGET; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->MOVE_STMT.SOURCE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->MOVE_STMT.MASK; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->MOVE_STMT.MASK; break;
case kSCATTER_STMT:
yyr = yyt->SCATTER_STMT.ARRAY; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SCATTER_STMT.BASE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SCATTER_STMT.INDEXES; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SCATTER_STMT.MASK; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->SCATTER_STMT.MASK; break;
case kREALIGN_STMT:
yyr = yyt->REALIGN_STMT.ALIGNEE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REALIGN_STMT.ALIGN_SOURCE; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REALIGN_STMT.ALIGN_TARGET; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->REALIGN_STMT.ALIGN_TARGET; break;
case kREDISTRIBUTE_STMT:
yyr = yyt->REDISTRIBUTE_STMT.DISTRIBUTEE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDISTRIBUTE_STMT.DISTRIBUTION; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDISTRIBUTE_STMT.TARGET; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->REDISTRIBUTE_STMT.TARGET; break;
case kREG_SHADOW_GET:
yyr = yyt->REG_SHADOW_GET.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REG_SHADOW_GET.SHADOW_SIZE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->REG_SHADOW_GET.SHADOW_SIZE; break;
case kREG_SHADOW_PUT:
yyr = yyt->REG_SHADOW_PUT.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REG_SHADOW_PUT.SHADOW_SIZE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->REG_SHADOW_PUT.SHADOW_SIZE; break;
case kREG_SHADOW_SET:
yyr = yyt->REG_SHADOW_SET.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REG_SHADOW_SET.SHADOW_SIZE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REG_SHADOW_SET.VAL; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->REG_SHADOW_SET.VAL; break;
case kIND_SHADOW_CREATE:
yyr = yyt->IND_SHADOW_CREATE.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IND_SHADOW_CREATE.INDEX_VAR; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IND_SHADOW_CREATE.TMP_INDEX_VAR; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IND_SHADOW_CREATE.MASK; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
yyt = yyt->IND_SHADOW_CREATE.MASK; break;
case kIND_SHADOW_FREE:
yyr = yyt->IND_SHADOW_FREE.INDEX_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->IND_SHADOW_FREE.INDEX_VAR; break;
case kIND_SHADOW_GET:
yyr = yyt->IND_SHADOW_GET.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->IND_SHADOW_GET.SHADOW_VAR; break;
case kIND_SHADOW_PUT:
yyr = yyt->IND_SHADOW_PUT.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->IND_SHADOW_PUT.SHADOW_VAR; break;
case kIND_SHADOW_SET:
yyr = yyt->IND_SHADOW_SET.SHADOW_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->IND_SHADOW_SET.VAL; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->IND_SHADOW_SET.VAL; break;
case kBROADCAST_STMT:
yyr = yyt->BROADCAST_STMT.ELEMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BROADCAST_STMT.OWNER_TOP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BROADCAST_STMT.OWNER_SPEC; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->BROADCAST_STMT.OWNER_SPEC; break;
case kREDUCTION_STMT:
yyr = yyt->REDUCTION_STMT.REDFUNC; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDUCTION_STMT.ELEMS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDUCTION_STMT.RED_TOP; yyr->yyHead.yyOffset = 4; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->REDUCTION_STMT.RED_SPEC; yyr->yyHead.yyOffset = 5; yyr->yyHead.yyParent = yyt;
yyt = yyt->REDUCTION_STMT.RED_SPEC; break;
case kDIRTY_STMT:
yyr = yyt->DIRTY_STMT.ELEMS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->DIRTY_STMT.ELEMS; break;
case kLABEL_LIST:
yyr = yyt->LABEL_LIST.Next; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->LABEL_LIST.Next; break;
case kCONST_EXP:
yyr = yyt->CONST_EXP.C; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->CONST_EXP.C; break;
case kOP_EXP:
yyr = yyt->OP_EXP.EXP_OP; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->OP_EXP.OPND1; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->OP_EXP.OPND2; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->OP_EXP.OPND2; break;
case kOP1_EXP:
yyr = yyt->OP1_EXP.EXP_OP1; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->OP1_EXP.OPND; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->OP1_EXP.OPND; break;
case kSLICE_EXP:
yyr = yyt->SLICE_EXP.START; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SLICE_EXP.STOP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SLICE_EXP.INC; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->SLICE_EXP.INC; break;
case kVAR_EXP:
yyr = yyt->VAR_EXP.V; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->VAR_EXP.V; break;
case kFUNC_CALL_EXP:
yyr = yyt->FUNC_CALL_EXP.FUNC_ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->FUNC_CALL_EXP.FUNC_PARAMS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->FUNC_CALL_EXP.FUNC_PARAMS; break;
case kNAMED_EXP:
yyr = yyt->NAMED_EXP.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->NAMED_EXP.VAL; break;
case kDO_EXP:
yyr = yyt->DO_EXP.DO_ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DO_EXP.RANGE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DO_EXP.BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->DO_EXP.BODY; break;
case kTYPE_EXP:
yyr = yyt->TYPE_EXP.ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->TYPE_EXP.ELEMENTS; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->TYPE_EXP.ELEMENTS; break;
case kARRAY_EXP:
yyr = yyt->ARRAY_EXP.ELEMENTS; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ARRAY_EXP.ELEMENTS; break;
case kPERM_EXP:
yyr = yyt->PERM_EXP.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PERM_EXP.VAL; break;
case kBOUND_EXP:
yyr = yyt->BOUND_EXP.VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->BOUND_EXP.VAR; break;
case kBTE_LIST:
yyr = yyt->BTE_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BTE_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->BTE_LIST.Next; break;
case kUSED_VAR:
yyr = yyt->USED_VAR.VARNAME; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->USED_VAR.VARNAME; break;
case kINDEXED_VAR:
yyr = yyt->INDEXED_VAR.IND_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->INDEXED_VAR.IND_EXPS; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->INDEXED_VAR.IND_EXPS; break;
case kSUBSTRING_VAR:
yyr = yyt->SUBSTRING_VAR.IND_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SUBSTRING_VAR.IND_EXP; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->SUBSTRING_VAR.IND_EXP; break;
case kSELECTED_VAR:
yyr = yyt->SELECTED_VAR.SELEC_VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->SELECTED_VAR.SELECTOR; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->SELECTED_VAR.SELECTOR; break;
case kLOOP_VAR:
yyr = yyt->LOOP_VAR.LOOP_VARNAME; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->LOOP_VAR.LOOP_VARNAME; break;
case kPERM_VAR:
yyr = yyt->PERM_VAR.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PERM_VAR.VAL; break;
case kADDR:
yyr = yyt->ADDR.E; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->ADDR.E; break;
case kDO_VAR:
yyr = yyt->DO_VAR.DO_ID; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DO_VAR.RANGE; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->DO_VAR.BODY; yyr->yyHead.yyOffset = 3; yyr->yyHead.yyParent = yyt;
yyt = yyt->DO_VAR.BODY; break;
case kREMOTE_VAR:
yyr = yyt->REMOTE_VAR.VAR; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->REMOTE_VAR.VAR; break;
case kBTV_LIST:
yyr = yyt->BTV_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BTV_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->BTV_LIST.Next; break;
case kVAR_PARAM:
yyr = yyt->VAR_PARAM.V; yyr->yyHead.yyOffset = 6; yyr->yyHead.yyParent = yyt;
yyt = yyt->VAR_PARAM.V; break;
case kVALUE_PARAM:
yyr = yyt->VALUE_PARAM.E; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->VALUE_PARAM.E; break;
case kNAMED_PARAM:
yyr = yyt->NAMED_PARAM.VAL; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->NAMED_PARAM.VAL; break;
case kPROC_PARAM:
yyr = yyt->PROC_PARAM.P; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->PROC_PARAM.P; break;
case kFUNC_PARAM:
yyr = yyt->FUNC_PARAM.F; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
yyt = yyt->FUNC_PARAM.F; break;
case kBTP_LIST:
yyr = yyt->BTP_LIST.Elem; yyr->yyHead.yyOffset = 1; yyr->yyHead.yyParent = yyt;
InitTree (yyr);
yyr = yyt->BTP_LIST.Next; yyr->yyHead.yyOffset = 2; yyr->yyHead.yyParent = yyt;
yyt = yyt->BTP_LIST.Next; break;
  default: return;
  }
 }
}

void BeginTree ()
{
}

void CloseTree ()
{
}
